Package: rayimage
Type: Package
Title: Image Processing for Simulated Cameras
Version: 0.26.1
Authors@R: person("Tyler", "Morgan-Wall", email = "tylermw@gmail.com",
    role = c("aut", "cph", "cre"), comment = c(ORCID = "0000-0002-3131-3814"))
Maintainer: Tyler Morgan-Wall <tylermw@gmail.com>
Description: Uses convolution-based techniques to generate simulated camera bokeh, depth of field, and other camera effects, using an image and an optional depth map. Accepts both filename inputs and in-memory array representations of images and matrices, including common raster formats such as 'JPEG', 'PNG', 'TIFF', 'TGA', 'BMP', 'PSD', 'GIF', 'HDR', 'PIC', 'PNM', 'DNG', and 'EXR'. Includes functions to perform 2D convolutions, color correction, colorspace conversion, image/matrix reorientation and resizing, image and text overlays, exposure adjustment, camera vignette effects, and image titles.
License: GPL-3
LazyData: true
Depends: R (>= 4.3.0)
Imports: Rcpp, png, jpeg, grDevices, grid, tiff, systemfonts, ragg
Suggests: magick, testthat (>= 3.0.0), libopenexr (>= 3.4.12-4), cli
LinkingTo: Rcpp, RcppArmadillo, progress, tinydng, stbimageheaders
Encoding: UTF-8
URL: https://www.rayimage.dev,
        https://github.com/tylermorganwall/rayimage
BugReports: https://github.com/tylermorganwall/rayimage/issues
Config/testthat/edition: 3
Config/build/compilation-database: true
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-06-12 05:44:36 UTC; tyler
Author: Tyler Morgan-Wall [aut, cph, cre] (ORCID:
    <https://orcid.org/0000-0002-3131-3814>)
Repository: CRAN
Date/Publication: 2026-06-12 16:30:02 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-06-12 17:31:22 UTC; unix
Archs: rayimage.so.dSYM

Package: rayimage
Type: Package
Title: Image Processing for Simulated Cameras
Version: 0.26.1
Authors@R: person("Tyler", "Morgan-Wall", email = "tylermw@gmail.com",
    role = c("aut", "cph", "cre"), comment = c(ORCID = "0000-0002-3131-3814"))
Maintainer: Tyler Morgan-Wall <tylermw@gmail.com>
Description: Uses convolution-based techniques to generate simulated camera bokeh, depth of field, and other camera effects, using an image and an optional depth map. Accepts both filename inputs and in-memory array representations of images and matrices, including common raster formats such as 'JPEG', 'PNG', 'TIFF', 'TGA', 'BMP', 'PSD', 'GIF', 'HDR', 'PIC', 'PNM', 'DNG', and 'EXR'. Includes functions to perform 2D convolutions, color correction, colorspace conversion, image/matrix reorientation and resizing, image and text overlays, exposure adjustment, camera vignette effects, and image titles.
License: GPL-3
LazyData: true
Depends: R (>= 4.3.0)
Imports: Rcpp, png, jpeg, grDevices, grid, tiff, systemfonts, ragg
Suggests: magick, testthat (>= 3.0.0), libopenexr (>= 3.4.12-4), cli
LinkingTo: Rcpp, RcppArmadillo, progress, tinydng, stbimageheaders
Encoding: UTF-8
URL: https://www.rayimage.dev,
        https://github.com/tylermorganwall/rayimage
BugReports: https://github.com/tylermorganwall/rayimage/issues
Config/testthat/edition: 3
Config/build/compilation-database: true
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-06-12 05:44:36 UTC; tyler
Author: Tyler Morgan-Wall [aut, cph, cre] (ORCID:
    <https://orcid.org/0000-0002-3131-3814>)
Repository: CRAN
Date/Publication: 2026-06-12 16:30:02 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-06-12 20:29:42 UTC; unix
Archs: rayimage.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method("[",rayimg)
S3method(print,rayimg)
export(CS_ACESCG)
export(CS_ADOBE)
export(CS_BT2020)
export(CS_P3D65)
export(CS_SRGB)
export(generate_2d_disk)
export(generate_2d_exponential)
export(generate_2d_gaussian)
export(get_string_dimensions)
export(interpolate_array)
export(plot_image)
export(plot_image_grid)
export(ray_read_image)
export(ray_write_image)
export(render_alpha_outline)
export(render_bokeh)
export(render_boolean_distance)
export(render_bw)
export(render_clamp)
export(render_color_correction)
export(render_convert_colorspace)
export(render_convolution)
export(render_convolution_fft)
export(render_crop)
export(render_exposure)
export(render_gamma_linear)
export(render_image_overlay)
export(render_padding)
export(render_reorient)
export(render_resized)
export(render_sprite_overlay)
export(render_stack)
export(render_text_image)
export(render_title)
export(render_to_display)
export(render_tonemap)
export(render_vibrance)
export(render_vignette)
export(render_white_balance)
import(grDevices)
import(grid)
import(png)
import(stats)
importFrom(Rcpp,evalCpp)
importFrom(ragg,agg_png)
useDynLib(rayimage, .registration = TRUE)

#'@title Preview Focus
#'
#'@description Displays the focal point
#'
#'@param image 3-layer RGB/4-layer RGBA array, `rayimg` class, or filename of an image.
#'@param depthmap Depth map filename or 1d array.
#'@param focus Defaults `0.5`. Depth in which to blur. Minimum 0, maximum 1.
#'@keywords internal
#'@examples
#'#Plot the dragon
preview_focus = function(image, depthmap, focus, imagetype, depthtype) {
  image = ray_read_image(image) #Always output RGBA array
  #Check if file or image before below:
  depthmap = ray_read_image(depthmap, convert_to_array = FALSE)
  if (length(dim(depthmap)) == 3) {
    depthmap = depthmap[,, 1]
  }

  isimage = max(depthmap, na.rm = TRUE) == 1
  if (isimage) {
    maxval = max(depthmap[depthmap != 1], na.rm = TRUE)
  } else {
    maxval = max(depthmap, na.rm = TRUE)
  }
  if (isimage) {
    depthmap[depthmap == 1] = maxval
  }
  depthrange = range(depthmap, na.rm = TRUE)
  range_depth_high = focus + depthrange[2] / 200
  range_depth_low = focus - depthrange[1] / 200
  if (range_depth_high >= maxval) {
    range_depth_high = range_depth_high - range(depthmap)[2] / 200
  }
  if (
    any(depthmap < range_depth_high, na.rm = TRUE) &
      any(depthmap > range_depth_low, na.rm = TRUE)
  ) {
    image[,, 1][depthmap < range_depth_high & depthmap > range_depth_low] = 1
    image[,, 2][depthmap < range_depth_high & depthmap > range_depth_low] = 0
    image[,, 3][depthmap < range_depth_high & depthmap > range_depth_low] = 0
    message(sprintf("Focal range: %g-%g", depthrange[1], depthrange[2]))
    plot_image(image)
  } else {
    message(sprintf(
      "Focus point (%g) not in focal range: %g-%g",
      focus,
      depthrange[1],
      depthrange[2]
    ))
  }
}

#' Flip Left-Right
#'
#' @param x Matrix
#'
#' @return Flipped matrix
#' @keywords internal
#'
#' @examples
#' #Fake example
fliplr = function(x) {
  if (length(dim(x)) == 2) {
    x[, ncol(x):1]
  } else {
    x[, ncol(x):1, ]
  }
}


#' Flip Up-Down
#'
#' @param x Matrix
#'
#' @return Flipped matrix
#' @keywords internal
#'
#' @examples
#' #Fake example
flipud = function(x) {
  if (length(dim(x)) == 2) {
    x[nrow(x):1, ]
  } else {
    x[nrow(x):1, , ]
  }
}

#' @title Build RGB-to-XYZ matrices from primaries + white
#' @param primaries Default `list(r=c(0.713,0.293), g=c(0.165,0.830), b=c(0.128,0.044))`. xy chromaticities (AP1).
#' @param white_xy Default `c(0.32168,0.33767)`. White xy (D60).
#' @return List with 3x3 numeric matrices: `rgb_to_xyz`, `xyz_to_rgb`.
#' @keywords internal
make_rgb_xyz_matrices = function(
  primaries = list(
    r = c(0.713, 0.293),
    g = c(0.165, 0.830),
    b = c(0.128, 0.044)
  ),
  white_xy = c(0.32168, 0.33767) # D60
) {
  xy_to_XYZ = function(xy) c(xy[1] / xy[2], 1, (1 - xy[1] - xy[2]) / xy[2])
  Xr = xy_to_XYZ(primaries$r)
  Xg = xy_to_XYZ(primaries$g)
  Xb = xy_to_XYZ(primaries$b)
  M = cbind(Xr, Xg, Xb)
  W = xy_to_XYZ(white_xy)
  S = solve(M, W)
  rgb_to_xyz = M %*% diag(S)
  xyz_to_rgb = solve(rgb_to_xyz)
  list(rgb_to_xyz = rgb_to_xyz, xyz_to_rgb = xyz_to_rgb)
}

#' @title Build a colorspace descriptor
#' @param name Default `"ACEScg"`.
#' @param primaries Default AP1.
#' @param white_xy Default D60.
#' @return list(name, rgb_to_xyz, xyz_to_rgb, white_xyz, white_name)
#' @keywords internal
make_colorspace = function(
  name = "ACEScg",
  primaries = list(
    r = c(0.713, 0.293),
    g = c(0.165, 0.830),
    b = c(0.128, 0.044)
  ),
  white_xy = c(0.32168, 0.33767), # D60
  white_name = "D60"
) {
  mats = make_rgb_xyz_matrices(primaries = primaries, white_xy = white_xy)
  xy_to_XYZ = function(xy) c(xy[1] / xy[2], 1, (1 - xy[1] - xy[2]) / xy[2])
  list(
    name = name,
    primaries = primaries,
    rgb_to_xyz = mats$rgb_to_xyz,
    xyz_to_rgb = mats$xyz_to_rgb,
    white_xyz = xy_to_XYZ(white_xy),
    white_name = white_name
  )
}

#' Prebuilt RGB Working/Display Color Spaces
#'
#' @title Prebuilt RGB Working/Display Color Spaces
#'
#' @description
#' These objects are **color space descriptors** used by `rayimage` to tag images
#' and to convert between RGB spaces. Each descriptor is a named list with:
#'
#' - `name`: Character. Human-readable name of the space.
#' - `primaries`: List with `r`, `g`, `b` entries giving xy chromaticities.
#' - `rgb_to_xyz`: 3x3 numeric matrix (linear RGB -> CIE XYZ, Y=1 white).
#' - `xyz_to_rgb`: 3x3 numeric matrix (CIE XYZ -> linear RGB).
#' - `white_xyz`: Length-3 numeric XYZ of the reference white (normalized s.t. Y = 1).
#' - `white_name`: Character label for the white (e.g., `"D60"`, `"D65"`).
#'
#' These are used as:
#' - `attr(img, "colorspace")`: the current working/display space for an image.
#' - `attr(img, "white_current")`: the current assumed scene/display white (XYZ, Y=1).
#'
#' @details
#' - **CS_ACESCG**: ACEScg (AP1 primaries, D60, scene-linear). Recommended working space.
#' - **CS_SRGB**: sRGB/Rec.709 (D65). Typical display space; used for PNG/JPEG/TIFF output.
#' - **CS_P3D65**: Display P3 (D65). Wide-gamut display space.
#' - **CS_BT2020**: BT.2020 (D65). Very wide-gamut; common for HDR mastering.
#' - **CS_ADOBE**: Adobe RGB (D65). Photo workflows.
#'
#' @seealso
#' - [render_convert_colorspace()] to convert images between spaces.
#' - [render_white_balance()] for Bradford CAT within a working space.
#' - [plot_image()], [ray_write_image()] for display/output conversion.
#'
#'@examplesIf interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")
#'   # Tag a raw array as sRGB, then convert to ACEScg:
#'   arr = array(runif(64*64*4), dim = c(64, 64, 4))
#'   ri  = ray_read_image(arr, assume_colorspace = CS_SRGB, assume_white = CS_SRGB$white_xyz)
#'   riA = render_convert_colorspace(ri, to_mats = CS_ACESCG)
#'   plot_image(riA) # display-converted to sRGB automatically
#'
#'   # Convert an sRGB JPEG to ACEScg on ingest:
#'   # img = ray_read_image("photo.jpg", normalize = FALSE)
#'   # img_aces = render_convert_colorspace(img, to_mats = CS_ACESCG)
#'
#' @name colorspace_descriptors
NULL

#' @rdname colorspace_descriptors
#' @export
CS_ACESCG = make_colorspace(
  name = "ACEScg",
  primaries = list(
    r = c(0.713, 0.293),
    g = c(0.165, 0.830),
    b = c(0.128, 0.044)
  ),
  white_xy = c(0.32168, 0.33767), # D60
  white_name = "D60"
)

#' @rdname colorspace_descriptors
#' @export
CS_SRGB = make_colorspace(
  name = "sRGB",
  primaries = list(
    r = c(0.64, 0.33),
    g = c(0.30, 0.60),
    b = c(0.15, 0.06)
  ),
  white_xy = c(0.3127, 0.3290), # D65
  white_name = "D65"
)

#' @rdname colorspace_descriptors
#' @export
CS_P3D65 = make_colorspace(
  name = "P3-D65",
  primaries = list(
    r = c(0.680, 0.320),
    g = c(0.265, 0.690),
    b = c(0.150, 0.060)
  ),
  white_xy = c(0.3127, 0.3290), # D65
  white_name = "D65"
)

#' @rdname colorspace_descriptors
#' @export
CS_BT2020 = make_colorspace(
  name = "BT.2020",
  primaries = list(
    r = c(0.708, 0.292),
    g = c(0.170, 0.797),
    b = c(0.131, 0.046)
  ),
  white_xy = c(0.3127, 0.3290), # D65
  white_name = "D65"
)

#' @rdname colorspace_descriptors
#' @export
CS_ADOBE = make_colorspace(
  name = "AdobeRGB",
  primaries = list(
    r = c(0.64, 0.33),
    g = c(0.21, 0.71),
    b = c(0.15, 0.06)
  ),
  white_xy = c(0.3127, 0.3290), # D65
  white_name = "D65"
)

#' @title Get String Dimensions
#'
#' @description
#' Calculates font metrics for a specified font, font size, and style.
#'
#' @param string The string to be measured.
#' @param font Default `"sans"`.
#' @param size A numeric value specifying the size of the font in points.
#' @param align Default `"left"``. The string alignment.
#' @param ... Other arguments to pass to `systemfonts::shape_string()``
#'
#' @return A data.frame listing the string dimensions.
#'
#' @details
#' The function renders specific characters (`"d"` for ascender, `"g"` for descender, `"x"` for neutral) using the specified font parameters. It calculates the bounding box of each character to determine the necessary adjustments for accurate text positioning.
#'
#' @import grDevices grid png
#' @export
#' @examples
#' # Get height of basic sans font
#' get_string_dimensions("This is a string", size=24)
get_string_dimensions = function(
  string,
  font = "sans",
  size = 12,
  align = "center",
  ...
) {
  size_metrics = systemfonts::shape_string(
    string,
    family = font,
    size = size,
    align = align,
    ...
  )[["metrics"]]
  return(size_metrics)
}

#' Internal wrapper for stb_image C++ loader
#'
#' @description
#' Internal helper that loads an image via the C++ `load_image_stb()` function
#' and returns a 3D RGBA numeric array with dimensions `c(height, width, channels)`.
#'
#' @param filename Default `NULL`. Path to an image file to load.
#' @param desired_channels Default `4L`. Number of channels to request from the loader.
#'
#' @return A numeric 3D array (height, width, `desired_channels`).
#' @keywords internal
#' @noRd
read_image_stb = function(filename = NULL, desired_channels = 4L) {
  if (!is.character(filename) || length(filename) != 1L) {
    stop("'read_image_stb' expects a single filename string.")
  }

  filename = normalizePath(filename, winslash = "/", mustWork = TRUE)

  # Dummy values required by the Rcpp-exported interface
  width = 0L
  height = 0L
  channels = 0L

  img = load_image_stb(
    filename = filename,
    width = width,
    height = height,
    channels = channels,
    desired_channels = as.integer(desired_channels)
  )

  if (!is.numeric(img)) {
    stop("stb loader did not return a numeric array.")
  }
  if (is.null(dim(img)) || length(dim(img)) != 3L) {
    stop("stb loader did not return a 3D array.")
  }
  if (dim(img)[3L] != desired_channels) {
    stop(
      sprintf(
        "stb loader returned %d channels; expected %d.",
        dim(img)[3L],
        desired_channels
      )
    )
  }

  storage.mode(img) = "double"
  img
}

#'@title Render Bokeh
#'
#'@description Takes an image and a depth map to render the image with depth of field
#'(i.e. similar to "Portrait Mode" in an iPhone). User can specify a custom bokeh shape,
#'or use one of the built-in bokeh types.
#'
#'@param image 3-layer RGB/4-layer RGBA array, `rayimg` class, or filename of an image.
#'@param depthmap Depth map filename or 1d array.
#'@param focus Defaults `0.5`. Depth in which to blur.
#'@param focallength Default `100`. Focal length of the virtual camera.
#'@param fstop Default `4`. F-stop of the virtual camera.
#'@param filename Default `NULL`. The filename of the image to be saved. If this is not given, the image will be plotted instead.
#'@param preview Default `TRUE`. If `FALSE`, it will not display the image and just return the RGB array.
#'@param preview_focus Default `FALSE`. If `TRUE`, a red line will be drawn across the image
#'showing where the camera will be focused.
#'@param bokehshape Default `circle`. Also built-in: `hex`. The shape of the bokeh. If the user
#'passes in a 2D matrix, that matrix will control the shape of the bokeh.
#'@param bokehintensity Default `1`. Intensity of the bokeh when the pixel intensity is greater than `bokehlimit`.
#'@param bokehlimit Default `0.8`. Limit after which the bokeh intensity is increased by `bokehintensity`.
#'@param rotation Default `0`. Number of degrees to rotate the hexagonal bokeh shape.
#'@param aberration Default `0`. Adds chromatic aberration to the image. Maximum of `1`.
#'@param progress Default `TRUE`. Whether to display a progress bar.
#'@param ... Additional arguments to pass to [plot_image()] if `preview = TRUE`.
#'@return A `rayimg` RGBA array.
#'@export
#'@examplesIf interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")
#'#Plot the dragon
#'plot_image(dragon)
#'#Plot the depth map
#' plot_image(dragondepth/1500)
#'#Preview the focal plane:
#'render_bokeh(dragon, dragondepth, focus=950, preview_focus = TRUE)
#'#Change the focal length:
#'render_bokeh(dragon, dragondepth, focus=950, focallength=300)
#'#Add chromatic aberration:
#'render_bokeh(dragon, dragondepth, focus=950, focallength=300, aberration = 0.5)
#'#Change the focal distance:
#'render_bokeh(dragon, dragondepth, focus=600, focallength=300)
#'render_bokeh(dragon, dragondepth, focus=1300, focallength=300)
#'#Change the bokeh shape to a hexagon:
#'render_bokeh(dragon, dragondepth, bokehshape = "hex",
#'             focallength=300, focus=600)
#'#Change the bokeh intensity:
#'render_bokeh(dragon, dragondepth,
#'             focallength=400, focus=900, bokehintensity = 1)
#'render_bokeh(dragon, dragondepth,
#'             focallength=400, focus=900, bokehintensity = 3)
#'#Rotate the hexagonal shape:
#'render_bokeh(dragon, dragondepth, bokehshape = "hex", rotation=15, bokehintensity=3,
#'             focallength=400, focus=800)
render_bokeh = function(
  image,
  depthmap,
  focus = 0.5,
  focallength = 100,
  fstop = 4,
  filename = NULL,
  preview = TRUE,
  preview_focus = FALSE,
  bokehshape = "circle",
  bokehintensity = 1,
  bokehlimit = 0.8,
  rotation = 0,
  aberration = 0,
  progress = interactive(),
  ...
) {
  temp_image = ray_read_image(image, reset_camera_settings = TRUE) #Always output RGBA array
  #Check if file or image before below:
  imagetype = attr(temp_image, "filetype")

  depthmap = ray_read_image(
    depthmap,
    convert_to_array = FALSE,
    reset_camera_settings = TRUE
  )
  depthtype = attr(depthmap, "filetype")

  if (preview_focus) {
    preview_focus(image, depthmap, focus, imagetype, depthtype)
    return(invisible())
  }
  if (is.matrix(bokehshape)) {
    custombokeh = bokehshape
    bokehshape = 2L
  } else {
    if (bokehshape == "circle") {
      bokehshape = 0L
    } else {
      bokehshape = 1L
    }
    custombokeh = matrix(1, 1, 1)
  }
  if (aberration >= 1 || aberration <= -1) {
    stop("aberration value must be less than 1 and greater than -1")
  }

  if (length(dim(depthmap)) == 3) {
    depthmap = depthmap[,, 1]
  }

  depthmap[is.na(depthmap)] = max(depthmap, na.rm = TRUE) * 2

  for (i in 1:3) {
    max_size = min(c(max(dim(depthmap)), 500))
    if (i == 1) {
      depthmap2 = calc_bokeh_size(
        depthmap,
        focus,
        focallength,
        fstop,
        1 + aberration
      )
    } else if (i == 2) {
      depthmap2 = calc_bokeh_size(depthmap, focus, focallength, fstop, 1)
    } else {
      depthmap2 = calc_bokeh_size(
        depthmap,
        focus,
        focallength,
        fstop,
        1 - aberration
      )
    }
    if (any(dim(depthmap2)[1:2] != dim(temp_image)[1:2])) {
      stop(sprintf(
        "{rayimage}: dimensions of image (%ix%i) and depth map (%ix%i) don't match",
        dim(temp_image)[1],
        dim(temp_image)[2],
        dim(depthmap2)[1],
        dim(depthmap2)[2]
      ))
    }
    depthmap2[depthmap2 > max_size] = max_size
    temp_image[,, i] = psf(
      temp_image[,, i],
      depthmap2,
      depthmap,
      focus,
      bokehshape,
      custombokeh = custombokeh,
      bokehintensity,
      bokehlimit,
      rotation,
      progbar = progress,
      channel = i
    )
  }
  handle_image_output(temp_image, filename = filename, preview = preview)
}

#' @title Render White Balance (Bradford CAT)
#' @description Chromatic-adapt an image from `reference_white` to `target_white` **in its working RGB space**.
#'
#' When `bake = FALSE` (the default), this should not result in any visible changes to the image when
#' plotted in rayimage via [plot_image()], as that function then corrects the white balance for
#' display. Set
#'
#' @param image Default `NULL`. 3-layer RGB/4-layer RGBA array, `rayimg`, or filename.
#' @param reference_white Default `NA`. Source white (XYZ Y=1 or named); when `NA`, uses `attr(img, "white_current")`.
#' @param target_white Default `"D60"`. Target white (XYZ Y=1 or named).
#' @param bake Default `FALSE`. `FALSE` = appearance-preserving
#'   (update tag to the new white). When `TRUE`, apply CAT but keep the original tag,
#'   producing a visible warmer/cooler result on display.
#' @param filename Default `NULL`. Output path.
#' @param preview Default `FALSE`. If `TRUE`, display the image.
#' @param bake Default `FALSE`. If `TRUE`, this will bake the color change into the image, and return the
#' old tag. This can be used for stylistic adjustments.
#' @export
#' @examples
#' # Not specifying a target white balance returns an unchanged image
#' render_white_balance(dragon, preview = TRUE)
#'
#' # Default white "D60", converting to D50 produces a warmer image
#' render_white_balance(dragon, preview = TRUE, target_white = "D50", bake = TRUE)
#'
#' # Default white "D60", converting to "D75" produces a cooler image
#' render_white_balance(dragon, preview = TRUE, target_white = "D75", bake = TRUE)
#'
#' # Set reference white to colder "D75", converting to "D50" produces a very warm image
#' render_white_balance(dragon, preview = TRUE,
#'                      reference_white= "D75", target_white = "D50", bake = TRUE)
#'
#' # With a real, non-rendered image:
#' render_white_balance(sunset_image, preview=TRUE)
#'
#' # Cooler
#' render_white_balance(sunset_image, preview=TRUE, target_white = "D75", bake = TRUE)
#'
#' # Warmer
#' render_white_balance(sunset_image, preview=TRUE, target_white = "D50", bake = TRUE)
render_white_balance = function(
	image,
	reference_white = NA,
	target_white = "D60",
	bake = FALSE,
	filename = NULL,
	preview = FALSE
) {
	src = ray_read_image(image, normalize = FALSE)
	cs = attr(src, "colorspace")

	if (
		is.atomic(reference_white) &&
			length(reference_white) == 1L &&
			is.na(reference_white)
	) {
		reference_white = attr(src, "white_current")
	}

	# If whites are identical, do nothing
	if (
		isTRUE(all.equal(
			get_whitepoint_xyz(reference_white)$value,
			get_whitepoint_xyz(target_white)$value
		))
	) {
		return(handle_image_output(src, filename = filename, preview = preview))
	}

	if (!bake) {
		# CAT + update tag (appearance-preserving)
		return(render_convert_colorspace(
			src,
			from_mats = cs,
			to_mats = cs,
			adapt_white = TRUE,
			from_white = reference_white,
			to_white = target_white,
			filename = filename,
			preview = preview
		))
	}

	# bake = TRUE, CAT but keep the original tag ----
	src_wp = get_whitepoint_xyz(reference_white)
	dst_wp = get_whitepoint_xyz(target_white)

	# Work in linear RGB
	if (!isTRUE(attr(src, "source_linear"))) {
		warning(
			"render_white_balance(bake = TRUE): input not linear; convert with render_gamma_linear(..., TRUE) first."
		)
	}

	# RGB(working) -> XYZ
	xyz = apply_color_matrix(src, cs$rgb_to_xyz)
	# CAT in XYZ
	CAT = compute_cat_bradford(src_wp$value, dst_wp$value)
	xyz = apply_color_matrix(xyz, CAT)
	# XYZ -> RGB(working)
	out = apply_color_matrix(xyz, cs$xyz_to_rgb)
	# preserve alpha
	d = dim(src)
	if (length(d) == 3L && d[3] == 4L) {
		out[,, 4] = src[,, 4]
	}

	out = rayimg(
		out,
		filetype = attr(src, "filetype"),
		source_linear = attr(src, "source_linear"),
		colorspace = cs, # primaries unchanged
		white_current = src_wp$value, # Keep the original tag to bake in the new color
		exposure = attr(src, "exposure", exact = TRUE),
		iso = attr(src, "iso", exact = TRUE)
	)
	handle_image_output(out, filename = filename, preview = preview)
}

#' @title Generate Text Image
#'
#' @description Generates an image which tightly fits text.
#'
#' @param text Text to turn into an image.
#' @param color Default `"black"`. String specifying the color of the text.
#' @param size Default `12`. Numeric value specifying the font size of the text.
#' @param font Default `"sans"`. String specifying the font family for the text.
#' Common options include `"sans"`, `"mono"`, `"serif"`, `"Times"`, `"Helvetica"`, etc.
#' @param just Default `"left"`. Horizontal alignment of the text: `"left"`,
#' `"center"`, or `"right"`.
#' @param lineheight Default `1`. Multiplier for the lineheight.
#' @param background_color Default `"white"`. Color of the background.
#' @param background_alpha Default `1`. Transparency of the background. A value
#' between `0` (fully transparent) and `1` (fully opaque).
#' @param use_ragg Default `TRUE`. Whether to use the `ragg` package as the graphics device. Required for emojis.
#' @param width Default `NA`. User-defined textbox width.
#' @param height Default `NA`. User-defined textbox height.
#' @param check_text_width Default `TRUE`. Whether to manually adjust the bounding
#' box of the resulting image to ensure the string bbox is wide enough for the text.
#' Not all systems provide accurate font sizes: this ensures the string is not cut off
#' at the edges, at the cost of needing to repeatedly render the image internally until
#' a suitable image is found.
#' @param check_text_height Default `FALSE`. Whether to manually adjust the bounding
#' box of the resulting image to ensure the string bbox is tall enough for the text.
#' This will ensure a tight vertical bounding box on the text.
#' @param trim Default `FALSE`. If `TRUE`, post-process the rendered image by
#' cropping it to the tight bounding box of the text or emoji pixels. The
#' bounding box is measured from a transparent text mask, so this can also trim
#' images with opaque backgrounds.
#' @param trim_padding Default `0`. Padding in pixels to add after trimming. Use
#' a scalar for equal padding on all sides, or `c(top, right, bottom, left)`.
#' Nonzero padding implies `trim = TRUE`.
#' @param filename Default `NULL`. String specifying the file path to save the resulting image.
#' If `NULL` and `preview = FALSE`, the function returns the processed RGB array.
#' @param preview Default `FALSE`. Boolean indicating whether to display the image after processing.
#' If `TRUE`, the image is displayed but not saved or returned.
#'
#' @return A 3-layer RGB array of the processed image if `filename = NULL` and `preview = FALSE`.
#' Otherwise, writes the image to the specified file or displays it if `preview = TRUE`.
#'
#'
#' @import grDevices grid
#' @export
#'
#'@examplesIf interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")
#'   #Generate an image of some text
#'   render_text_image("Some text", preview = TRUE)
#'   #Change the font size
#'   render_text_image("Some text", size = 100, preview = TRUE)
#'   #Change the font color
#'   render_text_image("Some text", size = 100, color="red",preview = TRUE)
#'   #Change the background color and transparency
#'   render_text_image("Some text", size = 50, color="purple",
#'                       background_color="purple", background_alpha = 0.5,
#'                       preview = TRUE)
#'   # Plot an emoji with the agg device.
#'   render_text_image("\U0001F600\U0001F680", size = 50, color = "purple", use_ragg = TRUE,
#'                     background_alpha = 0,
#'                     preview = TRUE)
#'
#'   # Plot an emoji with the agg device and adjust the height and width (which
#'   # is on by default) to be a tight fit.
#'   render_text_image("\U0001F600\U0001F680", size = 50, color = "purple", use_ragg = TRUE,
#'                     background_alpha = 0, check_text_width = TRUE,
#'                     check_text_height = TRUE,
#'                     preview = TRUE)
#'
#'   # Trim the image to the rendered glyphs and then add transparent padding.
#'   render_text_image("\U0001F409", size = 80, use_ragg = TRUE,
#'                     background_alpha = 0, trim = TRUE, trim_padding = 8,
#'                     preview = TRUE)
render_text_image = function(
  text,
  lineheight = 1,
  color = "black",
  size = 12,
  font = "sans",
  just = "left",
  background_color = "white",
  background_alpha = 1,
  use_ragg = TRUE,
  width = NA,
  height = NA,
  filename = NULL,
  check_text_width = TRUE,
  check_text_height = TRUE,
  trim = FALSE,
  trim_padding = 0,
  preview = FALSE
) {
  if (!is.logical(trim) || length(trim) != 1 || is.na(trim)) {
    stop("`trim` must be `TRUE` or `FALSE`.")
  }
  trim_padding = normalize_render_padding(trim_padding)
  trim_padding_values = unname(trim_padding)
  if (any(trim_padding_values > 0)) {
    trim = TRUE
  }
  text_metrics = systemfonts::shape_string(
    text,
    size = size,
    vjust = 0.5,
    res = 72,
    lineheight = lineheight,
    family = font
  )[["metrics"]]
  bg_col = convert_color(background_color, as_hex = TRUE)
  metrics_string = systemfonts::string_metrics_dev(
    text,
    unit = "inches",
    size = size
  ) *
    72
  max_total_width = ceiling(metrics_string$width) + 2
  max_total_height = ceiling(text_metrics$height)
  n_newlines = sum(strsplit(text, "") == "\n")
  if (!is.na(width)) {
    max_total_width = width
  }
  if (!is.na(height)) {
    max_total_height = height
  }
  final_array = array(
    1,
    dim = c(
      max_total_height,
      max_total_width,
      4
    )
  )
  temp_filename = tempfile(fileext = ".png")
  ray_write_image(final_array, temp_filename)

  temp_image = png::readPNG(temp_filename)
  image_width = ncol(temp_image)
  image_height = nrow(temp_image)

  render_text_alpha_mask = function(image_width, image_height) {
    png_device = linear_png_device(use_ragg = use_ragg)
    png_device(
      temp_filename,
      width = image_width,
      height = image_height,
      units = "px",
      pointsize = size,
      family = font,
      res = 72,
      bg = NA
    )
    dev_id = grDevices::dev.cur()
    close_device = function() {
      devs = grDevices::dev.list()
      if (!is.null(devs) && dev_id %in% devs) {
        grDevices::dev.off(which = dev_id)
      }
    }
    on.exit(close_device(), add = TRUE)

    #First write with no background
    grid::grid.newpage()
    grid::grid.rect(
      x = 0.5,
      y = 0.5,
      width = 1,
      height = 1,
      just = c("center"),
      default.units = "npc",
      gp = grid::gpar(
        fill = grDevices::adjustcolor("black", alpha.f = 0),
        col = NA
      )
    )
    grid::grid.text(
      label = text,
      x = 0.5,
      y = 0.5,
      default.units = "npc",
      just = c("center", "center"),
      gp = grid::gpar(
        fontsize = size,
        lineheight = lineheight,
        cex = 1,
        col = "black",
        fontfamily = font
      )
    )
    close_device()
    temp = png::readPNG(temp_filename)
    temp[,, 4]
  }

  test_edges = function(image_width, image_height) {
    text_alpha = render_text_alpha_mask(image_width, image_height)
    side_edges = text_alpha[c(1, nrow(text_alpha)), ]
    vert_edges = text_alpha[, c(1, ncol(text_alpha))]
    return(c(any(vert_edges > 0), any(side_edges > 0)))
  }

  #First write with no background
  if (check_text_width || check_text_height) {
    test_edge_vec = test_edges(image_width, image_height)
    if (!check_text_height) {
      test_edge_vec[2] = FALSE
    }
    if (!check_text_width) {
      test_edge_vec[1] = FALSE
    }
    while (any(test_edge_vec)) {
      if (test_edge_vec[1]) {
        image_width = image_width * 2
      }
      if (check_text_height) {
        if (test_edge_vec[2]) {
          image_height = image_height * 2
        }
      }
      test_edge_vec = test_edges(image_width, image_height)
    }

    temp = png::readPNG(temp_filename)
    stopifnot(any(temp[,, 4] != 0))

    vert_bbox = range(which(apply(temp[,, 4], 1, sum) != 0))
    hori_bbox = range(which(apply(temp[,, 4], 2, sum) != 0))

    vert_blank = (vert_bbox[1] - 1) + (image_height - vert_bbox[2])
    hori_blank = (hori_bbox[1] - 1) + (image_width - hori_bbox[2])
    if (check_text_height) {
      image_height = image_height - vert_blank + size
    }
    image_width = image_width - hori_blank + size
  }

  #If no edges, proceed with normal render
  png_device = linear_png_device(use_ragg = use_ragg)
  png_device(
    temp_filename,
    width = image_width,
    height = image_height,
    units = "px",
    pointsize = size,
    family = font,
    res = 72,
    bg = NA
  )

  dev_id = grDevices::dev.cur()
  grid::grid.newpage()
  grid::grid.rect(
    x = 0.5,
    y = 0.5,
    width = 1,
    height = 1,
    just = c("center"),
    default.units = "npc",
    gp = grid::gpar(
      fill = grDevices::adjustcolor(bg_col, alpha.f = background_alpha),
      col = NA
    )
  )

  grid::grid.text(
    label = text,
    x = 0.5,
    y = 0.5,
    default.units = "npc",
    just = c("center", "center"),
    gp = grid::gpar(
      fontsize = size,
      lineheight = lineheight,
      cex = 1,
      col = convert_color(color, as_hex = TRUE),
      fontfamily = font
    )
  )
  grDevices::dev.off(which = dev_id)

  temp = ray_read_image(temp_filename, reset_camera_settings = TRUE)
  if (trim) {
    text_alpha = render_text_alpha_mask(image_width, image_height)
    text_rows = which(rowSums(text_alpha > 0) > 0)
    text_cols = which(colSums(text_alpha > 0) > 0)
    if (length(text_rows) > 0 && length(text_cols) > 0) {
      temp = temp[
        min(text_rows):max(text_rows),
        min(text_cols):max(text_cols),
        ,
        drop = FALSE
      ]
    }
  }
  if (any(trim_padding_values > 0)) {
    temp = render_padding(
      temp,
      pad = trim_padding,
      color = background_color,
      alpha = background_alpha,
      preview = FALSE
    )
  }

  handle_image_output(temp, filename = filename, preview = preview)
}

#' @title Render Convert Colorspace
#' @description Convert between two RGB spaces via XYZ, with optional white adaptation. Operates on **linear** RGB.
#'
#' @param image 3-layer RGB/4-layer RGBA array, `rayimg`, or filename.
#' @param from_mats Default `NA`. Source space object; when `NA`, pulled from `attr(src,"colorspace")`.
#' @param to_mats   Default `CS_ACESCG`. Target space object.
#' @param adapt_white Default `TRUE`. Apply Bradford CAT from `from_white` to `to_white`.
#' @param from_white Default `NA`. Source white (name or XYZ, Y=1). When `NA`, uses `attr(src,"white_current")`.
#' @param to_white   Default `NA`. Target white (name or XYZ, Y=1). When `NA`, uses `to_mats$white_xyz`.
#' @param filename Default `NULL`. Output path.
#' @param preview  Default `FALSE`. If `TRUE`, display the image.
#' @return A `rayimg` RGBA array tagged with the **target** space.
#' @export
#'@examplesIf interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")
#' # Read photo, convert to ACEScg with CAT (scene)
#' 	photo = ray_read_image(sunset_image, normalize = FALSE)
#' 	photo_aces = render_convert_colorspace(
#' 		photo,
#' 		to_mats = CS_ACESCG,
#' 		adapt_white = TRUE
#' 	)
#' 	tmp_txt = tempfile(fileext = ".png")
#' 	render_text_image(
#' 		"Sunset",
#' 		size = 60,
#' 		filename = tmp_txt,
#' 		color = "#c300ffff",
#' 		background_alpha = 0
#' 	)
#' 	# Read logo (display-referred), convert primaries only (no CAT)
#' 	logo = ray_read_image(tmp_txt, normalize = FALSE) # sRGB/D65
#' 	logo_aces = render_convert_colorspace(
#' 		logo,
#' 		to_mats = CS_ACESCG,
#' 		adapt_white = FALSE
#' 	)
#'
#' 	# Composite in ACEScg, then display (plot_image converts to sRGB/D65 + OETF)
#' 	# Here, we also turn overlay conversion in [render_image_overlay()] off,
#'  # to show what happens when you don't account for the colorspace difference.
#'  # By default [render_image_overlay()] will do this for you.
#' 	comp1 = render_image_overlay(
#' 		photo_aces,
#' 		logo_aces,
#' 		convert_overlay_colorspace = FALSE
#' 	)
#' 	comp2 = render_image_overlay(
#' 		photo_aces,
#' 		logo,
#' 		convert_overlay_colorspace = FALSE
#' 	)
#'   #Note the color differences, which arise from the mismatched colorspace on the right.
#' 	 plot_image_grid(list(comp1, comp2), dim = c(1, 2))
render_convert_colorspace = function(
  image,
  from_mats = NA,
  to_mats = CS_ACESCG,
  adapt_white = TRUE,
  from_white = NA,
  to_white = NA,
  filename = NULL,
  preview = FALSE
) {
  og_src = ray_read_image(image, normalize = FALSE)
  d = dim(og_src)
  if (length(d) != 3L || d[3] < 3L) {
    return(handle_image_output(og_src, filename = filename, preview = preview))
  }
  src = ray_read_image(image, normalize = FALSE)

  if (is.atomic(from_mats) && length(from_mats) == 1L && is.na(from_mats)) {
    from_mats = attr(src, "colorspace")
  }
  stopifnot(is.list(from_mats), is.list(to_mats))

  if (is.atomic(from_white) && length(from_white) == 1L && is.na(from_white)) {
    from_white = attr(src, "white_current")
  }
  if (is.atomic(to_white) && length(to_white) == 1L && is.na(to_white)) {
    to_white = to_mats$white_xyz
  }

  src_wp = get_whitepoint_xyz(from_white)
  dst_wp = get_whitepoint_xyz(to_white)

  # Idempotence: same primaries, no CAT, and equal whites
  same_mats = isTRUE(all.equal(from_mats$rgb_to_xyz, to_mats$rgb_to_xyz)) &&
    isTRUE(all.equal(from_mats$xyz_to_rgb, to_mats$xyz_to_rgb))
  same_white = isTRUE(all.equal(src_wp$value, dst_wp$value))
  if (same_mats && (!adapt_white || same_white)) {
    return(handle_image_output(src, filename = filename, preview = preview))
  }

  # Assert linear data
  if (!isTRUE(attr(src, "source_linear"))) {
    warning(
      "render_convert_colorspace(): input is not linear; convert with render_gamma_linear(..., TRUE) first."
    )
  }

  # RGB(from) -> XYZ
  xyz = apply_color_matrix(src, from_mats$rgb_to_xyz)

  # Optional CAT in XYZ
  if (adapt_white && !isTRUE(all.equal(src_wp$value, dst_wp$value))) {
    CAT = compute_cat_bradford(src_wp$value, dst_wp$value)
    xyz = apply_color_matrix(xyz, CAT)
  }

  # XYZ -> RGB(to)
  out = apply_color_matrix(xyz, to_mats$xyz_to_rgb)
  if (d[3] == 4L) {
    out[,, 4] = src[,, 4]
  } # preserve alpha

  # Tag with the target space
  white_tag = if (adapt_white) dst_wp else src_wp
  to_mats$white_name = white_tag$white_name

  out = rayimg(
    out,
    filetype = attr(src, "filetype"),
    source_linear = TRUE,
    colorspace = to_mats,
    white_current = white_tag$value,
    exposure = attr(src, "exposure", exact = TRUE),
    iso = attr(src, "iso", exact = TRUE)
  )

  handle_image_output(out, filename = filename, preview = preview)
}

test_that("render_exposure bakes exposure and ISO into pixels", {
	img = ray_read_image(array(0.25, dim = c(2, 2, 4)))
	img[,, 4] = 0.5

	out = render_exposure(img, exposure = 1, iso = 200)

	expect_equal(unclass(out)[,, 1:3], array(1, dim = c(2, 2, 3)))
	expect_equal(unclass(out)[,, 4], matrix(0.5, 2, 2))
	expect_equal(attr(out, "exposure"), 1)
	expect_equal(attr(out, "iso"), 200)
})

test_that("plot and write do not apply informational exposure metadata", {
	img = ray_read_image(array(0.1, dim = c(2, 2, 3)))
	out = render_exposure(img, exposure = 1, iso = 200)

	tmp = tempfile(fileext = ".png")
	ray_write_image(out, tmp, write_linear = TRUE)
	written = png::readPNG(tmp)

	expect_equal(written[,, 1:3], array(0.4, dim = c(2, 2, 3)), tolerance = 1 / 255)
})

test_that("print.rayimg reports non-neutral exposure and ISO", {
	img = render_exposure(array(0.1, dim = c(2, 2, 3)), exposure = 1, iso = 200)

	printed = capture.output(print(img))

	expect_true(any(grepl("Exposure: \\+1 EV", printed)))
	expect_true(any(grepl("ISO: 200", printed)))
})

test_that("render_exposure auto uses luminance instead of max channel", {
	img = array(0, dim = c(2, 2, 3))
	img[,, 2] = 1
	img = ray_read_image(img)
	luma_weight = CS_ACESCG$rgb_to_xyz[2L, 2L]

	out = render_exposure(img, auto = TRUE, percentile = 1)

	expect_equal(attr(out, "exposure"), -log2(luma_weight))
	expect_equal(max(unclass(out)[,, 2]), 1 / luma_weight)
	expect_equal(max(unclass(out)[,, 2]) * luma_weight, 1)
})

test_that("render_exposure auto uses luminance percentile", {
	img = matrix(c(0.25, 0.25, 0.25, 100), nrow = 2)
	target_luminance = as.numeric(stats::quantile(as.numeric(img), 0.75, names = FALSE))

	out = render_exposure(img, auto = TRUE, percentile = 0.75)

	expect_equal(attr(out, "exposure"), -log2(target_luminance))
	expect_equal(
		as.numeric(stats::quantile(as.numeric(unclass(out)), 0.75, names = FALSE)),
		1
	)
	expect_gt(max(unclass(out)), 1)
})

test_that("render_exposure auto accounts for ISO and exposure compensation", {
	img = ray_read_image(array(0.25, dim = c(2, 2, 3)))

	out = render_exposure(img, iso = 200, auto = TRUE, percentile = 1)
	biased = render_exposure(img, exposure = 1, iso = 200, auto = TRUE, percentile = 1)

	expect_equal(attr(out, "exposure"), 1)
	expect_equal(max(unclass(out)), 1)
	expect_equal(attr(biased, "exposure"), 2)
	expect_equal(max(unclass(biased)), 2)
})

test_that("render_exposure auto verbose reports exposure info", {
	img = ray_read_image(array(0.25, dim = c(2, 2, 3)))

	expect_silent(render_exposure(img, auto = TRUE, percentile = 1))
	expect_message(
		render_exposure(img, auto = TRUE, percentile = 1, verbose = TRUE),
		"Auto exposure: \\+2.000 EV"
	)
})

test_that("ray_read_image can reset camera metadata", {
	img = render_exposure(array(0.25, dim = c(2, 2, 3)), exposure = 1, iso = 200)

	kept = ray_read_image(img)
	reset = ray_read_image(img, reset_camera_settings = TRUE)

	expect_equal(attr(kept, "exposure"), 1)
	expect_equal(attr(kept, "iso"), 200)
	expect_equal(attr(reset, "exposure"), 0)
	expect_equal(attr(reset, "iso"), 100)
	expect_equal(as.numeric(reset), as.numeric(img))
})

test_that("combining operations reset camera metadata", {
	img = render_exposure(array(0.25, dim = c(2, 2, 3)), exposure = 1, iso = 200)
	overlay = array(0, dim = c(2, 2, 4))
	overlay[,, 4] = 0.5

	out = render_image_overlay(img, overlay)

	expect_equal(attr(out, "exposure"), 0)
	expect_equal(attr(out, "iso"), 100)
})

test_that("colorspace and geometric operations preserve camera metadata", {
	img = render_exposure(array(0.25, dim = c(4, 4, 3)), exposure = 1, iso = 200)

	converted = render_convert_colorspace(img, to_mats = CS_SRGB)
	cropped = render_crop(img, width_limits = c(0, 0.5), height_limits = c(0, 0.5))

	expect_equal(attr(converted, "exposure"), 1)
	expect_equal(attr(converted, "iso"), 200)
	expect_equal(attr(cropped, "exposure"), 1)
	expect_equal(attr(cropped, "iso"), 200)
})

#'@title Read Image
#'
#'@description Reads an image from a file/array/matrix. From files, supports
#'`JPEG`, `PNG`, `TIFF`, `TGA`, `BMP`, `PSD`, `GIF`, `HDR`, `PIC`, `PNM`, `EXR`,
#'and `DNG` images.
#'
#' @param image 3-layer RGB/4-layer RGBA array, `rayimg` class, or filename of an image.
#' @param convert_to_array Default `FALSE`. Whether to convert 2D B&W images/matrices to RGBA arrays.
#' @param preview Default `FALSE`. If `TRUE`, it will display the image in addition
#'to returning it.
#' @param source_linear Default `NA`, automatically determined based on the image type.
#' Whether the image source is linear data or sRGB. `TRUE` for matrices, arrays, and
#' floating-point HDR formats (`EXR`, `HDR`), and `DNG`; `FALSE` for all other file-based formats
#' (e.g., `jpeg`, `png`, `tiff`, `tga`, `bmp`, `psd`, `gif`, `pic`, `pnm`).
#' @param normalize Default `FALSE`. If `TRUE`, convert to `normalize_to` space on read. Note that
#' rayimg inputs will keep their colorspace and ignore this option--use `render_convert_colorspace()`
#' to change an existing rayimg to a new colorspace.
#' @param dng_normalize Default `FALSE`. If `TRUE`, DNG data is normalized to 0..1,
#' using black/white levels; if `FALSE`, raw values are returned.
#' @param normalize_to Default `CS_ACESCG`. Target colorspace when `normalize=TRUE`.
#' @param normalize_adapt_white Default `TRUE`. If `TRUE`, Bradford-adapt source white to target white.
#' @param assume_colorspace Default `NULL`. A colorspace descriptor (e.g., `CS_SRGB`,
#'   `CS_ACESCG`). If given, the loaded image will be *tagged* with this space
#'   rather than the default inferred by file type. No pixel conversion is performed
#'   unless `normalize=TRUE`.
#' @param assume_white Default `NULL`. Scene/display white for the loaded image.
#'   Either a named white (`"D60","D65","D50","D55","D75","E"`) or XYZ with Y=1.
#'   If `NULL`, uses `assume_colorspace$white_xyz`.
#' @param reset_camera_settings Default `FALSE`. If `TRUE`, reset informational
#'   exposure/ISO metadata to neutral values (`exposure = 0`, `iso = 100`) when
#'   wrapping an existing `rayimg` or attributed array.
#'
#'@param ... Arguments to pass to `jpeg::readJPEG`, `png::readPNG`, `tiff::readTIFF`,
#'or `libopenexr::read_exr()` for supported formats.
#'
#'@return A `rayimg` RGBA array.
#'@import grDevices
#'@export
#'@examplesIf interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")
#'#Write as a png
#'tmparr = tempfile(fileext=".png")
#'ray_read_image(dragon) |>
#'  ray_write_image(tmparr)
#'ray_read_image(tmparr) |>
#'  plot_image()
#'#Write as a JPEG (passing quality arguments via ...)
#'tmparr = tempfile(fileext=".jpg")
#'ray_read_image(dragon) |>
#'  ray_write_image(tmparr, quality = 0.2)
#'ray_read_image(tmparr) |>
#'  plot_image()
#'#Write as a tiff
#'tmparr = tempfile(fileext=".tiff")
#'ray_read_image(dragon) |>
#'  ray_write_image(tmparr)
#'ray_read_image(tmparr) |>
#'   plot_image()
#'
#'#Write as an exr
#'tmparr = tempfile(fileext=".exr")
#'ray_read_image(dragon) |>
#'  ray_write_image(tmparr)
#'ray_read_image(tmparr) |>
#'   plot_image()
ray_read_image = function(
  image,
  convert_to_array = FALSE,
  preview = FALSE,
  source_linear = NA,
  normalize = FALSE,
  dng_normalize = FALSE,
  normalize_to = CS_ACESCG,
  normalize_adapt_white = TRUE,
  assume_colorspace = NULL,
  assume_white = NULL,
  reset_camera_settings = FALSE,
  ...
) {
  decode_if_needed = function(img, linear_flag) {
    if (isTRUE(linear_flag)) {
      return(list(image = img, source_linear = TRUE))
    }
    img[is.na(img)] = 0

    if (length(dim(img)) == 3) {
      if (dim(img)[3] >= 3) {
        img[,, 1:3] = to_linear(img[,, 1:3])
      } else {
        img = to_linear(img)
      }
    } else {
      img = to_linear(img)
    }
    list(image = img, source_linear = TRUE)
  }

  get_wp = function(white_point) {
    wp = get_whitepoint_xyz(white_point)
    if (is.null(wp)) {
      return(NULL)
    }
    wp$value
  }

  process_image_preview = function(image2) {
    if (convert_to_array) {
      image_val = array(1, dim = c(dim(image2)[1:2], 4))
      if (length(dim(image2)) == 2) {
        image_val[,, 1:3] = image2
      } else if (length(dim(image2)) == 3) {
        if (dim(image2)[3] == 2) {
          image_val[,, 1:3] = image2[,, 1]
          image_val[,, 4] = image2[,, 2]
        } else if (dim(image2)[3] == 3) {
          image_val[,, 1:3] = image2[,, 1:3]
        } else if (dim(image2)[3] == 4) {
          image_val = image2
        }
      }
    } else {
      image_val = image2
    }
    if (preview) {
      plot_image(image_val)
    }
    image_val
  }

  imagetype = if (
    is.character(image) &&
      tolower(tools::file_ext(image)) == "dng"
  ) {
    "dng"
  } else {
    get_file_type(image)
  }
  if (is.na(source_linear)) {
    # arrays, matrices, and floating-point HDR formats are treated as linear
    source_linear = imagetype %in% c("array", "matrix", "exr", "hdr", "dng")
  }

  finalize_image = function(img, img_type, linear_flag, default_cs) {
    decoded = decode_if_needed(img, linear_flag)
    cs = if (!is.null(assume_colorspace)) {
      assume_colorspace
    } else {
      switch(
        img_type,
        png = CS_SRGB,
        jpg = CS_SRGB,
        tif = CS_SRGB,
        tga = CS_SRGB,
        bmp = CS_SRGB,
        psd = CS_SRGB,
        gif = CS_SRGB,
        pic = CS_SRGB,
        pnm = CS_SRGB,
        hdr = CS_ACESCG,
        exr = CS_ACESCG,
        array = default_cs,
        matrix = default_cs,
        CS_ACESCG
      )
    }
    wc = get_wp(if (!is.null(assume_white)) assume_white else cs$white_xyz)
    ri = rayimg(
      process_image_preview(decoded$image),
      filetype = img_type,
      source_linear = decoded$source_linear,
      colorspace = cs,
      white_current = wc,
      exposure = if (reset_camera_settings) 0 else
        attr(img, "exposure", exact = TRUE),
      iso = if (reset_camera_settings) 100 else attr(img, "iso", exact = TRUE)
    )
    if (normalize) {
      ri = render_convert_colorspace(
        ri,
        from_mats = NA,
        to_mats = normalize_to,
        adapt_white = normalize_adapt_white,
        from_white = NA,
        to_white = normalize_to$white_xyz
      )
    }
    ri
  }

  if (inherits(image, "rayimg")) {
    cs_assigned = if (!is.null(assume_colorspace)) {
      assume_colorspace
    } else {
      attr(image, "colorspace")
    }
    wc_assigned = if (!is.null(assume_white)) {
      get_wp(assume_white)
    } else {
      attr(image, "white_current")
    }
    ri = rayimg(
      process_image_preview(image),
      filetype = attr(image, "filetype"),
      source_linear = attr(image, "source_linear"),
      colorspace = cs_assigned,
      white_current = wc_assigned,
      exposure = if (reset_camera_settings) 0 else
        attr(image, "exposure", exact = TRUE),
      iso = if (reset_camera_settings) 100 else attr(image, "iso", exact = TRUE)
    )
    return(ri) # no normalization on rayimg here
  }

  if (imagetype == "array") {
    return(finalize_image(image, imagetype, source_linear, CS_ACESCG))
  } else if (imagetype == "matrix") {
    if (length(dim(image)) == 3) {
      return(finalize_image(
        image[,, 1],
        imagetype,
        source_linear,
        CS_ACESCG
      ))
    }
    return(finalize_image(image, imagetype, source_linear, CS_ACESCG))
  } else if (imagetype == "dng") {
    out = read_dng_cpp(image, normalize = dng_normalize)
    pixels = out$pixels
    meta = out$meta
    if (
      !is.null(meta$orientation) &&
        !is.na(meta$orientation) &&
        meta$orientation != 1
    ) {
      orientation = as.integer(meta$orientation)
      flipx = FALSE
      flipy = FALSE
      transpose = FALSE
      if (orientation == 2L) {
        flipx = TRUE
      } else if (orientation == 3L) {
        flipx = TRUE
        flipy = TRUE
      } else if (orientation == 4L) {
        flipy = TRUE
      } else if (orientation == 5L) {
        transpose = TRUE
      } else if (orientation == 6L) {
        flipy = TRUE
        transpose = TRUE
      } else if (orientation == 7L) {
        flipx = TRUE
        flipy = TRUE
        transpose = TRUE
      } else if (orientation == 8L) {
        flipx = TRUE
        transpose = TRUE
      }
      if (flipx) {
        pixels = fliplr(pixels)
      }
      if (flipy) {
        pixels = flipud(pixels)
      }
      if (transpose) {
        if (length(dim(pixels)) == 2) {
          pixels = t(pixels)
        } else {
          pixels = aperm(pixels, c(2, 1, 3))
        }
        if (!is.null(meta$width) && !is.null(meta$height)) {
          tmp_dim = meta$width
          meta$width = meta$height
          meta$height = tmp_dim
        }
      }
      if (!is.null(meta$cfa_pattern)) {
        cfa = meta$cfa_pattern
        if (flipx) {
          cfa = cfa[, 2:1, drop = FALSE]
        }
        if (flipy) {
          cfa = cfa[2:1, , drop = FALSE]
        }
        if (transpose) {
          cfa = t(cfa)
        }
        meta$cfa_pattern = cfa
      }
      meta$orientation = 1L
    }
    attr(pixels, "dng") = meta
    if (!is.na(source_linear) && !source_linear) {
      pixels = decode_if_needed(pixels, source_linear)$image
    }
    return(finalize_image(pixels, "dng", TRUE, CS_ACESCG))
  } else if (imagetype == "png") {
    image = png::readPNG(image, ...)
    return(finalize_image(image, imagetype, source_linear, CS_SRGB))
  } else if (imagetype == "tif") {
    image = tiff::readTIFF(image, ...)
    return(finalize_image(image, imagetype, source_linear, CS_SRGB))
  } else if (imagetype == "jpg") {
    image = jpeg::readJPEG(image, ...)
    return(finalize_image(image, imagetype, source_linear, CS_SRGB))
  } else if (imagetype == "hdr") {
    image = read_image_stb(image, desired_channels = 4L)
    return(finalize_image(image, imagetype, source_linear, CS_ACESCG))
  } else if (imagetype %in% c("tga", "bmp", "psd", "gif", "pic", "pnm")) {
    image = read_image_stb(image, desired_channels = 4L)
    return(finalize_image(image, imagetype, source_linear, CS_SRGB))
  } else if (imagetype == "exr") {
    if (length(find.package("libopenexr", quiet = TRUE)) > 0) {
      tmp = libopenexr::read_exr(image, ...)
      image = array(1, dim = c(tmp$height, tmp$width, 4))
      image[,, 1] = tmp$r
      image[,, 2] = tmp$g
      image[,, 3] = tmp$b
      return(rayimg(
        process_image_preview(image),
        filetype = imagetype,
        source_linear = TRUE,
        colorspace = if (!is.null(assume_colorspace)) {
          assume_colorspace
        } else {
          CS_ACESCG
        },
        white_current = if (!is.null(assume_white)) {
          get_wp(assume_white)
        } else {
          CS_ACESCG$white_xyz
        }
      ))
    } else {
      stop("The 'libopenexr' package is required for EXR support.")
    }
  } else {
    stop("Unsupported/unknown input type.")
  }
}

test_that("Checking ray_read/ray_write", {
  rayfile_png = tempfile(fileext = ".png")
  clamped_dragon = render_clamp(dragon)
  rayimage::ray_write_image(filename = rayfile_png, image = clamped_dragon)
  dragon_png = rayimage::ray_read_image(image = rayfile_png)
  expect_true(compare_image(
    clamped_dragon,
    dragon_png,
    max_diff = 0.4,
    quantile_diff = 0.02
  ))

  rayfile_jpg = tempfile(fileext = ".jpg")
  rayimage::ray_write_image(
    filename = rayfile_jpg,
    image = clamped_dragon,
    quality = 1
  )
  dragon_jpg = rayimage::ray_read_image(
    image = rayfile_jpg,
    convert_to_array = TRUE
  )
  expect_true(compare_image(
    clamped_dragon,
    dragon_jpg,
    max_diff = 0.6,
    quantile_diff = 0.02
  ))

  rayfile_tiff = tempfile(fileext = ".tiff")
  expect_no_error(
    {
      rayimage::ray_write_image(
        filename = rayfile_tiff,
        image = rayimage::dragon
      )
    }
  )
  expect_warning(
    {
      dragon_tiff = rayimage::ray_read_image(image = rayfile_tiff)
    },
    regex = "Photometric"
  )
  expect_true(compare_image(
    clamped_dragon,
    dragon_tiff,
    quantile_diff = 0.02,
    max_diff = 0.5
  ))
})

test_that("common 8-bit formats via magick can be read and plotted", {
  skip_on_cran()
  skip_if_not_installed("magick")

  generate_and_test = function(ext, magick_format) {
    img = magick::image_blank(width = 16, height = 16, color = "red")
    tmp = tempfile(fileext = paste0(".", ext))

    write_res = try(
      magick::image_write(img, path = tmp, format = magick_format),
      silent = TRUE
    )
    if (inherits(write_res, "try-error")) {
      skip(paste("magick/ImageMagick not built with", magick_format, "support"))
    }

    ri = ray_read_image(tmp, convert_to_array = TRUE)
    expect_s3_class(ri, "rayimg")

    out = tempfile(fileext = ".png")
    grDevices::png(out, width = 200, height = 200, units = "px")
    expect_silent(plot_image(ri))
    grDevices::dev.off()
    expect_true(file.exists(out))
  }

  formats = list(
    list(ext = "jpg", magick_format = "jpeg"), # JPEG
    list(ext = "png", magick_format = "png"), # PNG
    list(ext = "tif", magick_format = "tiff"), # TIFF
    list(ext = "tga", magick_format = "tga"), # TGA
    list(ext = "bmp", magick_format = "bmp"), # BMP
    list(ext = "psd", magick_format = "psd"), # PSD
    list(ext = "gif", magick_format = "gif"), # GIF
    list(ext = "pnm", magick_format = "pnm") # PNM (PBM/PGM/PPM family)
  )

  for (f in formats) {
    generate_and_test(ext = f$ext, magick_format = f$magick_format)
  }
})

test_that("Radiance HDR files can be read and plotted", {
  skip_on_cran()
  skip_if_not_installed("magick")

  img = magick::image_blank(width = 16, height = 16, color = "red")
  tmp = tempfile(fileext = ".hdr")

  write_res = try(
    magick::image_write(img, path = tmp, format = "hdr"),
    silent = TRUE
  )
  if (inherits(write_res, "try-error")) {
    skip("magick/ImageMagick not built with Radiance HDR support")
  }

  ri = ray_read_image(tmp, convert_to_array = TRUE)
  expect_s3_class(ri, "rayimg")

  out = tempfile(fileext = ".png")
  grDevices::png(out, width = 200, height = 200, units = "px")
  expect_silent(plot_image(ri))
  grDevices::dev.off()
  expect_true(file.exists(out))
})

test_that("EXR files can be read and plotted", {
  skip_on_cran()
  skip_if_not_installed("libopenexr")

  w = 4L
  h = 4L
  r = matrix(seq(0, 1, length.out = w * h), nrow = h, ncol = w)
  g = matrix(0.5, nrow = h, ncol = w)
  b = matrix(1 - r, nrow = h, ncol = w)

  tmp = tempfile(fileext = ".exr")

  # libopenexr::write_exr(path, r, g, b, a)
  libopenexr::write_exr(
    tmp,
    r,
    g,
    b
  )

  ri = ray_read_image(tmp, convert_to_array = TRUE)
  expect_s3_class(ri, "rayimg")

  out = tempfile(fileext = ".png")
  grDevices::png(out, width = 200, height = 200, units = "px")
  expect_silent(plot_image(ri))
  grDevices::dev.off()
  expect_true(file.exists(out))
})

test_that("Softimage PIC files can be read and plotted (when supported)", {
  skip_on_cran()
  skip_if_not_installed("magick")

  img = magick::image_blank(width = 16, height = 16, color = "red")
  tmp = tempfile(fileext = ".pic")

  write_res = try(
    magick::image_write(img, path = tmp, format = "pic"),
    silent = TRUE
  )
  if (inherits(write_res, "try-error")) {
    skip("magick/ImageMagick not built with Softimage PIC support")
  }

  ri = ray_read_image(tmp, convert_to_array = TRUE)
  expect_s3_class(ri, "rayimg")

  out = tempfile(fileext = ".png")
  grDevices::png(out, width = 200, height = 200, units = "px")
  expect_silent(plot_image(ri))
  grDevices::dev.off()
  expect_true(file.exists(out))
})

test_that("DNG roundtrip mosaic preserves values", {
	h = 16L
	w = 16L
	mosaic = matrix(seq(0, 1, length.out = h * w), nrow = h, ncol = w)
	attr(mosaic, "dng") = list(
		cfa_pattern = matrix(c(0, 1, 1, 2), 2, 2, byrow = TRUE),
		black_level = c(0, 0, 0, 0),
		white_level = c(65535, 65535, 65535, 65535)
	)

	tmp = tempfile(fileext = ".dng")
	ray_write_image(mosaic, tmp)
	out = ray_read_image(tmp, dng_normalize = TRUE)

	diff = max(abs(as.matrix(out) - mosaic))
	expect_true(diff <= (1 / 65535 + 1e-6))
})

#'@title add_padding
#'
#'@description Adds padding to the matrix
#'
#'@param input A two-dimensional matrix or a 3D array (including `rayimg`).
#'@return Matrix with edges padded
#'@keywords internal
add_padding = function(input) {
  is_rayimg = inherits(input, "rayimg")
  input_array = if (is_rayimg) {
    unclass(input)
  } else {
    input
  }

  if (length(dim(input_array)) == 2) {
    temp = matrix(0, nrow = nrow(input_array) + 2, ncol = ncol(input_array) + 2)
    temp[2:(nrow(temp) - 1), 2:(ncol(temp) - 1)] = input_array
    temp[2:(nrow(temp) - 1), 1] = input_array[, 1]
    temp[1, 2:(ncol(temp) - 1)] = input_array[1, ]
    temp[2:(nrow(temp) - 1), ncol(temp)] = input_array[, ncol(input_array)]
    temp[nrow(temp), 2:(ncol(temp) - 1)] = input_array[nrow(input_array), ]
    temp[1, 1] = temp[1, 2]
    temp[1, ncol(temp)] = temp[1, ncol(temp) - 1]
    temp[nrow(temp), 1] = temp[nrow(temp) - 1, 2]
    temp[nrow(temp), ncol(temp)] = temp[nrow(temp) - 1, ncol(temp)]
  } else {
    d = dim(input_array)
    temp = array(0, dim = c(d[1] + 2, d[2] + 2, d[3]))
    temp[2:(d[1] + 1), 2:(d[2] + 1), ] = input_array
    temp[2:(d[1] + 1), 1, ] = input_array[, 1, ]
    temp[1, 2:(d[2] + 1), ] = input_array[1, , ]
    temp[2:(d[1] + 1), d[2] + 2, ] = input_array[, d[2], ]
    temp[d[1] + 2, 2:(d[2] + 1), ] = input_array[d[1], , ]
    temp[1, 1, ] = temp[1, 2, ]
    temp[1, d[2] + 2, ] = temp[1, d[2] + 1, ]
    temp[d[1] + 2, 1, ] = temp[d[1] + 1, 2, ]
    temp[d[1] + 2, d[2] + 2, ] = temp[d[1] + 1, d[2] + 2, ]
  }

  if (is_rayimg) {
    temp = rayimg(
      temp,
      filetype = attr(input, "filetype"),
      source_linear = attr(input, "source_linear"),
      colorspace = attr(input, "colorspace"),
      white_current = attr(input, "white_current"),
      exposure = attr(input, "exposure", exact = TRUE),
      iso = attr(input, "iso", exact = TRUE)
    )
  }
  temp
}

#'@title add_multi_padding
#'
#'@description Adds multiple levels padding to the matrix
#'
#'@param input A two-dimensional matrix or a 3D array (including `rayimg`).
#'@param pad Number of padding entries
#'@return Matrix with edges padded
#'@keywords internal
add_multi_padding = function(input, pad = 1) {
  if (length(dim(input)) == 2) {
    temp = matrix(0, nrow = nrow(input) + 2 * pad, ncol = ncol(input) + 2 * pad)
    temp[(1 + pad):(pad + nrow(input)), (1 + pad):(pad + ncol(input))] = input
    temp[(1 + pad):(pad + nrow(input)), 1:pad] = input[, 1]
    for (i in 1:pad) {
      temp[i, (1 + pad):(pad + ncol(input))] = input[1, ]
    }
    for (i in (pad + ncol(input) + 1):ncol(temp)) {
      temp[(1 + pad):(pad + nrow(input)), i] = input[, ncol(input)]
    }
    for (i in (pad + nrow(input) + 1):nrow(temp)) {
      temp[i, (1 + pad):(pad + ncol(input))] = input[nrow(input), ]
    }
    temp[1:pad, 1:pad] = input[1, 1]
    temp[1:pad, (pad + ncol(input) + 1):ncol(temp)] = input[1, ncol(input)]
    temp[(pad + nrow(input) + 1):nrow(temp), 1:pad] = input[nrow(input), 1]
    temp[
      (pad + nrow(input) + 1):nrow(temp),
      (pad + ncol(input) + 1):ncol(temp)
    ] = input[nrow(input), ncol(input)]
    return(temp)
  } else {
    temp_height = list()
    for (i in seq_len(dim(input)[3])) {
      temp_height[[i]] = add_multi_padding(input[,, i], pad)
    }
    temp_array = array(0, dim = c(dim(temp_height[[i]]), dim(input)[3]))
    for (i in seq_len(dim(input)[3])) {
      temp_array[,, i] = temp_height[[i]]
    }
    return(temp_array)
  }
}

#'@title trim_padding
#'
#'@description Trims padding
#'
#'@param input A two-dimensional matrix.
#'@param pad Number of padding entries
#'@return Matrix with edges trimmed
#'@keywords internal
trim_padding = function(input, pad = 1) {
  if (length(dim(input)) == 2) {
    return(input[(1 + pad):(nrow(input) - pad), (1 + pad):(ncol(input) - pad)])
  } else {
    return(input[
      (1 + pad):(nrow(input) - pad),
      (1 + pad):(ncol(input) - pad),
      ,
      drop = FALSE
    ])
  }
}

#'@title pad to fit
#'
#'@description Pads to a rect
#'
#'@return Hillshade with edges trimmed
#'@keywords internal
pad_to_fit = function(dim, kernel) {
  whichdim = which.min(dim)
  if (dim[1] != dim[2]) {
    if (whichdim == 1) {
      extra_rows = dim[2] - dim[1]
      if (extra_rows %% 2 == 0) {
        pad = matrix(0, nrow = extra_rows / 2, ncol = ncol(kernel))
        kernel = rbind(pad, kernel, pad)
      } else {
        pad1 = matrix(0, nrow = floor(extra_rows / 2) + 1, ncol = ncol(kernel))
        pad2 = matrix(0, nrow = floor(extra_rows / 2), ncol = ncol(kernel))
        kernel = rbind(pad1, kernel, pad2)
      }
    } else {
      extra_cols = dim[1] - dim[2]
      if (extra_cols %% 2 == 0) {
        pad = matrix(0, nrow = nrow(kernel), ncol = extra_cols / 2)
        kernel = cbind(pad, kernel, pad)
      } else {
        pad1 = matrix(0, nrow = nrow(kernel), ncol = floor(extra_cols / 2) + 1)
        pad2 = matrix(0, nrow = nrow(kernel), ncol = floor(extra_cols / 2))
        kernel = cbind(pad1, kernel, pad2)
      }
    }
  }
  return(kernel)
}

#'@title expand to fit
#'
#'@description Pads to a rect
#'
#'@return Hillshade with edges trimmed
#'@keywords internal
expand_to_fit = function(dim, kernel) {
  kdim = dim(kernel)
  if (!all(kdim <= dim)) {
    stop("kernel dimensions must be less than or equal to image")
  }
  if (kdim[1] != dim[1]) {
    extra_rows = dim[1] - kdim[1]
    if (extra_rows %% 2 == 0) {
      pad = matrix(0, nrow = extra_rows / 2, ncol = ncol(kernel))
      kernel = rbind(pad, kernel, pad)
    } else {
      pad1 = matrix(0, nrow = floor(extra_rows / 2) + 1, ncol = ncol(kernel))
      pad2 = matrix(0, nrow = floor(extra_rows / 2), ncol = ncol(kernel))
      kernel = rbind(pad1, kernel, pad2)
    }
  }
  if (kdim[2] != dim[2]) {
    extra_cols = dim[2] - kdim[2]
    if (extra_cols %% 2 == 0) {
      pad = matrix(0, nrow = nrow(kernel), ncol = extra_cols / 2)
      kernel = cbind(pad, kernel, pad)
    } else {
      pad1 = matrix(0, nrow = nrow(kernel), ncol = floor(extra_cols / 2) + 1)
      pad2 = matrix(0, nrow = nrow(kernel), ncol = floor(extra_cols / 2))
      kernel = cbind(pad1, kernel, pad2)
    }
  }
  return(kernel)
}

#' Print method for rayimg
#'
#' Displays the working colorspace and non-neutral camera metadata alongside a numeric preview.
#'
#' @param x Default `NULL`. A `rayimg` object.
#' @param preview_n Default `10`. Max rows/cols to display in numeric preview.
#' @param decimals Default `3`. Number of decimal places to display in numeric preview.
#' @param color Default `TRUE`. Colorize headers and R/G/B/A numeric cells when supported.
#' @param ... Default ``. Ignored.
#'
#' @return Invisibly returns `x`.
#' @export
print.rayimg = function(
  x = NULL,
  preview_n = 10,
  decimals = 3,
  color = TRUE,
  ...
) {
  ft = attr(x, "filetype")
  if (is.null(ft)) ft = "unknown"
  d = dim(x)
  image_type = attr(x, "filetype")
  exposure_attr = rayimg_exposure_value(attr(x, "exposure", exact = TRUE))
  iso_attr = rayimg_iso_value(attr(x, "iso", exact = TRUE))

  opt_override = getOption("rayimage.color", NA)
  want_color = isTRUE(color) && !isFALSE(opt_override)
  have_cli = want_color && requireNamespace("cli", quietly = TRUE)
  can_color = have_cli &&
    isTRUE(cli::num_ansi_colors() >= 8) &&
    isTRUE(cli::is_dynamic_tty()) &&
    !isTRUE(getOption("knitr.in.progress")) &&
    !isTRUE(getOption("testthat.in_progress"))
  truecolor_ok = rayimg_truecolor_supported()

  style_bold = function(z) if (can_color) cli::style_bold(z) else z
  col_magenta = function(z) if (can_color) cli::col_magenta(z) else z
  col_yellow = function(z) if (can_color) cli::col_yellow(z) else z
  col_cyan = function(z) if (can_color) cli::col_cyan(z) else z
  col_silver = function(z) if (can_color) cli::col_silver(z) else z
  col_red = function(z) if (can_color) cli::col_red(z) else z
  col_green = function(z) if (can_color) cli::col_green(z) else z
  col_blue = function(z) if (can_color) cli::col_blue(z) else z

  print_camera_settings = function() {
    if (isTRUE(all.equal(exposure_attr, 0)) && isTRUE(all.equal(iso_attr, 100))) {
      return(invisible(NULL))
    }
    cat(
      "  ",
      col_silver("Exposure: "),
      col_cyan(sprintf("%+.3g EV", exposure_attr)),
      "  ",
      col_silver("ISO: "),
      col_cyan(format(iso_attr, trim = TRUE, scientific = FALSE)),
      "\n",
      sep = ""
    )
    invisible(NULL)
  }

  hell = if (
    requireNamespace("cli", quietly = TRUE) &&
      isTRUE(cli::is_utf8_output())
  ) {
    cli::symbol$ellipsis
  } else {
    "..."
  }
  vell = rayimg_vertical_ellipsis(TRUE)

  # Vector fallback
  if (is.null(d)) {
    cat(
      col_magenta(style_bold("<rayimg>")),
      " ",
      col_silver(sprintf("[length %d]", length(x))),
      " ",
      col_yellow("<vector>"),
      "  ",
      "filetype: ",
      col_cyan(ft),
      "\n",
      sep = ""
    )
    print_camera_settings()
    y_plain = formatC(as.numeric(x), format = "f", digits = decimals)
    print(y_plain, quote = FALSE, right = TRUE)
    return(invisible(x))
  }

  h = d[1]
  w = d[2]
  c = if (length(d) >= 3) d[3] else 1L

  chan_labels = attr(x, "channels")

  if (!length(chan_labels)) {
    chan_labels = rayimg_detect_channels(x)
  }

  type_str = paste0(chan_labels, collapse = " + ")

  cat(
    col_magenta(style_bold("<rayimg>")),
    " ",
    col_silver(sprintf(
      "[%d x %d%s]",
      h,
      w,
      if (c > 1) paste0(" x ", c) else ""
    )),
    " ",
    col_yellow(paste0("<", type_str, ">")),
    "  ",
    "filetype: ",
    col_cyan(ft),
    "  Linear Data: ",
    col_green(as.character(attr(x, "source_linear"))),
    "\n",
    sep = ""
  )
  cs_attr = attr(x, "colorspace")
  white_attr = attr(x, "white_current")

  describe_colorspace = function(cs) {
    if (!is.list(cs)) return("unknown")
    nm = cs$name
    if (is.null(nm) || length(nm) != 1L) nm = "unknown"
    desc = as.character(nm)
    wn = cs$white_name
    if (!is.null(wn) && length(wn) == 1L) {
      desc = paste0(desc, " (white ", wn, ")")
    }
    desc
  }
  describe_white = function(white_vec) {
    if (
      is.numeric(white_vec) &&
        length(white_vec) == 3L &&
        all(is.finite(white_vec))
    ) {
      paste0(
        "[",
        paste(formatC(as.numeric(white_vec), format = "f", digits = 5), collapse = ", "),
        "]"
      )
    } else {
      "unknown"
    }
  }
  cs_desc = describe_colorspace(cs_attr)
  white_desc = describe_white(white_attr)
  white_col = if (identical(white_desc, "unknown")) col_silver else col_blue

  cat(
    "  ",
    col_silver("Working colorspace: "),
    col_blue(cs_desc),
    "  ",
    col_silver("White XYZ: "),
    white_col(white_desc),
    "\n",
    sep = ""
  )
  print_camera_settings()

  # Numeric preview
  rmax = min(h, preview_n)
  cmax = min(w, preview_n)
  more_rows = h > preview_n
  more_cols = w > preview_n
  if (more_rows || more_cols) {
    msg = sprintf(
      "showing [1:%d, 1:%d]%s plus last row/col",
      rmax,
      cmax,
      if (c > 1) ", all channels" else ""
    )
    cat(col_silver(msg), "\n", sep = "")
  }

  ux = unclass(x)

  # Alpha analysis based on displayed labels; get physical alpha plane correctly
  has_alpha = "Alpha" %in% chan_labels
  suppress_alpha = FALSE
  alpha_idx = -1
  if (has_alpha && length(dim(ux)) > 2) {
    alpha_idx = which("Alpha" == chan_labels)
    aplane = ux[,, alpha_idx, drop = FALSE]
    alpha_all_one = all(abs(aplane - 1) <= 1e-8, na.rm = TRUE)
    suppress_alpha = alpha_all_one
  }

  # Colorizers for numeric cells
  col_fun = function(lbl) {
    if (!can_color) return(identity)
    switch(
      lbl,
      "Red" = col_red,
      "Green" = col_green,
      "Blue" = col_blue,
      "Alpha" = col_cyan,
      identity
    )
  }
  hdr_col_fun = function(lbl) {
    if (!can_color) return(function(z) z)
    switch(
      lbl,
      "Red" = col_red,
      "Green" = col_green,
      "Blue" = col_blue,
      "Alpha" = col_cyan,
      function(z) z
    )
  }
  fmt = function(v) formatC(as.numeric(v), format = "f", digits = decimals)

  build_preview_pair = function(mat_numeric, f_color) {
    mat_numeric = unclass(mat_numeric)
    if (is.null(dim(mat_numeric))) {
      print_tl = mat_numeric
    } else {
      print_tl = mat_numeric[seq_len(rmax), seq_len(cmax), drop = TRUE]
    }
    tl_plain = matrix(
      fmt(print_tl),
      nrow = rmax,
      ncol = cmax
    )
    tl_disp = tl_plain

    if (more_cols) {
      last_col_plain = fmt(mat_numeric[seq_len(rmax), w, drop = TRUE])
      ell_col_plain = rep(hell, rmax)
      tl_plain = cbind(tl_plain, ell_col_plain, last_col_plain)
      tl_disp = cbind(
        tl_disp,
        if (can_color) col_silver(ell_col_plain) else ell_col_plain,
        last_col_plain
      )
    }

    if (more_rows) {
      last_row_first_plain = fmt(mat_numeric[h, seq_len(cmax), drop = TRUE])
      if (more_cols) {
        last_row_plain = c(
          last_row_first_plain,
          hell,
          fmt(mat_numeric[h, w, drop = TRUE])
        )
        last_row_disp = c(
          last_row_first_plain,
          if (can_color) col_silver(hell) else hell,
          fmt(mat_numeric[h, w, drop = TRUE])
        )
      } else {
        last_row_plain = last_row_first_plain
        last_row_disp = last_row_first_plain
      }
      ell_row_plain = rep(vell, ncol(tl_plain))
      ell_row_disp = if (can_color) col_silver(ell_row_plain) else ell_row_plain

      tl_plain = rbind(tl_plain, ell_row_plain, last_row_plain)
      tl_disp = rbind(tl_disp, ell_row_disp, last_row_disp)
    }

    cn_base = if (more_cols) {
      c(
        paste0("[,", seq_len(cmax), "]"),
        paste0("[,", hell, "]"),
        paste0("[,", w, "]")
      )
    } else {
      paste0("[,", seq_len(cmax), "]")
    }
    rn_base = if (more_rows) {
      c(
        paste0("[", seq_len(rmax), ",]"),
        paste0("[", vell, ",]"),
        paste0("[", h, ",]")
      )
    } else {
      paste0("[", seq_len(rmax), ",]")
    }

    dimnames(tl_plain) = list(rn_base, cn_base)
    dimnames(tl_disp) = dimnames(tl_plain)

    if (!identical(f_color, identity)) {
      mask_num = !(tl_disp %in% c(hell, vell))
      tl_disp[mask_num] = f_color(tl_disp[mask_num])
    }

    list(plain = tl_plain, disp = tl_disp)
  }

  print_matrix_ansi = function(mat_plain, mat_disp) {
    stopifnot(identical(dim(mat_plain), dim(mat_disp)))
    rn = rownames(mat_plain)
    cn = colnames(mat_plain)
    pad = function(s, w) {
      vis = if (can_color) {
        cli::ansi_nchar(s, type = "width")
      } else {
        nchar(s, type = "width")
      }
      if (vis < w) {
        return(paste0(strrep(" ", w - vis), s))
      } else {
        return(s)
      }
    }
    rn_w = max(nchar(rn, type = "width"))
    col_w = vapply(
      seq_len(ncol(mat_plain)),
      function(j) {
        max(nchar(c(cn[j], mat_plain[, j]), type = "width"))
      },
      integer(1)
    )

    cat(strrep(" ", rn_w + 1))
    for (j in seq_len(ncol(mat_plain))) {
      cat(" ", pad(cn[j], col_w[j]), sep = "")
    }
    cat("\n")
    for (i in seq_len(nrow(mat_plain))) {
      cat(pad(rn[i], rn_w), " ", sep = "")
      for (j in seq_len(ncol(mat_plain))) {
        cat(" ", pad(mat_disp[i, j], col_w[j]), sep = "")
      }
      cat("\n")
    }
  }

  # ---- 2-D (single-channel) path ----
  # This colors slices the appropriate color (e.g. red if it's the red slice)
  if (length(d) == 2L || c == 1L) {
    chan_hint = if (length(chan_labels)) chan_labels[1] else "Grayscale"
    cat(
      hdr_col_fun(chan_hint)(style_bold(paste0("[", chan_hint, "]"))),
      "\n",
      sep = ""
    )
    pair = build_preview_pair(ux, f_color = col_fun(chan_hint))
    print_matrix_ansi(pair$plain, pair$disp)
    return(invisible(x))
  }

  for (i in seq_along(chan_labels)) {
    if (has_alpha && suppress_alpha && i == alpha_idx) next
    lbl = chan_labels[i]
    cat(hdr_col_fun(lbl)(style_bold(paste0("[", lbl, "]"))), "\n", sep = "")
    pair = build_preview_pair(
      ux[,, i, drop = TRUE],
      f_color = col_fun(lbl)
    )
    print_matrix_ansi(pair$plain, pair$disp)

    if (
      i < length(chan_labels) &&
        !(suppress_alpha && (i + 1L) == alpha_idx)
    ) {
      cat("\n")
    }
  }
  if (
    suppress_alpha &&
      chan_labels[length(chan_labels)] == "Alpha"
  ) {
    cat(col_silver(style_bold("[A] No transparency (layer all 1.0)\n")))
  }

  invisible(x)
}

#'@title Plot Image
#'
#'@description Displays the image in the current device.
#'
#'@param image 3-layer RGB/4-layer RGBA array, `rayimg` class, or filename of an image.
#'@param draw_grid Default `FALSE`. If `TRUE`, this will draw a grid in the background to help
#'disambiguate the actual image from the device (helpful if the image background is the same as the
#'device's background).
#'@param asp Default `1`. Aspect ratio of the pixels in the plot. For example, an aspect ratio of `4/3` will
#'slightly widen the image.
#'@param new_page  Default `TRUE`. Whether to call `grid::grid.newpage()` before plotting the image.
#'@param ignore_alpha Default `FALSE`. Whether to ignore the alpha channel when plotting.
#'@param return_grob Default `FALSE`. Whether to return the grob object.
#'@param gp A `grid::gpar()` object to include for the grid viewport displaying the image.
#'@param angle Default `0`. Counter-clockwise rotation (in degrees) applied before plotting.
#'@param show_linear Default `FALSE`. Most data should be gamma corrected before displaying on a screen.
#' Set to `TRUE` to turn off this gamma correction.
#'@export
#'@examplesIf interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")
#'#Plot the dragon array
#'plot_image(dragon)
#'#Make pixels twice as wide as tall
#'plot_image(dragon, asp = 2)
#'#Plot non-square images
#'plot_image(dragon[1:100,,])
#'#Make pixels twice as tall as wide
#'plot_image(dragon[1:100,,], asp = 1/2)
plot_image = function(
  image,
  draw_grid = FALSE,
  ignore_alpha = FALSE,
  asp = 1,
  new_page = TRUE,
  return_grob = FALSE,
  gp = grid::gpar(),
  angle = 0,
  show_linear = FALSE
) {
  prepared = prepare_native_raster(
    image,
    angle = angle,
    show_linear = show_linear
  )
  nr = prepared$native_raster

  if (new_page) {
    grid::grid.newpage()
  }
  image_dim = prepared$display_dim

  if (draw_grid) {
    draw_grid_fxn = function() {
      grid::pushViewport(
        grid::viewport(
          layout = grid::grid.layout(
            1,
            1,
            widths = grid::unit(image_dim[1], "pt"),
            heights = grid::unit(image_dim[2], "pt")
          ),
          gp = gp
        )
      )
      grid_density = 0.01
      for (i in seq(-2, 2, by = grid_density)) {
        grid::grid.lines(
          x = c(0, 1),
          y = c(i, i + 1),
          default.units = "npc",
          gp = grid::gpar(col = "grey")
        )
        grid::grid.lines(
          x = c(0, 1),
          y = c(i, i - 1),
          default.units = "npc",
          gp = grid::gpar(col = "grey")
        )
      }
      grid::popViewport()
    }
    draw_grid_fxn()
  }
  return(plot_asp_native_raster(
    nr,
    asp = asp,
    return_grob = return_grob,
    gp = gp
  ))
}

#' @title Render Alpha Outline
#'
#' @description Creates a colored RGBA outline or halo from an image alpha
#' channel or a supplied mask. When `image` is supplied, the original image is
#' composited over the outline by default.
#'
#' @param image Default `NULL`. Image whose alpha channel will be used as the
#' mask. If `mask` is supplied, `image` is only used for metadata.
#' @param mask Default `NULL`. Optional logical or numeric matrix to use as the
#' outline mask instead of `image` alpha.
#' @param expand Default `0`. Number of pixels to expand the mask.
#' @param edge_softness Default `0.1`. Width of the softened halo edge
#' transition, in pixels.
#' @param blur Default `0`. Gaussian blur standard deviation applied to the
#' final outline alpha.
#' @param gap_fill Default `1`. Maximum alpha gap width, in pixels, to bridge
#' in the expanded outline alpha. This fills border-connected breaks and notches
#' caused by discrete distance quantization. Set to `0` to disable.
#' @param gap_fill_alpha_threshold Default `0.25`. Alpha threshold used to
#' classify border gaps. Gap filling only modifies pixels below this threshold
#' that are connected to the image exterior; enclosed interior holes surrounded
#' by pixels at or above this threshold are left unchanged.
#' @param color Default `"black"`. Outline color.
#' @param alpha Default `1`. Overall outline alpha multiplier.
#' @param pad Default `0`. Padding in pixels before computing the outline. If
#' `NULL`, padding is computed from `expand`, `edge_softness`, and `blur`.
#' @param composite Default `TRUE`. If `TRUE` and `image` is supplied, composite
#' the outline below the original image. Set to `FALSE` to return the outline
#' alone.
#' @param filename Default `NULL`. File to save the image to. If `NULL` and
#' `preview = FALSE`, returns the output image.
#' @param preview Default `FALSE`. If `TRUE`, display the outline image.
#'
#' @return A `rayimg` RGBA output image with a `padding` attribute.
#' @export
#' @examplesIf interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")
#'txt = render_text_image("\U0001F409",  size = 120, background_alpha = 0)
#'render_alpha_outline(txt, expand = 2, blur = 1, color = "purple") |>
#'  plot_image()
#' 
#' #Make a rainbow outline 
#'render_alpha_outline(txt, expand = 2, blur = 0.1, color = "darkgreen") |>
#'  render_alpha_outline(expand = 2, blur = 0.1, color = "yellow") |>
#'  render_alpha_outline(expand = 2, blur = 0.1, color = "orange") |>
#'  render_alpha_outline(expand = 2, blur = 0.1, color = "red") |>
#'  render_alpha_outline(expand = 2, blur = 0.1, color = "purple") |>
#'  render_alpha_outline(expand = 2, blur = 0.1, color = "blue") |>
#'  plot_image()
render_alpha_outline = function(
  image = NULL,
  mask = NULL,
  expand = 0,
  edge_softness = 0.1,
  blur = 0,
  gap_fill = 1,
  gap_fill_alpha_threshold = 0.25,
  color = "black",
  alpha = 1,
  pad = 0,
  composite = TRUE,
  filename = NULL,
  preview = FALSE
) {
  image_supplied = !is.null(image)
  if (!image_supplied && is.null(mask)) {
    stop("Either `image` or `mask` must be provided.")
  }
  if (!is.numeric(expand) || length(expand) != 1 || !is.finite(expand)) {
    stop("`expand` must be a finite numeric scalar.")
  }
  if (
    !is.numeric(edge_softness) ||
      length(edge_softness) != 1 ||
      !is.finite(edge_softness)
  ) {
    stop("`edge_softness` must be a finite numeric scalar.")
  }
  if (!is.numeric(blur) || length(blur) != 1 || !is.finite(blur)) {
    stop("`blur` must be a finite numeric scalar.")
  }
  if (!is.numeric(gap_fill) || length(gap_fill) != 1 || !is.finite(gap_fill)) {
    stop("`gap_fill` must be a finite numeric scalar.")
  }
  if (
    !is.numeric(gap_fill_alpha_threshold) ||
      length(gap_fill_alpha_threshold) != 1 ||
      !is.finite(gap_fill_alpha_threshold)
  ) {
    stop("`gap_fill_alpha_threshold` must be a finite numeric scalar.")
  }
  if (!is.numeric(alpha) || length(alpha) != 1 || !is.finite(alpha)) {
    stop("`alpha` must be a finite numeric scalar.")
  }
  if (!is.logical(composite) || length(composite) != 1 || is.na(composite)) {
    stop("`composite` must be `TRUE` or `FALSE`.")
  }

  expand = max(0, expand)
  edge_softness = max(.Machine$double.eps, edge_softness)
  blur = max(0, blur)
  gap_fill = as.integer(round(max(0, gap_fill)))
  if (gap_fill_alpha_threshold < 0 || gap_fill_alpha_threshold > 1) {
    stop("`gap_fill_alpha_threshold` must be in [0,1].")
  }
  alpha = min(max(alpha, 0), 1)
  if (is.null(pad)) {
    pad = ceiling(expand + edge_softness + 3 * blur)
  }
  pad = normalize_render_padding(pad)

  if (!is.null(mask) && (is.null(dim(mask)) || length(dim(mask)) != 2)) {
    stop("`mask` must be a matrix.")
  }
  if (!image_supplied) {
    image = array(0, dim = c(nrow(mask), ncol(mask), 4))
  }
  source_image = ray_read_image(image, convert_to_array = TRUE)
  padded_source = render_padding(
    source_image,
    pad = pad,
    color = "black",
    alpha = 0
  )

  if (is.null(mask)) {
    alpha_mask = unclass(padded_source[,, 4])
  } else {
    alpha_mask = render_padding(mask, pad = pad, color = "black")
    alpha_mask = unclass(alpha_mask)
  }
  alpha_mask[alpha_mask > 0] = 1
  alpha_mask[alpha_mask <= 0] = 0

  if (expand > 0) {
    booldistance = render_boolean_distance(alpha_mask)
    inner = expand - edge_softness
    outer = expand + edge_softness
    halo_alpha = matrix(
      0,
      nrow = nrow(alpha_mask),
      ncol = ncol(alpha_mask)
    )
    halo_alpha[booldistance <= inner] = 1
    band = booldistance > inner & booldistance < outer
    halo_alpha[band] = (outer - booldistance[band]) / (outer - inner)
  } else {
    halo_alpha = alpha_mask
  }
  if (gap_fill > 0 && expand > 0) {
    halo_alpha = fill_alpha_gaps(
      halo_alpha,
      max_gap = gap_fill,
      alpha_threshold = gap_fill_alpha_threshold
    )
  }

  halo_color = convert_color(color, single = TRUE)
  halo_image = array(0, dim = c(dim(alpha_mask), 4))
  halo_image[,, 1] = halo_color[1]
  halo_image[,, 2] = halo_color[2]
  halo_image[,, 3] = halo_color[3]
  halo_image[,, 4] = halo_alpha * alpha

  halo_image = rayimg(
    halo_image,
    filetype = attr(padded_source, "filetype"),
    source_linear = attr(padded_source, "source_linear"),
    colorspace = attr(padded_source, "colorspace"),
    white_current = attr(padded_source, "white_current"),
    exposure = attr(padded_source, "exposure", exact = TRUE),
    iso = attr(padded_source, "iso", exact = TRUE)
  )

  if (blur > 0) {
    halo_image = render_convolution(
      halo_image,
      kernel = generate_2d_gaussian(
        sd = blur,
        dim = 31,
        width = 30
      ),
      include_alpha = TRUE,
      preview = FALSE
    )
    halo_image[,, 1] = halo_color[1]
    halo_image[,, 2] = halo_color[2]
    halo_image[,, 3] = halo_color[3]
  }
  attr(halo_image, "padding") = pad
  if (composite && image_supplied) {
    if (!all(dim(padded_source)[1:2] == dim(halo_image)[1:2])) {
      stop(
        "`image` and `mask` must have matching dimensions when ",
        "`composite = TRUE`."
      )
    }
    halo_image = composite_alpha_outline_image(
      halo_image = halo_image,
      source_image = padded_source
    )
    attr(halo_image, "padding") = pad
  }
  handle_image_output(halo_image, filename = filename, preview = preview)
}

#' @title Composite Alpha Outline Image
#'
#' @description Internal helper that composites the original source image over
#' the generated outline using source-over alpha compositing. The outline is the
#' backdrop and the source image is the foreground. Colors are treated as
#' straight, non-premultiplied alpha, using
#' `Ao = Af + Ab * (1 - Af)` and
#' `Co = (Cf * Af + Cb * Ab * (1 - Af)) / Ao`. Fully transparent output pixels
#' are forced to black to avoid carrying arbitrary color values.
#' @param halo_image RGBA outline image used as the compositing backdrop.
#' @param source_image RGBA source image, padded to the same dimensions as
#' `halo_image`, used as the compositing foreground.
#' @return A `rayimg` RGBA image containing the source composited over the
#' outline.
#'
#' @keywords internal
composite_alpha_outline_image = function(halo_image, source_image) {
  halo_array = unclass(halo_image)
  source_array = unclass(source_image)

  Cf = source_array[,, 1:3]
  Af = source_array[,, 4]
  Cb = halo_array[,, 1:3]
  Ab = halo_array[,, 4]

  Af3 = array(Af, dim = c(dim(Af), 3))
  Ab3 = array(Ab, dim = c(dim(Ab), 3))
  Ao = pmin(pmax(Af + Ab * (1 - Af), 0), 1)
  num = Cf * Af3 + Cb * Ab3 * array(1 - Af, dim = c(dim(Af), 3))

  eps = 1e-8
  Co = num / array(pmax(Ao, eps), dim = c(dim(Ao), 3))
  if (any(Ao <= eps)) {
    transparent = Ao <= eps
    Co[,, 1][transparent] = 0
    Co[,, 2][transparent] = 0
    Co[,, 3][transparent] = 0
  }

  output = array(0, dim = dim(halo_array))
  output[,, 1:3] = Co
  output[,, 4] = Ao
  rayimg(
    output,
    filetype = attr(halo_image, "filetype"),
    source_linear = attr(halo_image, "source_linear"),
    colorspace = attr(halo_image, "colorspace"),
    white_current = attr(halo_image, "white_current"),
    exposure = attr(halo_image, "exposure", exact = TRUE),
    iso = attr(halo_image, "iso", exact = TRUE)
  )
}

#' @title Fill Border Alpha Gaps
#'
#' @description Internal post-process for small distance-transform
#' quantization artifacts on outline edges. The algorithm first calls
#' `find_border_alpha_gaps()` to mark only low-alpha pixels connected to the
#' image exterior. It then scans horizontal, vertical, and diagonal directions.
#' For each gap width up to `max_gap`, `offset_alpha()` samples the alpha values
#' on opposite sides of each candidate gap. A candidate pixel is raised to the
#' weaker of the two neighboring alpha values, and only if that pixel was marked
#' as border-connected. This bridges edge notches while leaving enclosed
#' interior holes unchanged.
#' @param alpha Numeric alpha matrix.
#' @param max_gap Default `1`. Maximum low-alpha run width, in pixels, to bridge.
#' @param alpha_threshold Default `0.25`. Alpha threshold used to decide which
#' pixels are low-alpha border gaps.
#' @return Numeric alpha matrix with eligible border gaps filled.
#'
#' @keywords internal
fill_alpha_gaps = function(alpha, max_gap = 1, alpha_threshold = 0.25) {
  max_gap = as.integer(round(max_gap))
  if (max_gap <= 0) {
    return(alpha)
  }
  alpha = pmin(pmax(alpha, 0), 1)
  border_gap = find_border_alpha_gaps(alpha, alpha_threshold)
  if (!any(border_gap)) {
    return(alpha)
  }
  directions = list(
    c(0L, 1L),
    c(1L, 0L),
    c(1L, 1L),
    c(1L, -1L)
  )
  for (direction in directions) {
    for (gap_width in seq_len(max_gap)) {
      for (left_distance in seq_len(gap_width)) {
        right_distance = gap_width + 1L - left_distance
        left_alpha = offset_alpha(
          alpha,
          row_offset = -left_distance * direction[1],
          col_offset = -left_distance * direction[2]
        )
        right_alpha = offset_alpha(
          alpha,
          row_offset = right_distance * direction[1],
          col_offset = right_distance * direction[2]
        )
        candidate_alpha = pmin(left_alpha, right_alpha)
        alpha[border_gap] = pmax(alpha[border_gap], candidate_alpha[border_gap])
      }
    }
  }
  pmin(pmax(alpha, 0), 1)
}

#' @title Find Border Alpha Gaps
#'
#' @description Internal helper that classifies which low-alpha pixels are part
#' of the exterior border region. Pixels below `alpha_threshold` are considered
#' low alpha. The algorithm starts from low-alpha pixels on the image boundary
#' and performs an 8-connected flood fill through other low-alpha pixels.
#' Low-alpha regions enclosed by pixels at or above `alpha_threshold` are not
#' reached, so they are protected from `fill_alpha_gaps()`.
#' @param alpha Numeric alpha matrix.
#' @param alpha_threshold Default `0.25`. Pixels below this alpha are considered
#' low-alpha gap pixels.
#' @return Logical matrix marking low-alpha pixels connected to the image
#' exterior.
#'
#' @keywords internal
find_border_alpha_gaps = function(alpha, alpha_threshold = 0.25) {
  alpha_threshold = min(max(alpha_threshold, 0), 1)
  low_alpha = alpha < alpha_threshold
  border_gap = matrix(FALSE, nrow = nrow(alpha), ncol = ncol(alpha))
  if (!any(low_alpha)) {
    return(border_gap)
  }

  border = matrix(FALSE, nrow = nrow(alpha), ncol = ncol(alpha))
  border[1, ] = TRUE
  border[nrow(alpha), ] = TRUE
  border[, 1] = TRUE
  border[, ncol(alpha)] = TRUE
  border_indices = which(low_alpha & border)
  if (length(border_indices) == 0) {
    return(border_gap)
  }

  nr = nrow(alpha)
  nc = ncol(alpha)
  queue = integer(length(alpha))
  queue[seq_along(border_indices)] = border_indices
  queue_start = 1L
  queue_end = length(border_indices)
  border_gap[border_indices] = TRUE

  while (queue_start <= queue_end) {
    idx = queue[queue_start]
    queue_start = queue_start + 1L
    row = ((idx - 1L) %% nr) + 1L
    col = ((idx - 1L) %/% nr) + 1L

    row_min = max(1L, row - 1L)
    row_max = min(nr, row + 1L)
    col_min = max(1L, col - 1L)
    col_max = min(nc, col + 1L)
    for (neighbor_row in row_min:row_max) {
      for (neighbor_col in col_min:col_max) {
        if (neighbor_row == row && neighbor_col == col) {
          next
        }
        neighbor_idx = neighbor_row + (neighbor_col - 1L) * nr
        if (low_alpha[neighbor_idx] && !border_gap[neighbor_idx]) {
          queue_end = queue_end + 1L
          queue[queue_end] = neighbor_idx
          border_gap[neighbor_idx] = TRUE
        }
      }
    }
  }
  border_gap
}

#' @title Offset Alpha Matrix
#'
#' @description Internal shift helper used by `fill_alpha_gaps()`. It returns a
#' same-size alpha matrix shifted by integer row and column offsets, filling
#' newly exposed pixels with zero instead of wrapping. This lets the gap-filling
#' scan compare pixels a fixed distance away on opposite sides of a candidate
#' gap.
#' @param alpha Numeric alpha matrix.
#' @param row_offset Integer row offset. Positive values sample from rows below
#' each output pixel.
#' @param col_offset Integer column offset. Positive values sample from columns
#' to the right of each output pixel.
#' @return Numeric alpha matrix with the same dimensions as `alpha`.
#'
#' @keywords internal
offset_alpha = function(alpha, row_offset, col_offset) {
  nr = nrow(alpha)
  nc = ncol(alpha)
  output = matrix(0, nrow = nr, ncol = nc)

  row_start = max(1L, 1L - row_offset)
  row_end = min(nr, nr - row_offset)
  col_start = max(1L, 1L - col_offset)
  col_end = min(nc, nc - col_offset)
  if (row_start > row_end || col_start > col_end) {
    return(output)
  }
  dest_rows = row_start:row_end
  dest_cols = col_start:col_end

  output[dest_rows, dest_cols] = alpha[
    dest_rows + row_offset,
    dest_cols + col_offset,
    drop = FALSE
  ]
  output
}

test_that("render_sprite_overlay preserves RGB output channels", {
  image = array(0, dim = c(4, 4, 3))
  overlay = array(0, dim = c(2, 2, 4))
  overlay[,, 1] = 1
  overlay[,, 4] = 0.5

  output = render_sprite_overlay(
    image,
    overlay,
    convert_overlay_colorspace = FALSE,
    overlay_coords = c(2, 2),
    preserve_channels = TRUE,
    preview = FALSE
  )

  expect_equal(dim(output), c(4, 4, 3))
  expect_equal(unclass(output)[2:3, 2:3, 1], matrix(0.5, 2, 2))
})

test_that("render_sprite_overlay preserves RGBA output alpha", {
  image = array(0, dim = c(3, 3, 4))
  image[,, 4] = 0.5
  overlay = array(0, dim = c(3, 3, 4))
  overlay[,, 1] = 1
  overlay[,, 4] = 0.5

  output = render_sprite_overlay(
    image,
    overlay,
    convert_overlay_colorspace = FALSE,
    preserve_channels = TRUE,
    preview = FALSE
  )

  expect_equal(dim(output), c(3, 3, 4))
  expect_equal(unclass(output)[,, 4], matrix(0.75, 3, 3))
})

test_that("render_sprite_overlay places overlays with numeric justification", {
  image = array(0, dim = c(6, 8, 3))
  overlay = array(0, dim = c(2, 3, 4))
  overlay[,, 1] = 1
  overlay[,, 4] = 1

  expect_overlay_region = function(hjust, expected_cols) {
    output = render_sprite_overlay(
      image,
      overlay,
      convert_overlay_colorspace = FALSE,
      overlay_coords = c(6.2, 3.1),
      hjust = hjust,
      vjust = 0,
      preserve_channels = TRUE,
      preview = FALSE
    )
    expected = matrix(FALSE, nrow = 6, ncol = 8)
    expected[3:4, expected_cols] = TRUE
    expect_identical(unclass(output)[,, 1] > 0.9, expected)
  }

  expect_overlay_region(0, 6:8)
  expect_overlay_region(0.5, 5:7)
  expect_overlay_region(1, 3:5)

  output = render_sprite_overlay(
    image,
    overlay,
    convert_overlay_colorspace = FALSE,
    overlay_coords = c(6.2, 3.1),
    overlay_just = c(1, 0),
    preserve_channels = TRUE,
    preview = FALSE
  )
  expected = matrix(FALSE, nrow = 6, ncol = 8)
  expected[3:4, 3:5] = TRUE
  expect_identical(unclass(output)[,, 1] > 0.9, expected)
})

test_that("render_sprite_overlay crops overlays outside each image edge", {
  image = array(0, dim = c(4, 4, 3))
  overlay = array(0, dim = c(3, 3, 4))
  overlay[,, 1] = 1
  overlay[,, 4] = 1

  expect_cropped_region = function(coords, rows, cols) {
    output = render_sprite_overlay(
      image,
      overlay,
      convert_overlay_colorspace = FALSE,
      overlay_coords = coords,
      hjust = 0,
      vjust = 0,
      preserve_channels = TRUE,
      preview = FALSE
    )
    expected = matrix(FALSE, nrow = 4, ncol = 4)
    expected[rows, cols] = TRUE
    expect_identical(unclass(output)[,, 1] > 0.9, expected)
  }

  expect_cropped_region(c(0, 2), 2:4, 1:2)
  expect_cropped_region(c(3, 2), 2:4, 3:4)
  expect_cropped_region(c(2, 0), 1:2, 2:4)
  expect_cropped_region(c(2, 3), 3:4, 2:4)
})

test_that("render_padding adds transparent padding without edge replication", {
  image = array(0, dim = c(2, 3, 4))
  image[,, 1] = matrix(1:6, nrow = 2) / 6
  image[,, 2] = 0.25
  image[,, 3] = 0.5
  image[,, 4] = 1

  padded = render_padding(image, pad = 1, preview = FALSE)
  padded_array = unclass(padded)

  expect_equal(dim(padded), c(4, 5, 4))
  expect_equal(padded_array[2:3, 2:4, ], image)
  expect_true(all(padded_array[c(1, 4), , 4] == 0))
  expect_true(all(padded_array[, c(1, 5), 4] == 0))
  expect_true(all(padded_array[1, , 1:3] == 0))
  expect_true(all(padded_array[, 1, 1:3] == 0))

  asymmetric = render_padding(
    image,
    pad = c(1, 2, 3, 4),
    preview = FALSE
  )
  expect_equal(dim(asymmetric), c(6, 9, 4))
  expect_equal(unclass(asymmetric)[2:3, 5:7, ], image)

  opaque_color = render_padding(
    image,
    pad = 1,
    color = "red",
    preview = FALSE
  )
  opaque_array = unclass(opaque_color)
  expect_true(all(opaque_array[c(1, 4), , 1] == 1))
  expect_true(all(opaque_array[c(1, 4), , 2:3] == 0))
  expect_true(all(opaque_array[c(1, 4), , 4] == 1))

  transparent_color = render_padding(
    image,
    pad = 1,
    color = "red",
    alpha = 0.25,
    preview = FALSE
  )
  transparent_array = unclass(transparent_color)
  expect_true(all(transparent_array[c(1, 4), , 1] == 1))
  expect_true(all(transparent_array[c(1, 4), , 4] == 0.25))

  scalar_color = render_padding(
    image,
    pad = 1,
    color = 0.25,
    alpha = 0.5,
    preview = FALSE
  )
  scalar_array = unclass(scalar_color)
  expect_true(all(scalar_array[c(1, 4), , 1:3] == 0.25))
  expect_true(all(scalar_array[c(1, 4), , 4] == 0.5))

  rgba_color = render_padding(
    image,
    pad = 1,
    color = c(1, 0, 0, 0.25),
    preview = FALSE
  )
  rgba_array = unclass(rgba_color)
  expect_true(all(rgba_array[c(1, 4), , 1] == 1))
  expect_true(all(rgba_array[c(1, 4), , 2:3] == 0))
  expect_true(all(rgba_array[c(1, 4), , 4] == 0.25))

  rgba_override = render_padding(
    image,
    pad = 1,
    color = c(1, 0, 0, 0.25),
    alpha = 0.5,
    preview = FALSE
  )
  rgba_override_array = unclass(rgba_override)
  expect_true(all(rgba_override_array[c(1, 4), , 1] == 1))
  expect_true(all(rgba_override_array[c(1, 4), , 2:3] == 0))
  expect_true(all(rgba_override_array[c(1, 4), , 4] == 0.5))

  hex_alpha_color = render_padding(
    image,
    pad = 1,
    color = "#FF000040",
    preview = FALSE
  )
  hex_alpha_array = unclass(hex_alpha_color)
  expect_true(all(hex_alpha_array[c(1, 4), , 1] == 1))
  expect_true(all(hex_alpha_array[c(1, 4), , 2:3] == 0))
  expect_equal(hex_alpha_array[c(1, 4), , 4], array(64 / 255, dim = c(2, 5)))

  explicit_alpha = render_padding(
    image,
    pad = 1,
    color = "#FF000040",
    alpha = 0.5,
    preview = FALSE
  )
  explicit_alpha_array = unclass(explicit_alpha)
  expect_true(all(explicit_alpha_array[c(1, 4), , 1] == 1))
  expect_true(all(explicit_alpha_array[c(1, 4), , 2:3] == 0))
  expect_true(all(explicit_alpha_array[c(1, 4), , 4] == 0.5))

  expect_equal(
    convert_color(c(1, 0, 0, 0.25), as_hex = TRUE),
    "#FF000040"
  )

  expect_error(
    render_padding(image, pad = 1, color = c(1, 0), preview = FALSE),
    "invalid color"
  )
})

test_that("render_stack uses shared color parsing for backgrounds", {
  image = array(1, dim = c(1, 1, 4))
  image[,, 4] = 1
  wider = array(1, dim = c(1, 3, 4))
  wider[,, 4] = 1

  stacked = render_stack(
    list(image, wider),
    stack = "v",
    align = "left",
    background = c(0.1, 0.2, 0.3, 0.4),
    convert_colorspace = FALSE,
    preview = FALSE
  )
  stacked_array = unclass(stacked)
  expect_equal(stacked_array[1, 2:3, 1], c(0.1, 0.1))
  expect_equal(stacked_array[1, 2:3, 2], c(0.2, 0.2))
  expect_equal(stacked_array[1, 2:3, 3], c(0.3, 0.3))
  expect_equal(stacked_array[1, 2:3, 4], c(0.4, 0.4))

  transparent = render_stack(
    list(image, wider),
    stack = "v",
    align = "left",
    background = "transparent",
    convert_colorspace = FALSE,
    preview = FALSE
  )
  transparent_array = unclass(transparent)
  expect_true(all(transparent_array[1, 2:3, 4] == 0))

  expect_error(
    render_stack(
      list(image, wider),
      stack = "v",
      background = c(0.1, 0.2),
      convert_colorspace = FALSE,
      preview = FALSE
    ),
    "invalid color"
  )
})

test_that("render_alpha_outline builds padded alpha halos", {
  mask = matrix(c(0, 1, 1, 0), nrow = 2)

  unpadded = render_alpha_outline(
    mask = mask,
    expand = 0,
    color = "red",
    alpha = 0.4,
    preview = FALSE
  )
  expect_equal(dim(unpadded), c(2, 2, 4))
  expect_equal(unclass(unpadded)[,, 4], mask * 0.4)

  halo = render_alpha_outline(
    mask = mask,
    expand = 0,
    color = "red",
    alpha = 0.4,
    pad = 1,
    preview = FALSE
  )
  halo_array = unclass(halo)

  expect_equal(dim(halo), c(4, 4, 4))
  expect_equal(unname(attr(halo, "padding")), rep(1L, 4))
  expect_equal(halo_array[2:3, 2:3, 4], mask * 0.4)
  expect_true(all(halo_array[c(1, 4), , 4] == 0))
  expect_true(all(halo_array[, c(1, 4), 4] == 0))
})

test_that("render_alpha_outline composites below source images by default", {
  image = array(0, dim = c(3, 3, 4))
  image[2, 2, 1] = 1
  image[2, 2, 4] = 1

  outline_only = render_alpha_outline(
    image,
    expand = 1,
    edge_softness = 0.1,
    color = c(0, 0, 1),
    composite = FALSE,
    preview = FALSE
  )
  composited = render_alpha_outline(
    image,
    expand = 1,
    edge_softness = 0.1,
    color = c(0, 0, 1),
    preview = FALSE
  )
  outline_array = unclass(outline_only)
  composited_array = unclass(composited)

  expect_equal(outline_array[2, 2, 1:3], c(0, 0, 1))
  expect_equal(composited_array[2, 2, 1:3], c(1, 0, 0))
  expect_equal(composited_array[2, 2, 4], 1)
  expect_equal(composited_array[2, 1, ], outline_array[2, 1, ])
})

test_that("render_alpha_outline fills single-pixel distance gaps", {
  mask = matrix(
    c(
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      TRUE,
      FALSE,
      TRUE
    ),
    nrow = 3
  )

  unfilled = render_alpha_outline(
    mask = mask,
    expand = 1,
    edge_softness = 0.1,
    gap_fill = 0,
    preview = FALSE
  )
  filled = render_alpha_outline(
    mask = mask,
    expand = 1,
    edge_softness = 0.1,
    gap_fill = 1,
    preview = FALSE
  )
  unfilled_alpha = unclass(unfilled)[,, 4]
  filled_alpha = unclass(filled)[,, 4]

  expect_equal(unfilled_alpha[2, 2], 0)
  expect_equal(filled_alpha[2, 2], 0.5)
  expect_equal(filled_alpha[, 1], unfilled_alpha[, 1])
})

test_that("render_alpha_outline gap_fill controls bridged gap width", {
  two_pixel_gap = matrix(c(1, 0, 0, 0.5), nrow = 1)
  one_pixel_fill = fill_alpha_gaps(two_pixel_gap, max_gap = 1)
  two_pixel_fill = fill_alpha_gaps(two_pixel_gap, max_gap = 2)

  expect_equal(one_pixel_fill, two_pixel_gap)
  expect_equal(two_pixel_fill, matrix(c(1, 0.5, 0.5, 0.5), nrow = 1))

  mask = matrix(c(TRUE, FALSE, FALSE, FALSE, FALSE, TRUE), nrow = 1)
  one_pixel_outline = render_alpha_outline(
    mask = mask,
    expand = 1,
    edge_softness = 0.1,
    gap_fill = 1,
    preview = FALSE
  )
  two_pixel_outline = render_alpha_outline(
    mask = mask,
    expand = 1,
    edge_softness = 0.1,
    gap_fill = 2,
    preview = FALSE
  )

  expect_equal(
    unclass(one_pixel_outline)[,, 4],
    c(1, 0.5, 0, 0, 0.5, 1)
  )
  expect_equal(
    unclass(two_pixel_outline)[,, 4],
    c(1, 0.5, 0.5, 0.5, 0.5, 1)
  )
})

test_that("render_alpha_outline gap_fill only fills border-connected gaps", {
  interior_gap = matrix(0.5, nrow = 3, ncol = 3)
  interior_gap[2, 2] = 0

  protected_gap = fill_alpha_gaps(
    interior_gap,
    max_gap = 1,
    alpha_threshold = 0.25
  )
  threshold_fill = fill_alpha_gaps(
    interior_gap,
    max_gap = 1,
    alpha_threshold = 0.75
  )

  expect_equal(protected_gap[2, 2], 0)
  expect_equal(threshold_fill[2, 2], 0.5)

  mask = matrix(TRUE, nrow = 5, ncol = 5)
  mask[2:4, 2:4] = FALSE
  outline = render_alpha_outline(
    mask = mask,
    expand = 1,
    edge_softness = 0.1,
    gap_fill = 1,
    gap_fill_alpha_threshold = 0.25,
    preview = FALSE
  )

  expect_equal(unclass(outline)[3, 3, 4], 0)
})

test_that("render_image_overlay default compositing math is unchanged", {
  image = array(0.2, dim = c(2, 2, 3))
  overlay = array(0, dim = c(2, 2, 4))
  overlay[,, 1] = 0.8
  overlay[,, 2] = 0.4
  overlay[,, 3] = 0.1
  overlay[,, 4] = matrix(c(0.25, 0.5, 0.75, 1), nrow = 2)

  default_output = render_image_overlay(
    image,
    overlay,
    convert_overlay_colorspace = FALSE,
    preview = FALSE
  )
  sprite_output = render_sprite_overlay(
    image,
    overlay,
    convert_overlay_colorspace = FALSE,
    overlay_coords = c(1, 1),
    preserve_channels = FALSE,
    preview = FALSE
  )

  expected = array(1, dim = c(2, 2, 4))
  for (channel in 1:3) {
    expected[,, channel] = overlay[,, channel] *
      overlay[,, 4] +
      image[,, channel] * (1 - overlay[,, 4])
  }
  default_array = array(unclass(default_output), dim = dim(default_output))
  sprite_array = array(unclass(sprite_output), dim = dim(sprite_output))

  expect_equal(default_array, expected)
  expect_equal(default_array, sprite_array)
})

test_that("render_image_overlay still resizes mismatched overlays by default", {
  image = array(0, dim = c(4, 5, 3))
  overlay = array(0, dim = c(2, 2, 4))
  overlay[,, 1] = 1
  overlay[,, 4] = 1

  output = render_image_overlay(
    image,
    overlay,
    convert_overlay_colorspace = FALSE,
    preview = FALSE
  )

  expect_equal(dim(output), c(4, 5, 4))
  expect_equal(
    unclass(output)[,, 1],
    matrix(1, nrow = 4, ncol = 5),
    tolerance = 1e-6
  )
})

test_that("render_image_overlay does not expose sprite placement arguments", {
  image = array(0, dim = c(2, 2, 3))
  overlay = array(0, dim = c(2, 2, 4))

  expect_error(
    render_image_overlay(
      image,
      overlay,
      overlay_coords = c(1, 1),
      preview = FALSE
    ),
    "unused argument"
  )
})

#'@title Render Boolean Distance
#'
#'@description Takes a logical matrix and returns the nearest distance to each TRUE value.
#'
#'@param boolean Logical matrix (or matrix of 1s and 0s), where distance will be measured to the `TRUE` values.
#'@param rescale Default `FALSE`. Rescales the calculated distance to a range of 0-1.
#'Useful for visualizing the distance matrix.
#'@return Matrix of distance values.
#'@export
#'@examplesIf interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")
#'#Measure distance to
#'plot_image(render_boolean_distance(t(volcano) > 150))
#'plot_image(render_boolean_distance(t(volcano) < 150))
#'#If we want to rescale this to zero to one (to visualize like an image), set rescale=TRUE
#'plot_image(render_boolean_distance(t(volcano) > 150,rescale=TRUE))
render_boolean_distance = function(boolean, rescale = FALSE) {
  if (any(is.na(boolean))) {
    stop("`boolean` must not have any NA values")
  }
  distval = get_boolean_distance(boolean)
  if (rescale) {
    if (min(distval) != max(distval)) {
      distval = (distval - min(distval)) / (max(distval) - min(distval))
    }
    return(distval)
  } else {
    return(distval)
  }
}

#' @title Handle Image Output
#'
#' @description Handles image output logic: returning, saving to a file, or previewing.
#'
#' @param image 3-layer RGB/4-layer RGBA array, `rayimg` class, or filename of an image.
#' @param filename Default `NULL`. String specifying the file path to save the image.
#' If `NULL`, the image is returned or displayed.
#' @param preview Default `FALSE`. If `TRUE`, the image is displayed.
#'
#' @return If `filename = NULL` and `preview = FALSE`, returns the image array invisibly.
#' Otherwise, saves to file or displays the image.
#'
#' @keywords internal
handle_image_output = function(image, filename = NULL, preview = FALSE) {
  if (is.null(filename) || is.na(filename)) {
    if (!preview) {
      return(image) # Return the image data
    }
    plot_image(render_clamp(image)) # Display image
    return(invisible(image))
  } else {
    ray_write_image(image, filename) # Save the image to file
    if (preview) {
      plot_image(render_clamp(image)) # Display saved image
    }
    return(invisible(image))
  }
}

#'@title Render Convolution FFT
#'
#'@description Takes an image and applies a convolution operation to it, using
#'a user-supplied or built-in kernel. This function uses a fast Fourier transform and
#'does the convolution in the frequency domain, so it should be faster for much larger kernels.
#'
#'@param image 3-layer RGB/4-layer RGBA array, `rayimg` class, or filename of an image.
#'@param filename Default `NULL`. The filename of the image to be saved. If this is not given, the image will be plotted instead.
#'@param kernel Default `gaussian`. By default, an 11x11 Gaussian kernel with a mean
#'of `0` and a standard deviation of `1`, running from `-kernel_extent` to `kernel_extent`.
#'If numeric, this will be the standard deviation of the normal distribution. If a
#'matrix, it will be used directly as the convolution kernel (but resized always to be an odd number
#'of columns and rows).
#'@param kernel_dim Default `c(11, 11)`. The dimension of the `gaussian` kernel. Ignored
#'if user specifies their own kernel.
#'@param kernel_extent Default `3`. Extent over which to calculate the kernel.
#'@param absolute Default `TRUE`. Whether to take the absolute value of the convolution.
#'@param pad Default `50`. Amount to pad the image to remove edge effects.
#'@param include_alpha Default `FALSE`. Whether to include the alpha channel in the convolution.
#'@param preview Default `FALSE`. Whether to plot the convolved image, or just to return the values.
#'@return A `rayimg` RGBA array.
#'@export
#'@examplesIf interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")
#'#Perform a convolution with the default gaussian kernel
#'plot_image(dragon)
#'#Perform a convolution with the default gaussian kernel
#'render_convolution_fft(dragon, kernel=0.1, preview = TRUE)
#'#Increase the width of the kernel
#'render_convolution_fft(dragon, kernel = 2, kernel_dim=21,kernel_extent=6, preview = TRUE)
#'#Use a built-in kernel:
#'render_convolution_fft(dragon, kernel = generate_2d_exponential(falloff=2, dim=31, width=21),
#'                       preview = TRUE)
#'#Perform edge detection
#'edge = matrix(c(-1,-1,-1,-1,8,-1,-1,-1,-1),3,3)
#'render_convolution_fft(render_bw(dragon), kernel = edge, preview = TRUE)
#'#Perform edge detection with Sobel matrices
#'sobel1 = matrix(c(1,2,1,0,0,0,-1,-2,-1),3,3)
#'sobel2 = matrix(c(1,2,1,0,0,0,-1,-2,-1),3,3,byrow=TRUE)
#'sob1 = render_convolution_fft(render_bw(dragon), kernel = sobel1)
#'sob2 = render_convolution_fft(render_bw(dragon), kernel = sobel2)
#'sob_all = sob1 + sob2
#'plot_image(sob1)
#'plot_image(sob2)
#'plot_image(sob_all)
#'#We can also apply this function to matrices:
#'volcano |> image()
#'volcano |>
#'  render_convolution_fft(kernel=generate_2d_gaussian(sd=1,dim=31)) |>
#'  image()
#'# Because this function uses the fast Fourier transform, large kernels will be much faster
#'# than the same size kernels in [render_convolution()]
#' render_convolution_fft(dragon, kernel_dim = c(200,200) , preview = TRUE)
#'#Use a custom kernel (in this case, an X shape):
#'custom = diag(10) + (diag(10)[,10:1])
#'#Normalize
#'custom = custom / 20
#'plot_image(custom*20)
#'render_convolution_fft(dragon, kernel = custom, preview = TRUE)
render_convolution_fft = function(
  image,
  kernel = "gaussian",
  kernel_dim = c(11, 11),
  kernel_extent = 3,
  absolute = TRUE,
  pad = 50,
  include_alpha = FALSE,
  filename = NULL,
  preview = FALSE
) {
  shift_fft = function(fft_mat) {
    nr = dim(fft_mat)[1]
    nc = dim(fft_mat)[2]
    if (nr %% 2 == 0) {
      nr_mid1 = nr / 2
      nr_mid2 = nr / 2 + 1
    } else {
      nr_mid1 = floor(nr / 2)
      nr_mid2 = floor(nr / 2) + 1
    }
    if (nc %% 2 == 0) {
      nc_mid1 = nc / 2
      nc_mid2 = nc / 2 + 1
    } else {
      nc_mid1 = floor(nc / 2)
      nc_mid2 = floor(nc / 2) + 1
    }
    fftcorn_nw = fft_mat[1:nr_mid1, 1:nc_mid1]
    fftcorn_ne = fft_mat[1:nr_mid1, nc_mid2:nc]
    fftcorn_sw = fft_mat[nr_mid2:nr, 1:nc_mid1]
    fftcorn_se = fft_mat[nr_mid2:nr, nc_mid2:nc]
    rbind(cbind(fftcorn_se, fftcorn_sw), cbind(fftcorn_ne, fftcorn_nw))
  }
  temp_image = ray_read_image(
    image,
    convert_to_array = FALSE,
    reset_camera_settings = TRUE
  )
  colorspace = attr(temp_image, "colorspace")
  whitepoint = attr(temp_image, "white_current")

  #Check if file or image before below:
  if (is.character(kernel)) {
    if (kernel == "gaussian") {
      kernel = generate_2d_gaussian(1, 1, kernel_dim, kernel_extent)
    }
  }
  if (is.numeric(kernel) && length(kernel) == 1) {
    kernel = generate_2d_gaussian(kernel, 1, kernel_dim, kernel_extent)
  }

  if (any(dim(kernel)[1:2] != dim(temp_image)[1:2])) {
    if (all(dim(kernel)[1:2] <= dim(temp_image)[1:2])) {
      kernel = expand_to_fit(dim(temp_image)[1:2], kernel)
    } else {
      stop("kernel can't have greater dimensions than image")
    }
  }
  if (pad != 0) {
    temp_image = add_multi_padding(temp_image, pad)
    kernel = add_multi_padding(kernel, pad)
  }

  if (any(dim(kernel) > dim(temp_image)[1:2] * 2 + 1)) {
    stop(
      "kernel dimensions: ",
      paste0(dim(kernel), collapse = "x"),
      " must not be greater than 2x image dimensions: ",
      paste0(dim(temp_image)[1:2], collapse = "x"),
      ", plus one (here, ",
      paste0(dim(temp_image)[1:2] * 2 + 1, collapse = "x"),
      ")."
    )
  }

  temp_fft = temp_image
  if (length(dim(temp_image)) == 2) {
    temp_fft = stats::fft(temp_image)
  } else if (length(dim(temp_image)) == 3) {
    channels = dim(temp_image)[3]
    if (!include_alpha && (channels == 2 || channels == 4)) {
      max_channel = channels - 1
    } else {
      max_channel = channels
    }
    for (i in seq_len(max_channel)) {
      temp_fft[,, i] = stats::fft(temp_image[,, i])
    }
  }

  kernal_fft = stats::fft(kernel)
  vals = Re(temp_fft)
  if (length(dim(temp_fft)) == 3) {
    channels = dim(temp_image)[3]
    if (!include_alpha && (channels == 2 || channels == 4)) {
      max_channel = channels - 1
    } else {
      max_channel = channels
    }
    for (i in seq_len(max_channel)) {
      vals[,, i] = shift_fft(
        Re(stats::fft(temp_fft[,, i] * kernal_fft, inverse = TRUE)) /
          length(vals[,, i])
      )
    }
  } else {
    vals = shift_fft(
      Re(stats::fft(temp_fft * kernal_fft, inverse = TRUE)) / length(vals)
    )
  }
  if (absolute) {
    vals = abs(vals)
  }

  if (pad != 0) {
    vals = trim_padding(vals, pad)
  }
  final_image = ray_read_image(
    vals,
    assume_colorspace = colorspace,
    assume_white = whitepoint,
    source_linear = TRUE
  )
  handle_image_output(final_image, filename = filename, preview = preview)
}

#' Apply exposure compensation and ISO gain
#'
#' @param image              3-layer RGB/4-layer RGBA array, `rayimg` class, or filename of an image.
#' @param exposure           Default `0`. Exposure compensation in stops; applied directly to RGB values.
#' @param iso                Default `NA`. ISO gain to apply directly to RGB values. `NA` applies no ISO gain
#'   and preserves the current informational ISO metadata.
#' @param auto               Default `FALSE`. If `TRUE`, automatically add the exposure needed to
#'   place a luminance percentile at 1.0 for LDR display/output. `exposure`
#'   is then treated as an additional compensation value.
#' @param percentile         Default `0.99`. Luminance percentile used when `auto = TRUE`.
#'   Must be greater than 0 and less than or equal to 1. Use `1` for max luminance.
#' @param verbose            Default `FALSE`. If `TRUE` and `auto = TRUE`, print the
#'   automatically computed exposure adjustment.
#' @param filename           Default `NA`. Output path.
#' @param preview            Default `FALSE`. If `TRUE`, display the image.
#' @param ...                Additional args passed to [plot_image()] (when `preview=TRUE`)
#'   or to [ray_write_image()] (when `filename` is given).
#'
#' @return A `rayimg` RGBA array.
#' @export
#'@examplesIf interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")
#' # LDR/sRGB (auto): decodes to linear, applies EV, records camera metadata
#'   render_exposure(dragon, exposure = +1, preview = TRUE)
#' # Force linear/HDR behavior
#'   render_exposure(dragon * 2, exposure = -1, preview = TRUE)
render_exposure = function(
  image,
  exposure = 0,
  iso = NA,
  auto = FALSE,
  percentile = 0.99,
  verbose = FALSE,
  filename = NA,
  preview = FALSE,
  ...
) {
  img = ray_read_image(image)
  exposure = rayimg_exposure_value(exposure)
  if (length(iso) != 1L) {
    stop("render_exposure(): 'iso' must be a single positive number or NA.")
  }
  iso_supplied = !is.na(iso)
  if (iso_supplied) {
    iso_to_apply = rayimg_iso_value(iso)
    iso_metadata = iso_to_apply
  } else {
    iso_to_apply = 100
    iso_metadata = rayimg_iso_value(attr(img, "iso", exact = TRUE))
  }
  if (!is.logical(auto) || length(auto) != 1L || is.na(auto)) {
    stop("render_exposure(): 'auto' must be TRUE or FALSE.")
  }
  if (
    !is.numeric(percentile) ||
      length(percentile) != 1L ||
      is.na(percentile) ||
      !is.finite(percentile) ||
      percentile <= 0 ||
      percentile > 1
  ) {
    stop("render_exposure(): 'percentile' must be a number in (0, 1].")
  }
  if (!is.logical(verbose) || length(verbose) != 1L || is.na(verbose)) {
    stop("render_exposure(): 'verbose' must be TRUE or FALSE.")
  }

  current_exposure = rayimg_exposure_value(attr(img, "exposure", exact = TRUE))
  auto_exposure = 0
  target_luminance = NA_real_
  if (auto) {
    img_values = unclass(img)
    d = dim(img_values)
    if (length(d) == 2L) {
      luminance_values = as.numeric(img_values)
    } else {
      channels = d[3L]
      color_channels = if (channels %in% c(2L, 4L)) {
        seq_len(channels - 1L)
      } else {
        seq_len(channels)
      }
      if (length(color_channels) == 1L) {
        luminance_values = as.numeric(img_values[,, color_channels])
      } else {
        rgb_values = img_values[,, color_channels, drop = FALSE]
        if (length(color_channels) >= 3L) {
          rgb_values = rgb_values[,, seq_len(3L), drop = FALSE]
          cs = attr(img, "colorspace", exact = TRUE)
          luma_weights = if (
            is.list(cs) &&
              is.matrix(cs$rgb_to_xyz) &&
              all(dim(cs$rgb_to_xyz) >= c(2L, 3L))
          ) {
            as.numeric(cs$rgb_to_xyz[2L, seq_len(3L)])
          } else {
            c(0.2126, 0.7152, 0.0722)
          }
          luminance_values =
            rgb_values[,, 1L] *
            luma_weights[1L] +
            rgb_values[,, 2L] * luma_weights[2L] +
            rgb_values[,, 3L] * luma_weights[3L]
        } else {
          luminance_values = as.numeric(rgb_values)
        }
      }
    }
    luminance_values = luminance_values[is.finite(luminance_values)]
    if (length(luminance_values)) {
      target_luminance = as.numeric(stats::quantile(
        luminance_values,
        probs = percentile,
        names = FALSE
      ))
      current_scale = iso_to_apply / 100
      if (target_luminance > 0 && current_scale > 0) {
        auto_exposure = -log2(target_luminance * current_scale)
        exposure = exposure + auto_exposure
      }
    }
    if (verbose) {
      message(sprintf(
        "Auto exposure: %+.3f EV (luminance p%.3g: %.6g, ISO: %.6g)",
        auto_exposure,
        percentile,
        target_luminance,
        iso_to_apply
      ))
    }
  }
  scale = (2^exposure) * (iso_to_apply / 100)
  out_values = unclass(img)
  if (!isTRUE(all.equal(scale, 1))) {
    d = dim(out_values)
    if (length(d) == 2L) {
      out_values = out_values * scale
    } else {
      channels = d[3L]
      color_channels = if (channels %in% c(2L, 4L)) {
        seq_len(channels - 1L)
      } else {
        seq_len(channels)
      }
      out_values[,, color_channels] = out_values[,, color_channels] * scale
    }
  }

  out = rayimg(
    out_values,
    filetype = attr(img, "filetype", exact = TRUE),
    source_linear = attr(img, "source_linear", exact = TRUE),
    colorspace = attr(img, "colorspace", exact = TRUE),
    white_current = attr(img, "white_current", exact = TRUE),
    exposure = current_exposure + exposure,
    iso = iso_metadata
  )
  out = ray_read_image(out)
  handle_image_output(out, filename = filename, preview = preview)
}

#' Render To Display
#' @description Convert an image from its working space to a display space, and optionally apply sRGB OETF for viewing.
#' @param image A `rayimg`, array, or filename.
#' @param display Default `CS_SRGB`. Target display RGB space.
#' @param encode Default `TRUE`. If `TRUE`, apply sRGB OETF (only valid when `display` is sRGB-like).
#' @param adapt_white Default `TRUE`. Whether to perform CAT from working white to display white.
#' @return A `rayimg` when `encode=FALSE`, or a `rayimg` with sRGB-encoded RGB when `encode=TRUE`.
#' @export
render_to_display = function(
	image,
	display = CS_SRGB,
	encode = TRUE,
	adapt_white = TRUE
) {
	img = ray_read_image(
		image,
		normalize = FALSE,
		reset_camera_settings = TRUE
	)
	# Ensure linear before color conversion
	if (!isTRUE(attr(img, "source_linear"))) {
		img = render_gamma_linear(img, srgb_to_linear = TRUE)
	}
	# Move primaries (and optionally white) to display
	moved = render_convert_colorspace(
		img,
		from_mats = NA,
		to_mats = display,
		adapt_white = adapt_white,
		from_white = NA,
		to_white = display$white_xyz
	)
	if (!encode) {
		return(moved)
	}
	# Only sRGB OETF is implemented here
	if (!identical(display$name, CS_SRGB$name)) {
		warning(
			"render_to_display(): encode=TRUE uses sRGB OETF; non-sRGB displays may require a different OETF/transfer."
		)
	}
	render_gamma_linear(moved, srgb_to_linear = FALSE)
}

#'@title Generate 2D exponential Distribution
#'
#'@description Generates a 2D exponential distribution, with an optional
#'argument to take the exponential to a user-defined power.
#'
#'@param falloff Default `1`. Falloff of the exponential.
#'@param dim Default `c(11, 11)`. The dimensions of the matrix.
#'@param width Default `3` (`-10` to `10`). The range in which to compute the distribution.
#'@param rescale_unity Default `FALSE`. If `TRUE`, this will rescale the max value to one. Useful
#'if wanting to plot the distribution with [plot_image()].
#'@import stats
#'@export
#'@examplesIf interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")
#'image(generate_2d_exponential(1,31,3), asp=1)
generate_2d_exponential = function(
  falloff = 1,
  dim = c(11, 11),
  width = 3,
  rescale_unity = FALSE
) {
  if (length(dim) == 1) {
    dim = c(dim, dim)
  }
  xy_ratio = dim[1] / dim[2]
  if (xy_ratio > 1) {
    x = seq(-width * xy_ratio, width * xy_ratio, length.out = dim[1])
    y = seq(-width, width, length.out = dim[2])
  } else {
    x = seq(-width, width, length.out = dim[1])
    y = seq(-width / xy_ratio, width / xy_ratio, length.out = dim[2])
  }
  testmat = matrix(0, dim[1], dim[2])
  for (i in 1:length(x)) {
    for (j in 1:length(y)) {
      testmat[i, j] = (dexp(sqrt(x[i]^2 + y[j]^2), rate = 1 / falloff))
    }
  }
  if (rescale_unity) {
    temp = testmat / sum(testmat)
    return(temp / max(temp))
  }
  testmat / sum(testmat)
}

#' Normalize/label a white point
#' @param white_point Default `NULL`. Named ("D65","D60","D55","D50","D75","E") or XYZ (length 3, any Y).
#' @returns list(value = XYZ with Y=1, white_name = name or "Custom")
#' @keywords internal
get_whitepoint_xyz = function(white_point = NULL) {
	std = list(
		"D65" = c(0.95047, 1, 1.08883),
		"D60" = c(0.95264, 1, 1.00827),
		"D55" = c(0.95560, 1, 0.92149),
		"D50" = c(0.96422, 1, 0.82521),
		"D75" = c(0.94972, 1, 1.22638),
		"E" = c(1, 1, 1)
	)
	if (is.null(white_point)) {
		return(NULL)
	}
	if (is.character(white_point) && length(white_point) == 1L) {
		wp = std[[toupper(white_point)]]
		if (is.null(wp)) {
			stop("Unknown white: ", white_point)
		}
		return(list(value = wp, white_name = white_point))
	}
	stopifnot(is.numeric(white_point), length(white_point) == 3L)
	Y = white_point[2]
	if (!is.finite(Y) || Y <= 0) {
		stop("White Y must be positive and finite.")
	}
	list(value = white_point / Y, white_name = "Custom")
}

#' Bradford CAT matrix from source to target white
#' @param source_xyz Default `c(0.95264,1,1.00827)`. Source XYZ white, Y=1.
#' @param target_xyz Default `c(0.95047,1,1.08883)`. Target XYZ white, Y=1.
#' @param eps Default `1e-8`. Small epsilon.
#' @return 3x3 numeric CAT matrix to apply in XYZ.
#' @keywords internal
compute_cat_bradford = function(
	source_xyz = c(0.95264, 1.0, 1.00827),
	target_xyz = c(0.95047, 1.0, 1.08883),
	eps = 1e-8
) {
	B = matrix(
		c(
			0.8951,
			0.2664,
			-0.1614,
			-0.7502,
			1.7135,
			0.0367,
			0.0389,
			-0.0685,
			1.0296
		),
		3,
		3,
		byrow = TRUE
	)
	src = B %*% source_xyz
	dst = B %*% target_xyz
	if (any(abs(src) < eps)) {
		stop("CAT: source LMS too close to zero")
	}
	S = diag(as.numeric(dst / src))
	solve(B) %*% S %*% B
}

#' Apply a 3x3 color transform to an image
#' @param img 3/4-channel array (RGB/A). 2D (grey) or 2-channel (grey+alpha) pass through.
#' @param M 3x3 matrix.
#' @return Array with transformed RGB; alpha preserved.
#' @keywords internal
apply_color_matrix = function(img, M = diag(3)) {
	stopifnot(is.array(img), is.matrix(M), all(dim(M) == c(3, 3)))
	d = dim(img)
	if (length(d) == 2L) {
		return(img)
	}
	ch = d[3]
	if (ch < 3L) {
		return(img)
	}
	flat = matrix(img[,, 1:3], ncol = 3)
	flat = flat %*% t(M)
	out = img
	out[,, 1:3] = array(flat, dim = c(d[1], d[2], 3))
	out
}

#' sRGB to Linear
#' @param x Values
#' @returns Transformed values
#' @keywords internal
to_linear = function(x) {
	out = x
	low = x <= 0.04045
	out[low] = x[low] / 12.92
	out[!low] = ((x[!low] + 0.055) / 1.055)^2.4
	out
}

#' Linear to sRGB
#' @param x Values
#' @returns Transformed values
#' @keywords internal
to_srgb = function(x) {
	x = pmin(pmax(x, 0), 1)
	out = x
	low = x <= 0.0031308
	out[low] = 12.92 * x[low]
	out[!low] = 1.055 * x[!low]^(1 / 2.4) - 0.055
	out
}

#' Encode RGB/RGBA to hex (sRGB domain)
#' @param x Values. 3 or 4 columns (R,G,B,A) in 0 to 1.
#' @return Hex vector (#RRGGBB or #RRGGBBAA).
#' @keywords internal
rgb_to_hex = function(x) {
	if (is.null(dim(x))) {
		x = matrix(x, ncol = length(x))
	}
	if (!(ncol(x) %in% c(3L, 4L))) {
		stop("rgb_to_hex(): expected 3 or 4 columns (R,G,B[,A]).")
	}
	x = pmin(pmax(x, 0), 1)
	ints = round(x * 255)
	apply(ints, 1, function(row) {
		paste0("#", paste(sprintf("%02X", row), collapse = ""))
	})
}

#' Array hex vector in column-major order
#' @param arr Array with c>=3
#' @return Hex vector length h*w
#' @keywords internal
array_rgb_to_hex_colmajor = function(arr) {
	d = dim(arr)
	if (length(d) != 3L || d[3] < 3L) {
		stop("array_rgb_to_hex_colmajor(): need h*w*(3|4).")
	}
	R = as.vector(arr[,, 1])
	G = as.vector(arr[,, 2])
	B = as.vector(arr[,, 3])
	if (d[3] >= 4L) {
		A = as.vector(arr[,, 4])
		return(rgb_to_hex(cbind(R, G, B, A)))
	}
	rgb_to_hex(cbind(R, G, B))
}

#'@title Matrix/Array Interpolation
#'
#'@description Given a series of X and Y coordinates and an array/matrix, interpolates the Z coordinate
#'using bilinear interpolation.
#'
#'@param image 3-layer RGB/4-layer RGBA array, `rayimg` class, or filename of an image.
#'@param x X indices (or fractional index) to interpolate.
#'@param y Y indices (or fractional index) to interpolate.
#'
#'@return Either a vector of values (if image is a matrix) or a list of interpolated values
#'from each layer.
#'@export
#'
#'@examples
#'#if(interactive()){
#'#Interpolate a matrix
#'interpolate_array(volcano,c(10,10.1,11),c(30,30.5,33))
#'#Interpolate a 3-layer array (returns list for each channel)
#'interpolate_array(dragon,c(10,10.1,11),c(30,30.5,33))
#'#end}
interpolate_array = function(image, x, y) {
  image = ray_read_image(image) #Always output RGBA array
  #Check if file or image before below:
  is_matrix = length(dim(image)) == 2
  xy = matrix(c(x, y), nrow = length(x), ncol = 2)

  if (is_matrix) {
    return(apply(xy, 1, (function(x) rayinterp2(image, x[1], x[2]))))
  }
  #Load and rotate images if image
  output = list()
  if (dim(image)[3] == 2) {
    output$grey = apply(
      xy,
      1,
      (function(x) rayinterp2(image[,, 1], x[1], x[2]))
    )
    output$alpha = apply(
      xy,
      1,
      (function(x) rayinterp2(image[,, 2], x[1], x[2]))
    )
  } else if (dim(image)[3] == 3) {
    output$red = apply(xy, 1, (function(x) rayinterp2(image[,, 1], x[1], x[2])))
    output$green = apply(
      xy,
      1,
      (function(x) rayinterp2(image[,, 2], x[1], x[2]))
    )
    output$blue = apply(
      xy,
      1,
      (function(x) rayinterp2(image[,, 3], x[1], x[2]))
    )
  } else if (dim(image)[3] == 4) {
    output$red = apply(xy, 1, (function(x) rayinterp2(image[,, 1], x[1], x[2])))
    output$green = apply(
      xy,
      1,
      (function(x) rayinterp2(image[,, 2], x[1], x[2]))
    )
    output$blue = apply(
      xy,
      1,
      (function(x) rayinterp2(image[,, 3], x[1], x[2]))
    )
    output$alpha = apply(
      xy,
      1,
      (function(x) rayinterp2(image[,, 4], x[1], x[2]))
    )
  }
  return(output)
}

#'@title Calculate Bokeh Size
#'
#'@description Calculates the amount of blurring at each point with camera characteristics.
#'
#'@param z Depth matrix.
#'@param zfocus Depth in which to blur. Minimum 0, maximum 1.
#'@param f Focal length of the virtual camera. .
#'@param N F-stop. Focal length of the virtual camera.
#'@keywords internal
#'@return Matrix of bokeh sizes.
calc_bokeh_size = function(z, zfocus, f, N, ramp) {
  ifelse(
    z - zfocus < 0,
    abs(f^2 * abs(z - zfocus) * ramp / ((zfocus - f) * z * N)),
    abs(f^2 * abs(z - zfocus) / ((zfocus - f) * z * N * ramp))
  )
}

#'@title Render Convolution
#'
#'@description Takes an image and applies a convolution operation to it, using
#'a user-supplied or built-in kernel. Edges are calculated by limiting the size
#'of the kernel to only that overlapping the actual image (renormalizing the
#'kernel for the edges).
#'
#'@param image 3-layer RGB/4-layer RGBA array, `rayimg` class, or filename of an image.
#'@param filename Default `NULL`. The filename of the image to be saved. If this is not given, the image will be plotted instead.
#'@param kernel Default `gaussian`. By default, an 11x11 Gaussian kernel with a mean
#'of `0` and a standard deviation of `1`, running from `-kernel_extent` to `kernel_extent`.
#'If numeric, this will be the standard deviation of the normal distribution. If a
#'matrix, it will be used directly as the convolution kernel (but resized always to be an odd number
#'of columns and rows).
#'@param kernel_dim Default `11`. The dimension of the `gaussian` kernel. Ignored
#'if user specifies their own kernel.
#'@param kernel_extent Default `3`. Extent over which to calculate the kernel.
#'@param absolute Default `TRUE`. Whether to take the absolute value of the convolution.
#'@param min_value Default `NULL`. If numeric, specifies the minimum value (for any color channel)
#'for a pixel to have the convolution performed.
#'@param include_alpha Default `FALSE`. Whether to include the alpha channel in the convolution.
#'@param preview Default `TRUE`. Whether to plot the convolved image, or just to return the values.
#'@param progress Default `TRUE`. Whether to display a progress bar.
#'@return A `rayimg` RGBA array.
#'@export
#'@examplesIf interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")
#'#Perform a convolution with the default gaussian kernel
#'plot_image(dragon)
#'#Perform a convolution with the default gaussian kernel
#'render_convolution(dragon, preview = TRUE)
#'#Increase the width of the kernel
#'render_convolution(dragon, kernel = 2, kernel_dim=21,kernel_extent=6, preview = TRUE)
#'#Perform edge detection using an edge detection kernel
#'edge = matrix(c(-1,-1,-1,-1,8,-1,-1,-1,-1),3,3)
#'render_convolution(render_bw(dragon), kernel = edge, preview = TRUE, absolute=FALSE)
#'#Perform edge detection with Sobel matrices
#'sobel1 = matrix(c(1,2,1,0,0,0,-1,-2,-1),3,3)
#'sobel2 = matrix(c(1,2,1,0,0,0,-1,-2,-1),3,3,byrow=TRUE)
#'sob1 = render_convolution(render_bw(dragon), kernel = sobel1)
#'sob2 = render_convolution(render_bw(dragon), kernel = sobel2)
#'sob_all = sob1 + sob2
#'plot_image(sob1)
#'plot_image(sob2)
#'plot_image(sob_all)
#'
#'#Only perform the convolution on bright pixels (bloom)
#'render_convolution(dragon, kernel = 5, kernel_dim=24, kernel_extent=24,
#'                   min_value=1, preview = TRUE)
#'#Use a built-in kernel:
#'render_convolution(dragon, kernel = generate_2d_exponential(falloff=2, dim=31, width=21),
#'                   preview = TRUE)
#'#We can also apply this function to matrices:
#'volcano |> image()
#'volcano |>
#'  render_convolution(kernel=generate_2d_gaussian(sd=1,dim=31)) |>
#'  image()
#'#Use a custom kernel (in this case, an X shape):
#'custom = diag(10) + (diag(10)[,10:1])
#'plot_image(custom)
#'render_convolution(dragon, kernel = custom, preview = TRUE)
render_convolution = function(
  image,
  kernel = "gaussian",
  kernel_dim = 11,
  kernel_extent = 3,
  absolute = TRUE,
  min_value = NULL,
  include_alpha = FALSE,
  filename = NULL,
  preview = FALSE,
  progress = FALSE
) {
  if (is.character(kernel)) {
    if (kernel == "gaussian") {
      kernel = generate_2d_gaussian(1, 1, kernel_dim, kernel_extent)
    }
  }
  if (is.numeric(kernel) && length(kernel) == 1) {
    kernel = generate_2d_gaussian(kernel, 1, kernel_dim, kernel_extent)
  }
  if (!is.null(dim(kernel)) && all(dim(kernel) >= 2)) {
    if (ncol(kernel) %% 2 == 0) {
      newkernel = matrix(0, ncol = ncol(kernel) + 1, nrow = nrow(kernel))
      newkernel[, 1:ncol(kernel)] = kernel
      kernel = newkernel
    }
    if (nrow(kernel) %% 2 == 0) {
      newkernel = matrix(0, ncol = ncol(kernel), nrow = nrow(kernel) + 1)
      newkernel[1:nrow(kernel), ] = kernel
      kernel = newkernel
    }
  }
  temp_image = ray_read_image(
    image,
    convert_to_array = FALSE,
    reset_camera_settings = TRUE
  )
  colorspace = attr(temp_image, "colorspace")
  whitepoint = attr(temp_image, "white_current")
  imagetype = attr(temp_image, "filetype")
  if (any(dim(kernel) > dim(temp_image)[1:2] * 2 + 1)) {
    stop(
      "kernel dimensions: ",
      paste0(dim(kernel), collapse = "x"),
      " must not be greater than 2x image dimensions: ",
      paste0(dim(temp_image)[1:2], collapse = "x"),
      ", plus one (here, ",
      paste0(dim(temp_image)[1:2] * 2 + 1, collapse = "x"),
      ")."
    )
  }

  bloom_matrix = matrix(TRUE, nrow = nrow(temp_image), ncol = ncol(temp_image))
  if (length(dim(temp_image)) == 3) {
    if (!is.null(min_value)) {
      bool_value = rep(TRUE, length = length(temp_image[,, 1]))
      for (i in seq_along(dim(temp_image)[3])) {
        bool_value = bool_value & temp_image[,, i] <= min_value
      }
      bloom_matrix[bool_value] = FALSE
    }
  } else {
    if (!is.null(min_value)) {
      bloom_matrix[temp_image <= min_value] = FALSE
    }
  }
  temp_image_final = temp_image
  if (imagetype != "matrix") {
    channels = dim(temp_image)[3]
    if (!include_alpha && (channels == 2 || channels == 4)) {
      max_channel = channels - 1
    } else {
      max_channel = channels
    }
    for (i in seq_len(max_channel)) {
      temp_image_final[,, i] = convolution_cpp(
        temp_image[,, i],
        kernel = kernel,
        progbar = progress,
        channel = i,
        bloom_matrix = bloom_matrix
      )
    }
  } else {
    temp_image_final = convolution_cpp(
      temp_image,
      kernel = kernel,
      progbar = progress,
      channel = 1,
      bloom_matrix = bloom_matrix
    )
  }
  if (absolute) {
    temp_image_final = abs(temp_image_final)
  }
  final_image = ray_read_image(
    temp_image_final,
    assume_colorspace = colorspace,
    assume_white = whitepoint,
    source_linear = TRUE
  )
  handle_image_output(final_image, filename = filename, preview = preview)
}

#' @title Render Color Correction Matrix (3x3, linear RGB)
#'
#' @description Apply a technical/look 3x3 matrix to an image
#'
#' @param image Default `NULL`. 3-layer RGB/4-layer RGBA array, `rayimg`, or filename.
#' @param matrix Default `diag(3)`. 3x3 numeric matrix.
#' @param filename Default `NULL`. Output path.
#' @param preview Default `FALSE`. If `TRUE`, display the image.
#' @return A `rayimg` RGBA array.
#' @export
#' @examples
#' # We will start with an image that's too warm--we want to correct it to
#' # match the color of the second as closely as possible.
#' dragon_D50 = render_white_balance(dragon, target_white = "D50", bake = TRUE)
#' dragon_D75 = render_white_balance(dragon, target_white = "D75", bake = TRUE)
#' plot_image(dragon_D50)
#' plot_image(dragon_D75)
#'
#' # Fucntion to fit a color correction matrix
#' fit_cc_matrix = function(src, tgt) {
#' 	stopifnot(ncol(src) == 3, ncol(tgt) == 3, nrow(src) == nrow(tgt))
#' 	M_t = apply(tgt, 2, function(y) qr.solve(src, y))
#' 	t(M_t)
#' }
#'
#' # Sample N RGBs from an image for fit
#' rand_samples = function(img, n = 5000, seed=1) {
#' 	set.seed(seed)
#' 	d = dim(img)
#' 	ys = sample.int(d[1], n, TRUE); xs = sample.int(d[2], n, TRUE)
#' 	cbind(img[cbind(ys,xs,1)], img[cbind(ys,xs,2)], img[cbind(ys,xs,3)])
#' }
#'
#' S = rand_samples(dragon_D50)
#' T = rand_samples(dragon_D75)
#'
#' M = fit_cc_matrix(S, T)
#' # Optionally, regularize toward identity to avoid overfitting
#' #  M = 0.8 * M + 0.2 * diag(3)
#'
#' updated_source = render_color_correction(dragon_D50, matrix = M)
#'
#' # Matching the old image (grey) to the target image (orange),
#' # giving a new image (black).
#' plot(S,pch=16,cex=0.5,col="grey")
#' points(T,pch=16,cex=0.5,col="orange")
#' points(rand_samples(updated_source,seed=2),pch=16,cex=0.5,col="black")
#'
#' # Plot the images
#' plot_image(render_title(dragon_D50, "OG Image",
#'            title_color = "white", title_size = 12))
#' plot_image(render_title(dragon_D75, "Target Image",
#'            title_color = "white", title_size = 12))
#' plot_image(render_title(updated_source, "Corrected OG Image",
#'            title_color = "white", title_size = 12))
render_color_correction = function(
	image,
	matrix = diag(3),
	filename = NULL,
	preview = FALSE
) {
	stopifnot(is.matrix(matrix), all(dim(matrix) == c(3, 3)), is.numeric(matrix))
	src = ray_read_image(
		image,
		convert_to_array = TRUE,
		normalize = FALSE,
		reset_camera_settings = TRUE
	)
	imagetype = attr(src, "filetype")
	img_source_linear = attr(src, "source_linear")
	colorspace = attr(src, "colorspace")
	white_current = attr(src, "white_current")

	if (!isTRUE(attr(src, "source_linear"))) {
		warning(
			"render_color_correction(): input is not linear; convert with render_gamma_linear(..., TRUE) first."
		)
	}
	d = dim(src)
	if (length(d) != 3L) {
		return(src)
	}
	out = apply_color_matrix(src, matrix)
	out[,, 1:3][out[,, 1:3] < 0] = 0
	out = ray_read_image(
		out,
		filetype = imagetype,
		source_linear = img_source_linear,
		assume_colorspace = colorspace,
		assume_white = white_current
	)
	handle_image_output(out, filename = filename, preview = preview)
}

.libopenexr_available = function() {
  requireNamespace("libopenexr", quietly = TRUE)
}

#'@title Write Image
#'
#'@description Takes an RGB array/filename and writes it to file.
#'
#'@param image 3-layer RGB/4-layer RGBA array, `rayimg` class, or filename of an image.
#'@param filename File to write to, with filetype determined by extension. Filetype can be
#'`PNG`, `JPEG`, `TIFF`, `EXR`, or `DNG`.
#'@param clamp Default `FALSE`, automatically determined. Whether to clamp the image to 0-1. If the file extension is `PNG` of `JPEG`,
#'this is forced to `TRUE`.
#'@param write_linear Default `NA`, automatically determined. By default, images will be gamma corrected (write_linear = FALSE) for
#' all file types except `EXR` and `DNG` (which are linear formats), unless otherwise specified.
#'@param ... Arguments to pass to either `jpeg::writeJPEG`, `png::writePNG`, `libopenexr::write_exr`, or `tiff::writeTIFF`.
#'@return A `rayimg` RGBA array.
#'@import grDevices
#'@export
#'@examplesIf interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")
#'#Write as a png
#'tmparr = tempfile(fileext=".png")
#'ray_read_image(dragon) |>
#'  ray_write_image(tmparr)
#'ray_read_image(tmparr) |>
#'  plot_image()
#'#Write as a JPEG (passing quality arguments via ...)
#'tmparr = tempfile(fileext=".jpg")
#'ray_read_image(dragon) |>
#'  ray_write_image(tmparr, quality = 0.2)
#'ray_read_image(tmparr) |>
#'  plot_image()
#'#Write as a tiff
#'tmparr = tempfile(fileext=".tiff")
#'ray_read_image(dragon) |>
#'  ray_write_image(tmparr)
#'ray_read_image(tmparr) |>
#'   plot_image()
ray_write_image = function(
  image,
  filename,
  clamp = FALSE,
  write_linear = NA,
  ...
) {
  if (missing(filename)) {
    stop("`filename` must be specified.")
  }
  image = ray_read_image(image) # rayimg RGBA, linear RGB
  fileext = tolower(tools::file_ext(filename))
  if (!fileext %in% c("png", "jpeg", "jpg", "tiff", "exr", "dng")) {
    stop(sprintf(
      "File extension (%s) must be one of `png`, `jpeg`, `jpg`, `exr`, `tiff`, or `dng`",
      fileext
    ))
  }

  if (fileext == "dng") {
    meta = attr(image, "dng")
    if (is.null(meta)) {
      meta = list()
    }
    write_dng_cpp(filename, image, meta = meta)
    return(invisible(image))
  }

  # Auto conversion to sRGB/D65 primaries for LDR outputs
  is_ldr = fileext %in% c("png", "jpeg", "jpg", "tiff")
  if (is_ldr && length(dim(image)) == 3 && dim(image)[3] >= 3) {
    cs_from = attr(image, "colorspace")
    white_curr = attr(image, "white_current")
    if (
      !is.null(cs_from) &&
        (!identical(cs_from$name, "sRGB") ||
          any(abs(white_curr - CS_SRGB$white_xyz) > 1e-8))
    ) {
      xyz = apply_color_matrix(image, cs_from$rgb_to_xyz)
      if (any(abs(white_curr - CS_SRGB$white_xyz) > 1e-8)) {
        CAT = compute_cat_bradford(white_curr, CS_SRGB$white_xyz)
        xyz = apply_color_matrix(xyz, CAT)
      }
      image = apply_color_matrix(xyz, CS_SRGB$xyz_to_rgb)
      image[,, 1:3][image[,, 1:3] < 0] = 0
      # Update attributes to reflect new working space at write boundary
      attr(image, "colorspace") = CS_SRGB
      attr(image, "white_current") = CS_SRGB$white_xyz
    }
  }

  # Clamp for integer formats
  if (clamp || fileext %in% c("png", "jpeg", "jpg", "tiff")) {
    image = render_clamp(image)
  }

  if (is.na(write_linear)) {
    # LDR formats default to gamma-encoded; EXR stays linear
    write_linear = !(fileext %in% c("png", "jpeg", "jpg", "tiff"))
  }

  is_matrix = length(dim(image)) == 2
  if (is_matrix) {
    if (!write_linear) image = to_srgb(image)
  } else {
    max_dim = dim(image)[3L]
    has_alpha = max_dim %in% c(2L, 4L)
    col_dims = if (has_alpha) seq_len(max_dim - 1L) else seq_len(max_dim)

    if (has_alpha) {
      alpha_channel = image[,, max_dim]
      alpha_channel[alpha_channel > 1] = 1
      alpha_channel[alpha_channel < 0] = 0
      image[,, max_dim] = alpha_channel
    }
    if (!write_linear) {
      image[,, col_dims] = to_srgb(image[,, col_dims]) # OETF only on RGB
    }
  }

  if (fileext %in% c("jpeg", "jpg")) {
    jpeg::writeJPEG(image, target = filename, ...)
  } else if (fileext == "png") {
    png::writePNG(image, target = filename, ...)
  } else if (fileext == "exr") {
    if (!.libopenexr_available()) {
      stop(
        "The 'libopenexr' package is required to write EXR files.",
        call. = FALSE
      )
    }
    rgba = array(1, dim = c(dim(image)[1:2], 4))
    if (is_matrix) {
      rgba[,, 1:3] = image
    } else if (dim(image)[3] == 1) {
      rgba[,, 1:3] = image[,, 1]
    } else if (dim(image)[3] == 2) {
      rgba[,, 1] = image[,, 1]
      rgba[,, 2] = image[,, 1]
      rgba[,, 3] = image[,, 1]
      rgba[,, 4] = image[,, 2]
    } else if (dim(image)[3] == 3) {
      rgba[,, 1:3] = image[,, 1:3]
    } else if (dim(image)[3] == 4) {
      rgba = image
    } else {
      stop(
        "EXR output requires a matrix or an array with 1, 2, 3, or 4 channels.",
        call. = FALSE
      )
    }
    libopenexr::write_exr(
      filename,
      r = rgba[,, 1],
      g = rgba[,, 2],
      b = rgba[,, 3],
      a = rgba[,, 4],
      ...
    )
  } else {
    tiff::writeTIFF(image, where = filename, ...)
  }

  # return the (possibly converted/encoded) rayimg invisibly
  invisible(image)
}

test_that("ray_write_image writes EXR when libopenexr is available", {
  skip_on_cran()
  skip_if_not_installed("libopenexr")

  rgb = array(0.25, dim = c(2, 2, 3))
  rgb[,, 1] = 0.1
  rgb[,, 2] = 0.4
  rgb[,, 3] = 0.8

  rgba = array(0.5, dim = c(2, 2, 4))
  rgba[,, 4] = matrix(c(0, 0.25, 0.75, 1), nrow = 2)

  for (image in list(rgb, rgba)) {
    tmp = tempfile(fileext = ".exr")
    on.exit(unlink(tmp), add = TRUE)

    expect_no_error(ray_write_image(image, tmp))
    expect_true(file.exists(tmp))
    expect_gt(file.info(tmp)$size, 0)
  }
})

test_that("ray_write_image errors clearly when libopenexr is unavailable", {
  skip_on_cran()
  testthat::local_mocked_bindings(
    .libopenexr_available = function() FALSE,
    .package = "rayimage"
  )

  tmp = tempfile(fileext = ".exr")
  expect_error(
    ray_write_image(array(0.5, dim = c(2, 2, 3)), tmp),
    "required to write EXR files"
  )
  expect_false(file.exists(tmp))
})

#' @title Render Tonemap
#'
#' @description Reinhard / Uncharted(Hable) / HBD operators in linear
#'
#' @param image Default `NULL`. 3/4-channel array, `rayimg`, or filename.
#' @param method Default `raw`. One of `"raw"`, `"reinhard"`, `"uncharted"`, `"hbd"`.
#' Chooses the global tone-mapping curve. `"raw"` leaves the image unchanged.
#' `"reinhard"` applies a classic shoulder that gently compresses highlights.
#' `"uncharted"` (Hable filmic) gives a film-style toe/shoulder and keeps mid-tones contrasty.
#' `"hbd"` is a fast filmic-like curve with a soft toe/shoulder.
#' All methods operate on **linear RGB** and return **display-referred linear** (no OETF).
#' @param exposure_bias Default `2`. Only for `"uncharted"`.
#' Scalar applied to the linear RGB channels **before** the Uncharted/Hable curve
#' (i.e., a pre-curve exposure gain). Values >1 brighten; <1 darken.
#' `exposure_bias=2` is about +1 stop; `0.5` is about -1 stop.
#' @param W Default `11.2`. White (linear scene value) that maps to 1.0 **after**
#' the Uncharted/Hable/Reinhard curve. Acts like a “white point” for the shoulder: larger values
#' preserve more highlight detail (later roll-off), smaller values roll off earlier and harder.
#' @param filename Default `NULL`. Output path.
#' @param preview Default `FALSE`. Show result.
#' @return A `rayimg` RGBA array.
#' @export
#' @examples
#' # Plot unchanged image
#' render_tonemap(dragon, preview = TRUE)
#' # Plot Reinhard tonemapped image
#' render_tonemap(dragon, method = "reinhard", preview = TRUE)
#' # Plot Uncharted/Hable tonemapped image
#' render_tonemap(dragon, method = "uncharted",preview = TRUE)
#' # Plot Uncharted/Hable tonemapped image, white point adjusted
#' render_tonemap(dragon, method = "uncharted", W=11.4, preview = TRUE)
#' # Plot Uncharted/Hable tonemapped image, exposure adjusted
#' render_tonemap(dragon, method = "uncharted", exposure_bias = 1,preview = TRUE)
#' # Plot Hejl-Burgess-Dawson tonemapped image, exposure adjusted
#' render_tonemap(dragon, method = "hbd",preview = TRUE)
render_tonemap = function(
  image,
  method = c("raw", "reinhard", "uncharted", "hbd"),
  exposure_bias = 2.0,
  W = 11.2,
  filename = NULL,
  preview = FALSE
) {
  method = match.arg(method)
  src = ray_read_image(
    image,
    convert_to_array = TRUE,
    reset_camera_settings = TRUE
  )
  imagetype = attr(src, "filetype")
  img_source_linear = attr(src, "source_linear")
  colorspace = attr(src, "colorspace")
  white_current = attr(src, "white_current")

  colorspace_luminance_row = attr(src, "colorspace")$rgb_to_xyz[2, ]
  d = dim(src)
  if (length(d) != 3) {
    return(src)
  }

  rgb = src[,, 1:3]

  if (method == "reinhard") {
    eps = 1e-6
    lw = colorspace_luminance_row

    L = lw[1] * rgb[,, 1] + lw[2] * rgb[,, 2] + lw[3] * rgb[,, 3]

    a = 0.18
    L_log_avg = exp(mean(log(pmax(L, eps))))

    Lm = (a / L_log_avg) * L
    Lwhite = W #White point
    Ld = (Lm * (1 + (Lm / (Lwhite * Lwhite)))) / (1 + Lm)

    scale = Ld / pmax(L, eps)

    rgb[,, 1] = rgb[,, 1] * scale
    rgb[,, 2] = rgb[,, 2] * scale
    rgb[,, 3] = rgb[,, 3] * scale
  } else if (method == "uncharted") {
    A = 0.15
    B = 0.50
    C = 0.10
    D = 0.20
    E = 0.02
    F = 0.30
    eps = 1e-6

    lw = colorspace_luminance_row
    L = lw[1] * rgb[,, 1] + lw[2] * rgb[,, 2] + lw[3] * rgb[,, 3]

    # Key normalization
    a = 0.18
    L_log_avg = exp(mean(log(pmax(L, eps))))
    pre_key = a / L_log_avg

    hable = function(x) {
      ((x * (A * x + C * B) + D * E) / (x * (A * x + B) + D * F) - E / F)
    }

    e = exposure_bias
    whiteScale = 1 / hable(e * W)

    Lm = e * pre_key * L
    Ld = hable(Lm) * whiteScale
    scale = Ld / pmax(L, eps)

    rgb[,, 1] = rgb[,, 1] * scale
    rgb[,, 2] = rgb[,, 2] * scale
    rgb[,, 3] = rgb[,, 3] * scale
  } else if (method == "hbd") {
    for (k in 1:3) {
      x = pmax(rgb[,, k] - 0.004, 0)
      rgb[,, k] = (x * (6.2 * x + 0.5)) / (x * (6.2 * x + 1.7) + 0.06)
    }
  } else {
    # Nothing
  }

  out = src
  out[,, 1:3] = rgb

  out = ray_read_image(
    out,
    filetype = imagetype,
    source_linear = img_source_linear,
    assume_colorspace = colorspace,
    assume_white = white_current
  )
  handle_image_output(out, filename = filename, preview = preview)
}

#'@title Generate 2D Disk
#'
#'@description Generates a 2D disk with a gradual falloff.
#'
#'Disk generated using the following formula:
#'
#'\ifelse{html}{\out{(-22.35 * cos(1.68 * r<sup>2</sup>) + 85.91 * sin(1.68 * r<sup>2</sup>) ) * exp(-4.89 * r<sup>2</sup>) +
#'(35.91 * cos(4.99 * r<sup>2</sup>) - 28.87 * sin(4.99 * r<sup>2</sup>)) * exp(-4.71 * r<sup>2</sup>) +
#'(-13.21 * cos(8.24 * r<sup>2</sup>) - 1.57 * sin(8.24 * r<sup>2</sup>)) * exp(-4.05 * r<sup>2</sup>) +
#'(0.50 * cos(11.90 * r<sup>2</sup>) + 1.81 * sin(11.90 * r<sup>2</sup>)) * exp(-2.92 * r<sup>2</sup>) +
#'(0.13 * cos(16.11 * r<sup>2</sup>) - 0.01 * sin(16.11 * r<sup>2</sup>)) * exp(-1.51 * r<sup>2</sup>)}}{(-22.35 cos(1.68 r^2) + 85.91 sin(1.68 r^2) ) exp(-4.89 r^2) +
#'(35.91 cos(4.99 r^2) - 28.87 sin(4.99 r^2)) exp(-4.71 r^2) +
#'(-13.21 cos(8.24 r^2) - 1.57 sin(8.24 r^2)) exp(-4.05 r^2) +
#'(0.50 cos(11.90 r^2) + 1.81 sin(11.90 r^2)) exp(-2.92 r^2) +
#'(0.13 cos(16.11 r^2) - 0.01 sin(16.11 r^2)) exp(-1.51 r^2)}
#'
#'The origin of the coordinate system is the center of the matrix.
#'
#'@param dim Default `c(11, 11)`. The dimensions of the matrix.
#'@param radius Default `1`. Radius of the disk, compared to the dimensions. Should be less than one.
#'@param rescale_unity Default `FALSE`. If `TRUE`, this will rescale the max value to one. Useful
#'if wanting to plot the distribution with [plot_image()].
#'@export
#'@examplesIf interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")
#'image(generate_2d_disk(101), asp=1)
generate_2d_disk = function(
  dim = c(11, 11),
  radius = 1,
  rescale_unity = FALSE
) {
  dim = rev(dim)
  mindim = min(dim)
  add_offset_x = FALSE
  add_offset_y = FALSE
  if (length(dim) == 2) {
    if (dim[2] - dim[1] > 0) {
      if (abs(dim[2] - dim[1]) %% 2 != 0) {
        add_offset_x = TRUE
      }
    } else if (dim[2] - dim[1] < 0) {
      if (abs(dim[2] - dim[1]) %% 2 != 0) {
        add_offset_y = TRUE
      }
    }
  }
  disk = generate_disk((1 / radius) * 1.18, mindim, add_offset_x, add_offset_y)
  if (length(dim) == 2) {
    disk = pad_to_fit(dim, disk)
  }
  disk = (disk - min(disk)) / (max(disk) - min(disk))
  disk = disk / sum(disk)
  if (rescale_unity) {
    return(disk / max(disk))
  }
  return(disk)
}

#' @title Render Vibrance
#'
#' @description
#' Adjusts image vibrance (adaptive saturation), similar in spirit to the iPhone "Vibrance" control:
#' it boosts muted colors more than already-saturated colors, and reduces saturation more strongly
#' in already-saturated regions when negative. Works on HDR arrays as well (values can exceed 1).
#'
#' @param image Image filename, 3-layer RGB array, 4-layer RGBA array, or matrix. If a filename,
#' it will be read via `ray_read_image()`.
#' @param vibrance Default `0`. Numeric scalar in -1..1 (values are clamped). Positive values
#' increase vibrance primarily in low-saturation regions; negative values decrease vibrance primarily
#' in high-saturation regions.
#' @param protect_luminance Default `0.25`. Numeric scalar in 0..1. Reduces the vibrance effect
#' in deep shadows and bright highlights to avoid harsh color shifts. Set to `0` to disable.
#' @param filename Default `NULL`. If not `NULL`, the processed image will be written to this file.
#' @param preview Default `FALSE`. Whether to plot the processed image.
#'
#' @return 4-layer RGBA array (or matrix if passed a matrix and `ray_read_image()` is not used).
#' @export
#'
#'@examplesIf interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")
#'   dragon |>
#'     render_vibrance(vibrance = 0.4, preview = TRUE)
render_vibrance = function(
  image,
  vibrance = 0,
  protect_luminance = 0.25,
  filename = NULL,
  preview = FALSE
) {
  if (length(vibrance) != 1 || !is.numeric(vibrance) || !is.finite(vibrance)) {
    stop("`vibrance` must be a single finite numeric value.")
  }
  if (
    length(protect_luminance) != 1 ||
      !is.numeric(protect_luminance) ||
      !is.finite(protect_luminance)
  ) {
    stop("`protect_luminance` must be a single finite numeric value.")
  }
  if (protect_luminance < 0 || protect_luminance > 1) {
    stop("`protect_luminance` must be in [0,1].")
  }

  vibrance = max(min(vibrance, 1), -1)

  temp_image = ray_read_image(image, reset_camera_settings = TRUE) # returns RGBA array in rayimage

  nr = dim(temp_image)[1]
  nc = dim(temp_image)[2]

  r = as.vector(temp_image[,, 1])
  g = as.vector(temp_image[,, 2])
  b = as.vector(temp_image[,, 3])

  na_mask = is.na(r) | is.na(g) | is.na(b)
  if (any(na_mask)) {
    r[na_mask] = 0
    g[na_mask] = 0
    b[na_mask] = 0
  }

  # Avoid negative values breaking HSV math; keep HDR highs intact.
  r = pmax(r, 0)
  g = pmax(g, 0)
  b = pmax(b, 0)

  maxc = pmax(r, pmax(g, b))
  minc = pmin(r, pmin(g, b))
  delta = maxc - minc

  v = maxc
  s = rep(0, length(v))
  nz = maxc > 0
  s[nz] = delta[nz] / maxc[nz]

  h = rep(0, length(v))
  idx = delta > 0

  if (any(idx)) {
    idx_r = idx & (maxc == r)
    idx_g = idx & (maxc == g) & !idx_r
    idx_b = idx & (maxc == b) & !(idx_r | idx_g)

    if (any(idx_r)) {
      h[idx_r] = (g[idx_r] - b[idx_r]) / delta[idx_r]
      h[idx_r][h[idx_r] < 0] = h[idx_r][h[idx_r] < 0] + 6
    }
    if (any(idx_g)) {
      h[idx_g] = (b[idx_g] - r[idx_g]) / delta[idx_g] + 2
    }
    if (any(idx_b)) {
      h[idx_b] = (r[idx_b] - g[idx_b]) / delta[idx_b] + 4
    }
    h[idx] = h[idx] / 6
    h = h - floor(h) # wrap into [0,1)
  }

  if (protect_luminance > 0) {
    v01 = pmin(v, 1)
    w = 1 - protect_luminance * abs(2 * v01 - 1)
    w = pmax(w, 0)
  } else {
    w = 1
  }

  if (vibrance >= 0) {
    s2 = s + vibrance * (1 - s) * w
  } else {
    s2 = s + vibrance * s * w
  }
  s2 = pmin(pmax(s2, 0), 1)

  # HSV -> RGB (vectorized), preserving HDR v
  r2 = v
  g2 = v
  b2 = v

  sat = s2
  idx_sat = sat > 0

  if (any(idx_sat)) {
    h6 = h * 6
    i = floor(h6)
    f = h6 - i

    p = v * (1 - sat)
    q = v * (1 - sat * f)
    t = v * (1 - sat * (1 - f))

    im = i %% 6

    k0 = idx_sat & (im == 0)
    k1 = idx_sat & (im == 1)
    k2 = idx_sat & (im == 2)
    k3 = idx_sat & (im == 3)
    k4 = idx_sat & (im == 4)
    k5 = idx_sat & (im == 5)

    if (any(k0)) {
      r2[k0] = v[k0]
      g2[k0] = t[k0]
      b2[k0] = p[k0]
    }
    if (any(k1)) {
      r2[k1] = q[k1]
      g2[k1] = v[k1]
      b2[k1] = p[k1]
    }
    if (any(k2)) {
      r2[k2] = p[k2]
      g2[k2] = v[k2]
      b2[k2] = t[k2]
    }
    if (any(k3)) {
      r2[k3] = p[k3]
      g2[k3] = q[k3]
      b2[k3] = v[k3]
    }
    if (any(k4)) {
      r2[k4] = t[k4]
      g2[k4] = p[k4]
      b2[k4] = v[k4]
    }
    if (any(k5)) {
      r2[k5] = v[k5]
      g2[k5] = p[k5]
      b2[k5] = q[k5]
    }
  }

  if (any(na_mask)) {
    r2[na_mask] = NA_real_
    g2[na_mask] = NA_real_
    b2[na_mask] = NA_real_
  }

  temp_image[,, 1] = matrix(r2, nrow = nr, ncol = nc)
  temp_image[,, 2] = matrix(g2, nrow = nr, ncol = nc)
  temp_image[,, 3] = matrix(b2, nrow = nr, ncol = nc)

  handle_image_output(temp_image, filename = filename, preview = preview)
}

#'@title Clamp Image
#'
#'@description Clamps an image to a user-specified range, ignoring the alpha channel.
#'
#'@param image 3-layer RGB/4-layer RGBA array, `rayimg` class, or filename of an image.
#'@param min_value Default `0`. Minimum value to clamp the RGB channels in the image to.
#'@param max_value Default `1`. Maximum value to clamp the RGB channels in the image to.
#'@param filename Default `NA`. Filename
#'@param preview Default `FALSE`. If `TRUE`, it will display the image in addition
#'to returning it.
#'@param ... Arguments to pass to either `jpeg::readJPEG`, `png::readPNG`, or `tiff::readTIFF`.
#'
#'@return A `rayimg` RGBA array.
#'@import grDevices
#'@export
#'@examplesIf interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")
#'#The image of the unchanged image
#'range(dragon)
#'#Clamp the maximum and minimum values to one and zero
#'render_clamp(dragon) |>
#'  range()
render_clamp = function(
  image,
  min_value = 0,
  max_value = 1,
  filename = NA,
  preview = FALSE,
  ...
) {
  image = ray_read_image(image, reset_camera_settings = TRUE) #Always output RGBA array
  imagetype = attr(image, "filetype")
  img_source_linear = attr(image, "source_linear")
  colorspace = attr(image, "colorspace")
  white_current = attr(image, "white_current")

  d = dim(image)
  alpha_idx = NA
  if (length(d) == 3) {
    if (d[3] == 4) {
      alpha_idx = 4
    }
    if (d[3] == 2) {
      alpha_idx = 2
    }
  }
  array_image = unclass(image)
  if (length(d) == 2) {
    array_image[image < min_value] = min_value
    array_image[image > max_value] = max_value
    array_image = array(array_image, dim = d)
  } else {
    if (is.na(alpha_idx)) {
      num_chans = 1:3
    } else {
      num_chans = seq_len(alpha_idx - 1)
    }
    array_image = unclass(image)
    for (chan in num_chans) {
      tmp_chan = array_image[,, chan]
      mat_d = dim(tmp_chan)
      tmp_chan[array_image[,, chan] < min_value] = min_value
      tmp_chan[array_image[,, chan] > max_value] = max_value
      array_image[,, chan] = array(tmp_chan, dim = mat_d)
    }
  }
  new_image = ray_read_image(
    array_image,
    filetype = imagetype,
    source_linear = img_source_linear,
    assume_colorspace = colorspace,
    assume_white = white_current
  )
  handle_image_output(new_image, filename = filename, preview = preview)
}

#' @title Render a Title on an Image
#'
#' @description Adds a title with optional styling and a title bar to an image.
#' The image can be previewed or saved to a file. Supports both the `grid`-based
#' method and (deprecated) `magick` package for rendering the title.
#'
#' @param image 3-layer RGB/4-layer RGBA array, `rayimg` class, or filename of an image.
#' @param title_text Default `""`. Text string to be added as the title to the image.
#' @param title_offset Default `c(15,15)`. Numeric vector specifying the horizontal
#' and vertical offset of the title text, relative to its anchor position.
#' @param title_lineheight Default `1`. Multiplier for the lineheight.
#' @param title_color Default `"black"`. String specifying the color of the title text.
#' @param title_size Default `30`. Numeric value specifying the font size of the title text.
#' @param title_font Default `"Arial"`. String specifying the font family for the title text.
#' Common options include `"sans"`, `"mono"`, `"serif"`, `"Times"`, `"Helvetica"`, etc.
#' @param title_style Default `"plain"`. String specifying the font style, such as
#' `"plain"`, `"italic"`, or `"bold"`.
#' @param title_bar_color Default `NULL`. Color of the optional title bar. If `NULL`, no bar is added.
#' @param title_bar_alpha Default `0.5`. Transparency level of the title bar. A value
#' between `0` (fully transparent) and `1` (fully opaque).
#' @param title_bar_width Default `NULL`. Numeric value for the height of the title bar in pixels.
#' If `NULL`, it is automatically calculated based on the text size and line breaks.
#' @param title_position Default `"northwest"`. String specifying the position of the title text.
#' Only used when `use_magick = TRUE`. Common options include `"northwest"`, `"center"`, `"south"`, etc.
#' @param title_just Default `"left"`. Horizontal alignment of the title text: `"left"`,
#' `"center"`, or `"right"`.
#' @param use_magick Default `FALSE`. Boolean indicating whether to use the `magick` package for
#' rendering titles. This option will be deprecated in future versions.
#' @param filename Default `NULL`. String specifying the file path to save the resulting image.
#' If `NULL` and `preview = FALSE`, the function returns the processed RGB array.
#' @param preview Default `FALSE`. Boolean indicating whether to display the image after processing.
#' If `TRUE`, the image is displayed but not saved or returned.
#'
#' @return A 3-layer RGB array of the processed image if `filename = NULL` and `preview = FALSE`.
#' Otherwise, writes the image to the specified file or displays it if `preview = TRUE`.
#'
#' @note The `use_magick` parameter and all functionality tied to the `magick` package are
#' planned for deprecation. It is recommended to use the `grid`-based method for
#' future compatibility.
#'
#' @import grDevices grid
#' @export
#'
#'@examplesIf interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")
#'#Plot the dragon
#'render_title(dragon, preview = TRUE, title_text = "Dragon", title_size=20)
#'#That's hard to see--let's add a title bar:
#'render_title(dragon, preview = TRUE, title_text = "Dragon", title_size=20,
#'          title_bar_color="white")
#'#Change the width of the bar:
#'render_title(dragon, preview = TRUE, title_text = "Dragon", title_size=20,
#'          title_bar_color="white", title_offset = c(8,8))
#'#The width of the bar will also automatically adjust for newlines:
#'render_title(dragon, preview = TRUE, title_text = "Dragon\n(Blue)", title_size=20,
#'          title_bar_color="white")
#'#Change the color and title color:
#'render_title(dragon, preview = TRUE, title_text = "Dragon", title_size=20,
#'          title_bar_color="red", title_color = "white")
#'#Change the transparency:
#'render_title(dragon, preview = TRUE, title_text = "Dragon",
#'             title_size=20, title_bar_alpha = 0.8,
#'             title_bar_color="red", title_color = "white")
#' #Read directly from a file
#' temp_image = tempfile(fileext = ".png")
#' ray_write_image(dragon, temp_image)
#' render_title(temp_image, preview = TRUE, title_text = "Dragon",
#'              title_size=20, title_bar_alpha = 0.8,
#'              title_bar_color="red", title_color = "white")
render_title = function(
  image,
  title_text = "",
  title_size = 30,
  title_offset = rep(title_size / 2, 2),
  title_lineheight = 1,
  title_color = "black",
  title_font = "Arial",
  title_style = "plain",
  title_bar_color = NA,
  title_bar_alpha = 0.5,
  title_bar_width = NULL,
  title_position = NA,
  title_just = "left",
  use_magick = FALSE,
  filename = NULL,
  preview = FALSE
) {
  #Always output RGBA array, need to ensure alpha is preserved
  image = ray_read_image(
    image,
    convert_to_array = TRUE,
    reset_camera_settings = TRUE
  )

  temp = tempfile(fileext = ".png")
  ray_write_image(
    image = ray_read_image(array(0, dim = dim(image))),
    filename = temp
  )

  if (use_magick) {
    if (title_style == "plain") {
      title_style = "normal"
    }
    if (is.na(title_bar_color)) {
      title_bar_color = NULL
    }
    if (is.na(title_position)) {
      title_position = "northwest"
    }
    dimensions = dim(image)
    if (!("magick" %in% rownames(utils::installed.packages()))) {
      stop("`magick` package required for adding title")
    }

    if (!is.null(title_bar_color)) {
      title_bar_color = col2rgb(title_bar_color) / 255
      title_bar = array(0, c(dimensions[1], dimensions[2], 4))
      if (is.null(title_bar_width)) {
        #Detect newlines and adjust title bar width
        char_vector = unlist(strsplit(title_text, ""))
        newline_indices = which(char_vector == "\n")
        newline_indices = newline_indices[
          newline_indices != length(char_vector)
        ]
        newlines = length(newline_indices)
        title_bar_width = 2 * title_offset[2] + title_size * (1 + newlines)
      }
      if (title_bar_width > dimensions[1]) {
        message(paste0(
          c(
            "Input title_bar_width (",
            title_bar_width,
            ") greater than image height (",
            dimensions[1],
            "), reducing size."
          ),
          collapse = ""
        ))
        title_bar_width = dimensions[1]
      }
      if (title_position %in% c("northwest", "north", "northeast")) {
        title_bar[1:title_bar_width, , 1] = title_bar_color[1]
        title_bar[1:title_bar_width, , 2] = title_bar_color[2]
        title_bar[1:title_bar_width, , 3] = title_bar_color[3]
        title_bar[1:title_bar_width, , 4] = title_bar_alpha
      } else if (title_position %in% c("southwest", "south", "southeast")) {
        title_bar[
          (nrow(title_bar) - title_bar_width):nrow(title_bar),
          ,
          1
        ] = title_bar_color[1]
        title_bar[
          (nrow(title_bar) - title_bar_width):nrow(title_bar),
          ,
          2
        ] = title_bar_color[2]
        title_bar[
          (nrow(title_bar) - title_bar_width):nrow(title_bar),
          ,
          3
        ] = title_bar_color[3]
        title_bar[
          (nrow(title_bar) - title_bar_width):nrow(title_bar),
          ,
          4
        ] = title_bar_alpha
      } else {
        title_bar[
          (nrow(title_bar) / 2 - title_bar_width / 2):(nrow(title_bar) /
            2 +
            title_bar_width / 2),
          ,
          1
        ] = title_bar_color[1]
        title_bar[
          (nrow(title_bar) / 2 - title_bar_width / 2):(nrow(title_bar) /
            2 +
            title_bar_width / 2),
          ,
          2
        ] = title_bar_color[2]
        title_bar[
          (nrow(title_bar) / 2 - title_bar_width / 2):(nrow(title_bar) /
            2 +
            title_bar_width / 2),
          ,
          3
        ] = title_bar_color[3]
        title_bar[
          (nrow(title_bar) / 2 - title_bar_width / 2):(nrow(title_bar) /
            2 +
            title_bar_width / 2),
          ,
          4
        ] = title_bar_alpha
      }
      title_bar_temp = paste0(tempfile(), ".png")
      png::writePNG(title_bar, title_bar_temp)
      magick::image_read(temp) |>
        magick::image_composite(magick::image_read(title_bar_temp), ) |>
        magick::image_write(path = temp, format = "png")
    }
    magick::image_read(temp) |>
      magick::image_annotate(
        title_text,
        location = paste0("+", title_offset[1], "+", title_offset[2]),
        size = title_size,
        color = title_color,
        style = title_style,
        font = title_font,
        gravity = title_position
      ) |>
      magick::image_write(path = temp, format = "png")
    new_image = render_image_overlay(image, temp)
  } else {
    draw_title_card = function(
      image,
      title,
      padding_x = 10,
      padding_y = 10,
      gp_text = grid::gpar(col = "white", fontsize = title_size),
      bg_color = "black",
      bg_alpha = 0.65,
      title_just = c("left", "top")
    ) {
      close_device_safely = function(dev_id) {
        try(grid::upViewport(0), silent = TRUE)
        devs = grDevices::dev.list()
        if (!is.null(devs) && dev_id %in% devs) {
          grDevices::dev.off(dev_id)
        }
      }

      png_device = linear_png_device()
      png_device(
        temp,
        width = ncol(image),
        height = nrow(image),
        pointsize = 12,
        family = title_font,
        units = "px",
        res = 72
      )
      dev_id = grDevices::dev.cur()
      pushed = 0L
      on.exit(
        {
          if (pushed > 0L) {
            try(grid::popViewport(pushed), silent = TRUE)
          }
          close_device_safely(dev_id)
        },
        add = TRUE
      )

      grid::grid.newpage()

      prepared = prepare_native_raster(image)
      nr = prepared$native_raster
      image_dim = prepared$display_dim

      image_width = image_dim[2]
      image_height = image_dim[1]

      image_vp = grid::viewport(
        name = "image",
        xscale = c(0, image_width),
        yscale = c(image_height, 0),
        gp = gp_text
      )
      grid::pushViewport(image_vp)
      pushed = pushed + 1L

      image_grob = grid::rasterGrob(
        nr,
        interpolate = FALSE,
        width = grid::unit(1, "npc"),
        height = grid::unit(1, "npc")
      )
      grid::grid.draw(image_grob)

      title_hjust = title_just[1]
      if (!title_hjust %in% c("left", "center", "right")) {
        stop(
          'Invalid title_just value for horizontal alignment. Must be "left", "center", or "right".'
        )
      }

      # Available width for text depends on horizontal justification.
      # For left/right aligned text, `padding_x` is only applied on one side.
      available_width = if (identical(title_hjust, "center")) {
        image_width - padding_x * 2
      } else {
        image_width - padding_x
      }
      available_width = max(0, available_width)

      # Wrap text into lines based on available width
      wrap_text = function(text, available_width) {
        # Split text into segments by newline characters
        segments = unlist(strsplit(text, "\n"))
        all_lines = character(0)

        for (segment in segments) {
          words = strsplit(segment, "\\s+")[[1]]
          if (length(words) == 0) {
            all_lines = c(all_lines, "") # Add an empty line for consecutive newlines
            next
          }

          lines = character(0)
          current_line = words[1]

          for (word in words[-1]) {
            test_line = paste(current_line, word)
            test_width = grid::convertWidth(
              grid::stringWidth(test_line),
              "native",
              valueOnly = TRUE
            )
            if (test_width > available_width) {
              lines = c(lines, current_line)
              current_line = word
            } else {
              current_line = test_line
            }
          }

          lines = c(lines, current_line)
          all_lines = c(all_lines, lines)
        }

        return(all_lines)
      }

      # Wrap the title text
      lines = wrap_text(title, available_width)
      lines_with_newlines = paste0(lines, collapse = "\n")

      text_size = systemfonts::shape_string(
        lines_with_newlines,
        size = title_size,
        vjust = 0,
        res = 72,
        lineheight = title_lineheight,
        family = title_font
      )[["metrics"]]
      # Position adjustments
      if (title_hjust == "left") {
        x = padding_x
        just = c("left", "top")
      } else if (title_hjust == "center") {
        x = image_width / 2
        just = c("center", "top")
      } else if (title_hjust == "right") {
        x = image_width - padding_x
        just = c("right", "top")
      }
      height_with_padding = text_size$height -
        text_size$top_bearing -
        text_size$bottom_bearing -
        text_size$top_border +
        padding_y * 2
      # Draw background rectangle
      grid::grid.rect(
        x = 0,
        y = 0,
        width = image_width,
        height = height_with_padding,
        just = c("left", "bottom"),
        default.units = "native",
        gp = grid::gpar(
          fill = grDevices::adjustcolor(bg_color, alpha.f = bg_alpha),
          col = NA
        )
      )
      grid::grid.text(
        label = lines_with_newlines,
        x = x,
        y = padding_y,
        default.units = "native",
        just = just,
        gp = grid::gpar(
          col = title_color,
          fontsize = title_size,
          lineheight = title_lineheight,
          cex = 1,
          fontfamily = title_font
        )
      )
      grid::popViewport(pushed)
      pushed = 0L
    }
    if (!is.na(title_position)) {
      warning("Title position is ignored when not using {magick}")
    }
    if (is.null(title_bar_color)) {
      title_bar_color = NA
    }
    if (is.null(title_bar_alpha)) {
      title_bar_alpha = NA
    }
    draw_title_card(
      image,
      title = title_text,
      padding_x = title_offset[1],
      padding_y = title_offset[2],
      gp_text = grid::gpar(
        col = title_color,
        fontsize = title_size,
        fontfamily = title_font,
        lineheight = title_lineheight,
        fontface = title_style,
        fill = NA
      ),
      bg_color = title_bar_color,
      bg_alpha = title_bar_alpha,
      title_just = title_just
    )
    new_image = ray_read_image(temp)
  }
  handle_image_output(new_image, filename = filename, preview = preview)
}

#'@title Plot Image Grid
#'
#'@description Displays the image in the current device.
#'
#'@param input_list List of 3-layer RGB/4-layer RGBA array, `rayimg` class, or image filenames.
#'@param dim Default `c(1,1)`. Width by height of output grid.
#'@param asp Default `1`. Aspect ratio of the pixels(s). For example, an aspect ratio of `4/3` will
#'slightly widen the image. This can also be a vector the same length of
#'`input_list` to specify an aspect ratio for each image in the grid.
#'@param draw_grid Default `FALSE`. If `TRUE`, this will draw a grid in the background to help
#'disambiguate the actual image from the device (helpful if the image background is the same as the
#'device's background).
#'@param gp A `grid::gpar()` object to include for the grid viewport displaying the image.
#'@export
#'@examplesIf interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")
#'#Plot the dragon array
#'plot_image_grid(list(dragon, 1-dragon), dim = c(1,2))
#'plot_image_grid(list(dragon, 1-dragon), dim = c(2,1))
#'plot_image_grid(list(dragon, NULL, 1-dragon), dim = c(2,2), asp = c(2,1,1/2))
#'plot_image_grid(list(dragon, NULL, NULL, dragon), dim = c(2,2), asp = c(2,1,1,1/2))
#'#Plot alongside the depth matrix
#'dragon_depth_reoriented = render_reorient(dragondepth,
#'                                          transpose = TRUE,
#'                                          flipx = TRUE)/2000
#'plot_image_grid(list(dragondepth/2000, dragon, dragon, dragondepth/2000),
#'                dim = c(2,2))
plot_image_grid = function(
  input_list,
  dim = c(1, 1),
  asp = 1,
  draw_grid = FALSE,
  gp = grid::gpar()
) {
  if (length(dim) != 2) stop("length of `dim` argument must equal 2")
  if (!inherits(input_list, "list"))
    stop("`input_list` must be a list of image arrays")

  if (length(asp) == 1) {
    asp = rep(asp, length(input_list))
  } else if (length(asp) != length(input_list)) {
    stop(sprintf(
      "length of `asp` (%i) not equal to length of `input_list` (%i) or length-1",
      length(asp),
      length(input_list)
    ))
  }

  if (length(input_list) > dim[1] * dim[2]) {
    warning(sprintf(
      "`input_list` of length %i, but grid only has %ix%i=%i slots: truncating list of images",
      length(input_list),
      dim[1],
      dim[2],
      dim[1] * dim[2]
    ))
    input_list = input_list[seq_len(dim[1] * dim[2])]
  }

  grid::grid.newpage()

  # optional background grid — push, draw, then pop
  if (draw_grid) {
    grid::pushViewport(
      grid::viewport(layout = grid::grid.layout(1, 1), name = "ray_bg_grid")
    )
    grid_density = 0.01
    for (i in seq(-2, 2, by = grid_density)) {
      grid::grid.lines(
        x = c(0, 1),
        y = c(i, i + 1),
        default.units = "npc",
        gp = grid::gpar(col = "grey")
      )
      grid::grid.lines(
        x = c(0, 1),
        y = c(i, i - 1),
        default.units = "npc",
        gp = grid::gpar(col = "grey")
      )
    }
    grid::upViewport() # pop "ray_bg_grid"
  }

  # parent layout viewport
  grid::pushViewport(grid::viewport(
    name = "ray_image_grid",
    layout = grid::grid.layout(nrow = dim[1], ncol = dim[2]),
    gp = gp
  ))
  on.exit(
    {
      # ensure we fully unwind no matter what
      try(grid::upViewport(0), silent = TRUE)
    },
    add = TRUE
  )

  layout_idx = data.frame(
    y = rep(seq_len(dim[1]), each = dim[2]),
    x = rep(seq_len(dim[2]), dim[1])
  )

  for (i in seq_len(length(input_list))) {
    # enter the parent (in case caller changed current vp)
    grid::seekViewport("ray_image_grid")
    # push cell viewport, draw, then pop it
    grid::pushViewport(grid::viewport(
      layout.pos.row = layout_idx$y[i],
      layout.pos.col = layout_idx$x[i],
      name = sprintf("ray_cell_%i_%i", layout_idx$x[i], layout_idx$y[i]),
      gp = gp
    ))
    if (!is.null(input_list[[i]])) {
      rayimage::plot_image(input_list[[i]], asp = asp[i], new_page = FALSE)
    }
    grid::upViewport() # pop cell
  }

  grid::upViewport() # pop "ray_image_grid"
  invisible(NULL)
}

rgb_to_hex = function(x) {
  if (is.null(dim(x))) {
    x = matrix(x, ncol = length(x), byrow = TRUE)
  }
  if (!(ncol(x) %in% c(3, 4))) {
    stop("rgb_to_hex(): expected 3 or 4 columns (R,G,B[,A]).")
  }
  x = pmin(pmax(x, 0), 1)
  ints = round(x * 255)
  paste0(
    "#",
    apply(ints, 1, function(row) {
      paste(sprintf("%02X", row), collapse = "")
    })
  )
}

#' Convert Color
#'
#' @param color The color to convert. Can be either an R color string, or a
#' numeric scalar, RGB vector, RGBA vector, or RGB/RGBA matrix with values
#' between `0` and `1`.
#' @param as_hex Default `FALSE`. If `TRUE`, return hex colors.
#' @param single Default `FALSE`. If `TRUE`, require exactly one color.
#'
#' @return RGB or RGBA color vector, or a color matrix for multiple colors.
#' @keywords internal
#'
#' @examples
#' #none
convert_color = function(color, as_hex = FALSE, single = FALSE) {
  if (is.null(color) || length(color) == 0) {
    stop("invalid color")
  }
  if (is.character(color)) {
    color = t(grDevices::col2rgb(color, alpha = TRUE)) / 255
  } else {
    if (!is.numeric(color)) {
      stop("invalid color")
    }
    color_dim = dim(color)
    if (is.null(color_dim)) {
      if (length(color) == 1) {
        color = rep(color, 3)
      }
      if (
        !(length(color) %in%
          c(3, 4) ||
          length(color) %% 3 == 0 ||
          length(color) %% 4 == 0)
      ) {
        stop("invalid color")
      }
      color_channels = if (length(color) %% 3 == 0) 3 else 4
      color = matrix(
        color,
        ncol = color_channels,
        byrow = TRUE
      )
    } else {
      if (length(color_dim) != 2 || !(color_dim[2] %in% c(3, 4))) {
        stop("invalid color")
      }
    }
  }
  if (!all(is.finite(color))) {
    stop("invalid color")
  }
  if (!all(color >= 0 & color <= 1)) {
    stop("invalid color")
  }
  if (isTRUE(single) && nrow(color) != 1L) {
    stop("invalid color")
  }
  if (as_hex) {
    return(rgb_to_hex(color))
  }
  # return vector for single row, matrix otherwise
  if (nrow(color) == 1L) as.numeric(color[1, ]) else color
}

#'@title Stack Images
#'
#'@description Stacks a list of images vertically or horizontally into a single image.
#'
#'@param input_list List of 3-layer RGB/4-layer RGBA array, `rayimg` class, or image filenames.
#'@param stack Default `"v"`. Stack direction: `"v"`/`"vertical"` or `"h"`/`"horizontal"`.
#'@param align Default `"center"`. Alignment on the non-stacking axis. Options:
#'`"start"`/`"left"`/`"top"`, `"center"`, `"end"`/`"right"`/`"bottom"`.
#'@param background Default `c(0,0,0,0)`. Background color for padding, as numeric
#'`c(r,g,b[,a])` in 0..1 or a color string.
#'@param convert_colorspace Default `TRUE`. Whether to convert all images to match
#'the colorspace of the first image in the list.
#'@param filename Default `NULL`. File to save the image to. If `NULL` and `preview = FALSE`,
#'returns an RGB array.
#'@param preview Default `FALSE`. If `TRUE`, it will display the image in addition
#'to returning it.
#'
#'@return A `rayimg` RGBA array.
#'@export
#'@examplesIf interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")
#'render_stack(list(dragon, 1 - dragon), stack = "h") |>
#'  plot_image()
#'render_stack(list(dragon, render_bw(dragon)), stack = "v") |>
#'  plot_image()
render_stack = function(
  input_list,
  stack = c("v", "h", "vertical", "horizontal"),
  align = c("center", "start", "end", "left", "right", "top", "bottom"),
  background = c(0, 0, 0, 0),
  convert_colorspace = TRUE,
  filename = NULL,
  preview = FALSE
) {
  if (!inherits(input_list, "list") || length(input_list) == 0) {
    stop("`input_list` must be a non-empty list of images.")
  }
  if (any(vapply(input_list, is.null, logical(1)))) {
    stop("`input_list` cannot contain NULL entries.")
  }

  if (missing(stack)) {
    stack = "v"
  } else {
    stack = tolower(stack)
  }
  stack = match.arg(stack, c("v", "h", "vertical", "horizontal"))
  is_vertical = stack %in% c("v", "vertical")
  if (missing(align)) {
    align = "center"
  } else {
    align = tolower(align)
  }
  align = match.arg(
    align,
    c("center", "start", "end", "left", "right", "top", "bottom")
  )
  align = switch(
    align,
    left = "start",
    top = "start",
    right = "end",
    bottom = "end",
    align
  )

  images = lapply(input_list, function(img) {
    ray_read_image(
      img,
      convert_to_array = TRUE,
      reset_camera_settings = TRUE
    )
  })

  base_image = images[[1]]
  base_colorspace = attr(base_image, "colorspace")
  base_white = attr(base_image, "white_current")

  if (convert_colorspace) {
    for (i in seq_along(images)) {
      images[[i]] = render_convert_colorspace(
        images[[i]],
        to_mats = base_colorspace,
        to_white = base_white
      )
    }
    base_white = attr(images[[1]], "white_current")
  }

  heights = vapply(images, function(img) dim(img)[1], numeric(1))
  widths = vapply(images, function(img) dim(img)[2], numeric(1))
  max_height = max(heights)
  max_width = max(widths)

  if (is_vertical) {
    out_height = sum(heights)
    out_width = max_width
  } else {
    out_height = max_height
    out_width = sum(widths)
  }

  bg = convert_color(background, single = TRUE)
  if (length(bg) == 3) {
    bg = c(bg, 1)
  }
  stacked = array(0, dim = c(out_height, out_width, 4))
  stacked[,, 1] = bg[1]
  stacked[,, 2] = bg[2]
  stacked[,, 3] = bg[3]
  stacked[,, 4] = bg[4]

  align_offset = function(total, size) {
    extra = total - size
    if (extra < 0) {
      stop("stacked size cannot be smaller than component size.")
    }
    switch(
      align,
      start = 1L,
      center = floor(extra / 2) + 1L,
      end = extra + 1L
    )
  }

  if (is_vertical) {
    row_start = 1L
    for (i in seq_along(images)) {
      img = unclass(images[[i]])
      h = dim(img)[1]
      w = dim(img)[2]
      col_start = align_offset(max_width, w)
      rows = row_start:(row_start + h - 1L)
      cols = col_start:(col_start + w - 1L)
      stacked[rows, cols, ] = img
      row_start = row_start + h
    }
  } else {
    col_start = 1L
    for (i in seq_along(images)) {
      img = unclass(images[[i]])
      h = dim(img)[1]
      w = dim(img)[2]
      row_start = align_offset(max_height, h)
      rows = row_start:(row_start + h - 1L)
      cols = col_start:(col_start + w - 1L)
      stacked[rows, cols, ] = img
      col_start = col_start + w
    }
  }

  stacked = ray_read_image(
    stacked,
    assume_colorspace = base_colorspace,
    assume_white = base_white,
    source_linear = TRUE
  )
  handle_image_output(stacked, filename = filename, preview = preview)
}

#'@title Add Overlay
#'
#'@description Takes an RGB array/filename and composites an image overlay over
#'the full image. For pixel-positioned sprite, label, glyph, or billboard
#'overlays, use [render_sprite_overlay()].
#'
#'@param image 3-layer RGB/4-layer RGBA array, `rayimg` class, or filename of an image.
#'@param image_overlay Default `NULL`. 3-layer RGB/4-layer RGBA array, `rayimg` class, or filename of an image.
#'This image will be resized to the dimensions of `image` if it does not match
#'exactly, unless `rescale_original = TRUE`.
#'@param rescale_original Default `FALSE`. If `TRUE`, function will resize the original image to match
#'the overlay.
#'@param convert_overlay_colorspace Default `TRUE`. Whether to convert the overlay's colorspace
#'to match the underlying image.
#'@param alpha Default `NA`, using overlay's alpha channel. Otherwise, this sets the alpha transparency
#'by multiplying the existing alpha channel by this value (between 0 and 1).
#'@param filename Default `NULL`. File to save the image to. If `NULL` and `preview = FALSE`,
#'returns an RGB array.
#'@param preview Default `FALSE`. If `TRUE`, it will display the image in addition
#'to returning it.
#'@return A `rayimg` RGBA array.
#'@import grDevices
#'@export
#'@examplesIf interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")
#'#Plot the dragon
#'plot_image(dragon)
#'#Add an overlay of a red semi-transparent circle:
#'circlemat = generate_2d_disk(min(dim(dragon)[1:2]))
#'circlemat = circlemat/max(circlemat)
#'
#'#Create RGBA image, with a transparency of 0.5
#'rgba_array = array(1, dim=c(nrow(circlemat),ncol(circlemat),4))
#'rgba_array[,,1] = circlemat
#'rgba_array[,,2] = 0
#'rgba_array[,,3] = 0
#'dragon_clipped = dragon
#'dragon_clipped[dragon_clipped > 1] = 1
#'render_image_overlay(dragon_clipped, image_overlay = rgba_array,
#'                  alpha=0.5, preview = TRUE)
#'
#' 	# Read photo, convert to ACEScg with CAT (scene)
#' 	photo = ray_read_image(sunset_image, normalize = FALSE)
#' 	photo_aces = render_convert_colorspace(
#' 		photo,
#' 		to_mats = CS_ACESCG,
#' 		adapt_white = TRUE
#' 	)
#' 	tmp_txt = tempfile(fileext = ".png")
#' 	render_text_image(
#' 		"Sunset",
#' 		size = 60,
#' 		filename = tmp_txt,
#' 		color = "#c300ff",
#' 		background_alpha = 0
#' 	)
#' 	# Read logo (display-referred), convert primaries only (no CAT)
#' 	logo = ray_read_image(tmp_txt, normalize = FALSE) # sRGB/D65
#' 	logo_aces = render_convert_colorspace(
#' 		logo,
#' 		to_mats = CS_ACESCG,
#' 		adapt_white = FALSE
#' 	)
#'
#' 	# Composite in ACEScg, then display (plot_image converts to sRGB/D65 + OETF)
#' 	# Here, we also turn overlay conversion in [render_image_overlay()] off,
#'  # to show what happens when you don't account for the colorspace difference.
#'  # By default [render_image_overlay()] will do this for you.
#' 	comp1 = render_image_overlay(
#' 		photo_aces,
#' 		logo_aces,
#' 		convert_overlay_colorspace = FALSE
#' 	) |>
#'    render_title(title_text = "#c300ff: Match",
#'                 title_bar_color = "white", title_color = "#c300ff", title_bar_alpha=1)
#' 	comp2 = render_image_overlay(
#' 		photo_aces,
#' 		logo,
#' 		convert_overlay_colorspace = FALSE
#' 	) |>
#' 	 render_title(title_text = "#c300ff: Incorrect",
#'                title_bar_color = "white", title_color = "#c300ff", title_bar_alpha=1)
#'
#' 	plot_image_grid(list(comp1, comp2), dim = c(1, 2))
render_image_overlay = function(
  image,
  image_overlay = NULL,
  rescale_original = FALSE,
  convert_overlay_colorspace = TRUE,
  alpha = NA,
  filename = NULL,
  preview = FALSE
) {
  render_image_overlay_impl(
    image = image,
    image_overlay = image_overlay,
    rescale_original = rescale_original,
    convert_overlay_colorspace = convert_overlay_colorspace,
    alpha = alpha,
    filename = filename,
    preview = preview
  )
}

#'@title Add Sprite Overlay
#'
#'@description Places an image overlay at a pixel location without resizing it to
#'the full destination image. This is intended for sprite-like or billboard-like
#'overlays such as labels, icons, glyphs, and screen-space annotations.
#'
#'@param image 3-layer RGB/4-layer RGBA array, `rayimg` class, or filename of an image.
#'@param image_overlay Default `NULL`. 3-layer RGB/4-layer RGBA array, `rayimg`
#'class, or filename of the sprite overlay.
#'@param convert_overlay_colorspace Default `TRUE`. Whether to convert the overlay's colorspace
#'to match the underlying image.
#'@param alpha Default `NA`, using overlay's alpha channel. Otherwise, this sets the alpha transparency
#'by multiplying the existing alpha channel by this value (between 0 and 1).
#'@param overlay_coords Default `c(1, 1)`. Pixel coordinate `c(x, y)` used to
#'position the overlay. `x` increases from left to right and `y` from top to
#'bottom, starting at 1.
#'@param overlay_dims Default `NULL`. Dimensions for the overlay in pixels. If
#'provided, the overlay is resized with `render_resized()` before compositing.
#'@param overlay_anchor Default `"nw"`. Which corner of the overlay is placed at
#'`overlay_coords` when no numeric justification is supplied. Options: `"nw"`,
#'`"ne"`, `"sw"`, `"se"`.
#'@param overlay_just Default `NULL`. Optional numeric `c(hjust, vjust)`
#'justification used to place the overlay relative to `overlay_coords`. If set,
#'the overlay's left/top pixel is computed as
#'`round(x - hjust * overlay_width)` and
#'`round(y - vjust * overlay_height)`.
#'@param hjust Default `NULL`. Optional horizontal justification. Overrides the
#'first value of `overlay_just`.
#'@param vjust Default `NULL`. Optional vertical justification. Overrides the
#'second value of `overlay_just`.
#'@param preserve_channels Default `TRUE`. If `TRUE`, RGB input images return
#'RGB output and RGBA input images return RGBA output.
#'@param filename Default `NULL`. File to save the image to. If `NULL` and
#'`preview = FALSE`, returns an image array.
#'@param preview Default `FALSE`. If `TRUE`, it will display the image in addition
#'to returning it.
#'
#'@return A `rayimg` array.
#'@export
#'@examplesIf interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")
#'dragon_clipped = dragon
#'dragon_clipped[dragon_clipped > 1] = 1
#'
#'# Render a transparent dragon emoji and center it on a pixel with hjust/vjust:
#'dragon_emoji = render_text_image(
#'  "\U0001F409",
#'  size = 80,
#'  background_alpha = 0,
#'  use_ragg = TRUE,
#'  trim = TRUE,
#'  trim_padding = 8
#')
#'render_sprite_overlay(
#'  dragon_clipped,
#'  image_overlay = dragon_emoji,
#'  overlay_coords = c(ncol(dragon_clipped) / 2, nrow(dragon_clipped) / 2),
#'  hjust = 0.5,
#'  vjust = 0.5,
#'  preview = TRUE
#')
#'
#'# The same justification can be supplied as overlay_just = c(hjust, vjust):
#'render_sprite_overlay(
#'  dragon_clipped,
#'  image_overlay = dragon_emoji,
#'  overlay_coords = c(ncol(dragon_clipped), nrow(dragon_clipped)),
#'  overlay_just = c(1, 1),
#'  preview = TRUE
#')
render_sprite_overlay = function(
  image,
  image_overlay = NULL,
  convert_overlay_colorspace = TRUE,
  alpha = NA,
  overlay_coords = c(1, 1),
  overlay_dims = NULL,
  overlay_anchor = "nw",
  overlay_just = NULL,
  hjust = NULL,
  vjust = NULL,
  preserve_channels = TRUE,
  filename = NULL,
  preview = FALSE
) {
  render_image_overlay_impl(
    image = image,
    image_overlay = image_overlay,
    rescale_original = FALSE,
    convert_overlay_colorspace = convert_overlay_colorspace,
    alpha = alpha,
    overlay_coords = overlay_coords,
    overlay_dims = overlay_dims,
    overlay_anchor = overlay_anchor,
    overlay_just = overlay_just,
    hjust = hjust,
    vjust = vjust,
    preserve_channels = preserve_channels,
    filename = filename,
    preview = preview
  )
}

#' @keywords internal
render_image_overlay_impl = function(
  image,
  image_overlay = NULL,
  rescale_original = FALSE,
  convert_overlay_colorspace = TRUE,
  alpha = NA,
  overlay_coords = NULL,
  overlay_dims = NULL,
  overlay_anchor = "nw",
  overlay_just = NULL,
  hjust = NULL,
  vjust = NULL,
  preserve_channels = FALSE,
  filename = NULL,
  preview = FALSE
) {
  if (is.null(image_overlay)) {
    stop("Need to pass in image to image_overlay argument.")
  }

  original_channels = {
    image_dim = dim(image)
    if (is.null(image_dim)) {
      NA_integer_
    } else if (length(image_dim) == 2) {
      1L
    } else {
      image_dim[3]
    }
  }

  # Load as rayimg (RGBA + attrs)
  image = ray_read_image(
    image,
    convert_to_array = TRUE,
    reset_camera_settings = TRUE
  )
  image_colorspace = attr(image, "colorspace")
  image_whitepoint = attr(image, "white_current")
  image_overlay = ray_read_image(
    image_overlay,
    convert_to_array = TRUE,
    reset_camera_settings = TRUE
  )
  if (convert_overlay_colorspace) {
    image_overlay = render_convert_colorspace(
      image_overlay,
      to_mats = image_colorspace
    )
  }

  img_type = attr(image, "filetype")
  over_type = attr(image_overlay, "filetype")

  if (!is.null(overlay_dims)) {
    stopifnot(length(overlay_dims) >= 2)
    overlay_dims = as.integer(round(overlay_dims[1:2]))
    stopifnot(all(overlay_dims > 0))
    image_overlay = render_resized(image_overlay, dims = overlay_dims)
  }
  placement_mode = !is.null(overlay_coords) ||
    !is.null(overlay_dims) ||
    !is.null(overlay_just) ||
    !is.null(hjust) ||
    !is.null(vjust)
  if (rescale_original) {
    target_dims = if (is.null(overlay_dims)) {
      dim(image_overlay)[1:2]
    } else {
      overlay_dims
    }
    if (!all(dim(image)[1:2] == target_dims)) {
      image = render_resized(image, dims = target_dims)
    }
  } else if (!placement_mode) {
    if (!all(dim(image)[1:2] == dim(image_overlay)[1:2])) {
      image_overlay = render_resized(image_overlay, dims = dim(image))
    }
  }
  is_matrix_image = length(dim(image)) == 2
  is_matrix_image_overlay = length(dim(image_overlay)) == 2
  process_image = function(image_input) {
    is_matrix_image = length(dim(image_input)) == 2
    if (is_matrix_image) {
      image_tmp = array(image_input, dim = c(dim(image_input), 4))
      image_tmp[,, 4] = 1
    } else {
      if (dim(image_input)[3] == 2) {
        #Greyscale with alpha
        image_tmp = array(image_input[,, 1], dim = c(dim(image_input)[1:2], 4))
        image_tmp[,, 4] = image_input[,, 2]
      } else if (dim(image_input)[3] == 3) {
        #RGB
        image_tmp = array(1, dim = c(dim(image_input)[1:2], 4))
        image_tmp[,, 4] = 1
      } else {
        image_tmp = image_input
      }
    }
    return(image_tmp)
  }
  image = process_image(image)
  image_overlay = process_image(image_overlay)

  if (!is.na(alpha)) {
    stopifnot(alpha >= 0, alpha <= 1)
    image_overlay[,, 4] = image_overlay[,, 4] * alpha
  }

  Cb = image[,, 1:3]
  Ab = image[,, 4]

  # Place overlay on a blank canvas at the requested location, cropping to bounds
  if (is.null(overlay_coords)) {
    overlay_coords = c(1, 1)
  }
  stopifnot(length(overlay_coords) >= 2)
  if (any(!is.finite(overlay_coords))) {
    stop("`overlay_coords` must be finite.")
  }
  overlay_size = dim(image_overlay)[1:2]
  use_overlay_just = !is.null(overlay_just) ||
    !is.null(hjust) ||
    !is.null(vjust)
  if (use_overlay_just) {
    if (is.null(overlay_just)) {
      overlay_just = c(0, 0)
    }
    if (!is.numeric(overlay_just) || length(overlay_just) < 2) {
      stop("`overlay_just` must be a numeric vector of length 2 or greater.")
    }
    overlay_just = overlay_just[1:2]
    if (!is.null(hjust)) {
      if (!is.numeric(hjust) || length(hjust) != 1 || !is.finite(hjust)) {
        stop("`hjust` must be a finite numeric scalar.")
      }
      overlay_just[1] = hjust
    }
    if (!is.null(vjust)) {
      if (!is.numeric(vjust) || length(vjust) != 1 || !is.finite(vjust)) {
        stop("`vjust` must be a finite numeric scalar.")
      }
      overlay_just[2] = vjust
    }
    if (any(!is.finite(overlay_just))) {
      stop("`overlay_just` must be finite.")
    }
    start_col = as.integer(round(
      overlay_coords[1] - overlay_just[1] * overlay_size[2]
    ))
    start_row = as.integer(round(
      overlay_coords[2] - overlay_just[2] * overlay_size[1]
    ))
  } else {
    valid_anchors = c("nw", "ne", "sw", "se")
    overlay_anchor = tolower(overlay_anchor)
    if (!(overlay_anchor %in% valid_anchors)) {
      stop(
        "`overlay_anchor` must be one of: ",
        paste(valid_anchors, collapse = ", ")
      )
    }
    overlay_coords = as.integer(round(overlay_coords[1:2]))
    start_row = overlay_coords[2]
    start_col = overlay_coords[1]
    if (overlay_anchor %in% c("sw", "se")) {
      start_row = overlay_coords[2] - overlay_size[1] + 1
    }
    if (overlay_anchor %in% c("ne", "se")) {
      start_col = overlay_coords[1] - overlay_size[2] + 1
    }
  }
  end_row = start_row + overlay_size[1] - 1
  end_col = start_col + overlay_size[2] - 1
  target_rows = seq(start_row, end_row)
  target_cols = seq(start_col, end_col)
  rows_in_bounds = target_rows >= 1 & target_rows <= dim(image)[1]
  cols_in_bounds = target_cols >= 1 & target_cols <= dim(image)[2]
  overlay_canvas = array(0, dim = dim(image))
  if (any(rows_in_bounds) && any(cols_in_bounds)) {
    dest_rows = target_rows[rows_in_bounds]
    dest_cols = target_cols[cols_in_bounds]
    src_rows = which(rows_in_bounds)
    src_cols = which(cols_in_bounds)
    overlay_canvas[dest_rows, dest_cols, ] = image_overlay[
      src_rows,
      src_cols,
      ,
      drop = FALSE
    ]
  }
  image_overlay = overlay_canvas

  Cf = image_overlay[,, 1:3]
  Af = image_overlay[,, 4]

  Af3 = array(Af, dim = c(dim(Af), 3))
  Ab3 = array(Ab, dim = c(dim(Ab), 3))

  Ao = pmin(pmax(Af + Ab * (1 - Af), 0), 1)
  num = Cf * Af3 + Cb * Ab3 * array(1 - Af, dim = c(dim(Af), 3))

  eps = 1e-8
  Co_lin = num / array(pmax(Ao, eps), dim = c(dim(Ao), 3))

  # ensure fully transparent pixels are black
  if (any(Ao <= eps)) {
    m = Ao <= eps
    Co_lin[,, 1][m] = 0
    Co_lin[,, 2][m] = 0
    Co_lin[,, 3][m] = 0
  }

  composite_image = array(1, dim = dim(image))
  composite_image[,, 4] = Ao
  composite_image[,, 1:3] = Co_lin

  composite_image = ray_read_image(
    composite_image,
    assume_colorspace = image_colorspace,
    assume_white = image_whitepoint,
    source_linear = TRUE
  )
  if (isTRUE(preserve_channels) && identical(original_channels, 3L)) {
    composite_image = composite_image[,, 1:3, drop = FALSE]
  }
  handle_image_output(composite_image, filename = filename, preview = preview)
}

#' Dragon Image
#'
#' @format An RGB 3-layer HDR rayimg array with 200 rows and 200 columns,
#' generated using the rayrender package, with gamma correction.
"dragon"

#' Dragon Depthmap
#'
#' @format An matrix with 200 rows and 200 columns, representing
#' the depth into the `dragon` image scene. Generated using the rayrender package.
#' Distances range from 847 to 1411.
"dragondepth"

#' Sunset Image
#'
#' @format An RGBA 4-layer LDR rayimg array from a JPEG image
#' with 400 rows and 400 columns.
#' @author Tyler Morgan-Wall
"sunset_image"

#'@title Resize Image
#'
#'@description Resizes an image or a matrix, using bilinear interpolation.
#'
#'@param image 3-layer RGB/4-layer RGBA array, `rayimg` class, or filename of an image.
#'@param mag Default `1`. Amount to magnify the image, preserving aspect ratio. Overridden if
#'`dim` is not `NULL`.
#'@param dims Default `NULL`. Exact resized dimensions.
#'@param filename Default `NULL`. The filename of the image to be saved. If this is not given, the image will be plotted instead.
#'@param preview Default `FALSE`. Whether to plot the convolved image, or just to return the values.
#'@param method Default `trilinear`. Filters to up/downsample the image. Options: `bilinear`, `box`, `trilinear`,
#' `catmull`, `mitchell`.
#'@return A `rayimg` RGBA array.
#'@export
#'@examplesIf interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")
#'#Plot the image with a title
#'dragon |>
#'  render_title("Dragon", title_offset=c(10,10), title_bar_color="black",
#'            title_size=20, title_color = "white") |>
#'  plot_image()
#'#Half of the resolution
#'render_resized(dragon, mag = 1/2) |>
#'  render_title("Dragon (half res)", title_offset=c(5,5), title_bar_color="black",
#'            title_size=10, title_color = "white") |>
#'  plot_image()
#'#Double the resolution
#'render_resized(dragon, mag = 2) |>
#'  render_title("Dragon (2x res)", title_offset=c(20,20), title_bar_color="black",
#'            title_size=40, title_color = "white") |>
#'  plot_image()
#'#Specify the exact resulting dimensions
#'render_resized(dragon, dim = c(160,320)) |>
#'  render_title("Dragon (custom size)", title_offset=c(10,10), title_bar_color="black",
#'            title_size=20, title_color = "white") |>
#'  plot_image()
render_resized = function(
  image,
  mag = 1,
  dims = NULL,
  filename = NULL,
  preview = FALSE,
  method = "tri"
) {
  temp_image = ray_read_image(image)
  orig_dims = dim(temp_image)[1:2]
  downscale = if (!is.null(dims)) {
    any(dims[1:2] < orig_dims)
  } else {
    isTRUE(mag < 1)
  }
  use_bilinear = method %in% c("bi", "bilinear") && !downscale
  if (!use_bilinear && method %in% c("bi", "bilinear")) {
    method = "box"
  }
  #Check if file or image before below:
  imagetype = attr(temp_image, "filetype")
  img_source_linear = attr(temp_image, "source_linear")
  colorspace = attr(temp_image, "colorspace")
  white_current = attr(temp_image, "white_current")

  if (use_bilinear) {
    if (!is.null(dims)) {
      dims = dims[1:2] / dim(temp_image)[1:2]
    }
    temp_list = list()
    if (length(dim(temp_image)) == 3) {
      channels = dim(temp_image)[3]
      for (i in seq_len(channels)) {
        if (is.null(dims)) {
          temp_list[[i]] = resize_image(temp_image[,, i], mag)
        } else {
          x1 = seq(1, nrow(temp_image), length.out = nrow(temp_image) * dims[1])
          y1 = seq(1, ncol(temp_image), length.out = ncol(temp_image) * dims[2])
          temp_list[[i]] = resize_image_xy(temp_image[,, i], x1, y1)
        }
      }
      temp_image = array(
        0,
        dim = c(nrow(temp_list[[1]]), ncol(temp_list[[1]]), channels)
      )
      for (i in seq_len(channels)) {
        temp_image[,, i] = temp_list[[i]]
      }
    } else {
      if (is.null(dims)) {
        temp_image = resize_image(t(flipud(temp_image)), mag)
      } else {
        x1 = seq(1, nrow(temp_image), length.out = nrow(temp_image) * dims[1])
        y1 = seq(1, ncol(temp_image), length.out = ncol(temp_image) * dims[2])
        temp_image = resize_image_xy(t(flipud(temp_image)), x1, y1)
      }
    }
  } else {
    method = tolower(method)
    method = switch(
      method,
      "default" = 0,
      "box" = 1,
      "triangle" = 2,
      "tri" = 3,
      "trilinear" = 3,
      "cubic" = 3,
      "catmull" = 4,
      "mitchell" = 5,
      3
    )
    if (is.null(dims)) {
      dims = mag * dim(temp_image)[1:2]
    }
    temp_list = list()
    if (length(dim(temp_image)) == 3) {
      for (i in seq_len(dim(temp_image)[3])) {
        temp_list[[i]] = resize_matrix_stb(
          unclass(temp_image[,, i]),
          dims[1],
          dims[2],
          method
        )
      }
      temp_image = array(0, dim = c(dims[1], dims[2], dim(temp_image)[3]))
      for (i in seq_len(dim(temp_image)[3])) {
        temp_image[,, i] = temp_list[[i]]
      }
    } else {
      bare_matrix = unclass(temp_image)
      temp_image = resize_matrix_stb(
        bare_matrix,
        dims[1],
        dims[2],
        method
      )
    }
  }
  temp_image = ray_read_image(
    temp_image,
    filetype = imagetype,
    source_linear = img_source_linear,
    assume_colorspace = colorspace,
    assume_white = white_current
  )
  handle_image_output(temp_image, filename = filename, preview = preview)
}

#'@title Reorient Image
#'
#'@description Reorients an image or matrix. Transformations are applied in this order: x, y, and transpose.
#'
#'@param image 3-layer RGB/4-layer RGBA array, `rayimg` class, or filename of an image.
#'@param flipx Default `FALSE`. Flip horizontally
#'@param flipy Default `FALSE`. Flip vertically.
#'@param transpose Default `FALSE`. Transpose image.
#'@param filename Default `NULL`. The filename of the image to be saved. If this is not given, the image will be plotted instead.
#'@param preview Default `FALSE`. Whether to plot the convolved image, or just to return the values.
#'@return 3-layer RGB reoriented array or matrix.
#'@export
#'@examplesIf interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")
#'#Original orientation
#'plot_image(dragon)
#'#Flip the dragon image horizontally
#'dragon |>
#'  render_reorient(flipx = TRUE) |>
#'  plot_image()
#'#Flip the dragon image vertically
#'dragon |>
#'  render_reorient(flipy = TRUE) |>
#'  plot_image()
#'#Transpose the dragon image
#'dragon |>
#'  render_reorient(transpose = TRUE) |>
#'  plot_image()
render_reorient = function(
  image,
  flipx = FALSE,
  flipy = FALSE,
  transpose = FALSE,
  filename = NULL,
  preview = FALSE
) {
  #Check if file or image before below:
  temp_image = ray_read_image(image, convert_to_array = FALSE)
  imagetype = attr(temp_image, "filetype")
  img_source_linear = attr(temp_image, "source_linear")
  colorspace = attr(temp_image, "colorspace")
  white_current = attr(temp_image, "white_current")

  if (flipx) {
    temp_image = fliplr(temp_image)
  }
  if (flipy) {
    temp_image = flipud(temp_image)
  }
  if (transpose) {
    if (imagetype == "matrix") {
      temp_image = t(temp_image)
    } else {
      temp_image = aperm(temp_image, c(2, 1, 3))
    }
  }
  temp_image = ray_read_image(
    temp_image,
    filetype = imagetype,
    source_linear = img_source_linear,
    assume_colorspace = colorspace,
    assume_white = white_current
  )

  handle_image_output(temp_image, filename = filename, preview = preview)
}

#' Vertical ellipsis symbol
#'
#' @param use_cli Default `TRUE`. If `TRUE`, consult `cli::is_utf8_output()` for Unicode support.
#' @return vertical ellipses if supported
#' @keywords internal
rayimg_vertical_ellipsis = function(use_cli = TRUE) {
  have_cli = use_cli && requireNamespace("cli", quietly = TRUE)
  unicode_ok = if (have_cli) isTRUE(cli::is_utf8_output()) else
    isTRUE(l10n_info()[["UTF-8"]])
  if (unicode_ok) "\u22EE" else "..."
}

#' @keywords internal
rayimg_truecolor_supported = function() {
  if (!requireNamespace("cli", quietly = TRUE)) return(FALSE)
  dyn = isTRUE(cli::is_dynamic_tty())
  utf8 = isTRUE(cli::is_utf8_output())
  ncol = cli::num_ansi_colors()
  tc = grepl("truecolor|24bit", Sys.getenv("COLORTERM", ""), ignore.case = TRUE)
  dyn && utf8 && (tc || isTRUE(ncol >= 256))
}

#' Compose RGB over checkerboard using alpha
#' @keywords internal
rayimg_composite_checker = function(
  rgb,
  alpha = NULL,
  light = 0.92,
  dark = 0.75
) {
  d = dim(rgb)
  H = d[1]
  W = d[2]
  if (is.null(alpha)) return(rgb)
  pat = outer(seq_len(H), seq_len(W), function(i, j) (i + j) %% 2L == 0L)
  bg = array(dark, dim = c(H, W, 3))
  bg[pat] = light
  out = array(0, dim = c(H, W, 3))
  for (k in 1:3) {
    out[,, k] = rgb[,, k] * alpha + bg[,, k] * (1 - alpha)
  }
  return(out)
}

# Truecolor detection
rayimg_truecolor_supported = function() {
  if (!requireNamespace("cli", quietly = TRUE)) return(FALSE)
  dyn = isTRUE(cli::is_dynamic_tty())
  utf8 = isTRUE(cli::is_utf8_output())
  # Require an explicit hint for 24-bit (COLORTERM), or a TERM that advertises it (rare)
  tc = grepl("truecolor|24bit", Sys.getenv("COLORTERM", ""), ignore.case = TRUE)
  term = Sys.getenv("TERM", "")
  direct = grepl("direct", term, ignore.case = TRUE) # e.g. xterm-direct/kitty-direct
  dyn && utf8 && (tc || direct)
}

# Map RGB [0, 255] to nearest xterm-256 index
rayimg_rgb_to_xterm256 = function(r, g, b) {
  # 6x6x6 color cube
  lev = c(0, 95, 135, 175, 215, 255)
  idx = function(x) {
    which.min((lev - x)^2) - 1L
  }
  ri = idx(r)
  gi = idx(g)
  bi = idx(b)
  cube_idx = 16L + 36L * ri + 6L * gi + bi
  cube_rgb = c(lev[ri + 1L], lev[gi + 1L], lev[bi + 1L])

  # grayscale ramp [232-255] (approx [8-238] step 10)
  gray_val = round((r + g + b) / 3)
  gray_i = as.integer(pmin(23L, pmax(0L, round((gray_val - 8) / 10))))
  gray_idx = 232L + gray_i
  gray_rgb = rep(8L + 10L * gray_i, 3L)

  # choose closer
  d_cube = sum((c(r, g, b) - cube_rgb)^2)
  d_gray = sum((c(r, g, b) - gray_rgb)^2)
  if (d_gray < d_cube) {
    return(gray_idx)
  } else {
    return(cube_idx)
  }
}

# One ANSI cell with 24-bit or 256-color fallback
# style: "auto" pick bg/fg based on luminance; "block"/"bg" force mode; "bg256"/"block256" force 256-color
rayimg_ansi_cell = function(
  r,
  g,
  b,
  style = c("auto", "block", "bg", "bg256", "block256")
) {
  style = match.arg(style)
  r = as.integer(pmin(pmax(round(r), 0L), 255L))
  g = as.integer(pmin(pmax(round(g), 0L), 255L))
  b = as.integer(pmin(pmax(round(b), 0L), 255L))

  tc_ok = rayimg_truecolor_supported()

  # luminance coefs
  Y = 0.2126 * r + 0.7152 * g + 0.0722 * b
  if (style == "auto") {
    style = if (tc_ok) {
      (if (Y < 80) "block" else "bg")
    } else {
      (if (Y < 80) "block256" else "bg256")
    }
  }

  # 24-bit variants
  if (style == "bg" && tc_ok) {
    return(paste0("\033[48;2;", r, ";", g, ";", b, "m", "  ", "\033[0m"))
  }
  if (style == "block" && tc_ok) {
    return(paste0(
      "\033[38;2;",
      r,
      ";",
      g,
      ";",
      b,
      "m",
      "\u2588\u2588",
      "\033[0m"
    ))
  }

  # 256-color fallback
  idx = rayimg_rgb_to_xterm256(r, g, b)
  if (style %in% c("bg256", "bg")) {
    return(paste0("\033[48;5;", idx, "m", "  ", "\033[0m"))
  }
  if (style %in% c("block256", "block")) {
    return(paste0("\033[38;5;", idx, "m", "\u2588\u2588", "\033[0m"))
  }

  # last resort
  "\u2588\u2588"
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

get_boolean_distance <- function(input) {
    .Call(`_rayimage_get_boolean_distance`, input)
}

encode_native_image_rcpp_1 <- function(bw_image) {
    .Call(`_rayimage_encode_native_image_rcpp_1`, bw_image)
}

encode_native_image_rcpp_3 <- function(r_image, g_image, b_image) {
    .Call(`_rayimage_encode_native_image_rcpp_3`, r_image, g_image, b_image)
}

encode_native_image_rcpp_4 <- function(r_image, g_image, b_image, a_image) {
    .Call(`_rayimage_encode_native_image_rcpp_4`, r_image, g_image, b_image, a_image)
}

read_dng_cpp <- function(filename, normalize = TRUE, pick = "largest") {
    .Call(`_rayimage_read_dng_cpp`, filename, normalize, pick)
}

dng_apply_active_area_cpp <- function(pixels, active_area) {
    .Call(`_rayimage_dng_apply_active_area_cpp`, pixels, active_area)
}

dng_exposure_normalize_cpp <- function(pixels, iso, exposure_time) {
    .Call(`_rayimage_dng_exposure_normalize_cpp`, pixels, iso, exposure_time)
}

dng_undo_baseline_exposure_cpp <- function(pixels, baseline_exposure) {
    .Call(`_rayimage_dng_undo_baseline_exposure_cpp`, pixels, baseline_exposure)
}

write_dng_cpp <- function(filename, pixels, meta, bitdepth = 16L, compress = TRUE) {
    .Call(`_rayimage_write_dng_cpp`, filename, pixels, meta, bitdepth, compress)
}

load_image_stb <- function(filename, width, height, channels, desired_channels) {
    .Call(`_rayimage_load_image_stb`, filename, width, height, channels, desired_channels)
}

subsample <- function(circle, size) {
    .Call(`_rayimage_subsample`, circle, size)
}

rayinterp2 <- function(image, XI, YI) {
    .Call(`_rayimage_rayinterp2`, image, XI, YI)
}

resize_image <- function(image, mag) {
    .Call(`_rayimage_resize_image`, image, mag)
}

resize_image_xy <- function(image, XI, YI) {
    .Call(`_rayimage_resize_image_xy`, image, XI, YI)
}

resize_matrix_stb <- function(image, width, height, method) {
    .Call(`_rayimage_resize_matrix_stb`, image, width, height, method)
}

generate_disk <- function(radius, dim, offsetx, offsety) {
    .Call(`_rayimage_generate_disk`, radius, dim, offsetx, offsety)
}

gen_ellipse <- function(intensity, width, height) {
    .Call(`_rayimage_gen_ellipse`, intensity, width, height)
}

subsample_rect <- function(rect, binsx, binsy) {
    .Call(`_rayimage_subsample_rect`, rect, binsx, binsy)
}

gen_circle_psf <- function(radius) {
    .Call(`_rayimage_gen_circle_psf`, radius)
}

is_inside <- function(sizehex, positionx, positiony, sinval, cosval) {
    .Call(`_rayimage_is_inside`, sizehex, positionx, positiony, sinval, cosval)
}

gen_hex_psf <- function(radius, rotation) {
    .Call(`_rayimage_gen_hex_psf`, radius, rotation)
}

psf <- function(image, blurmatrix, depthmap, depth, custombokeh, type, bokehintensity, bokehlimit, rotation, progbar, channel) {
    .Call(`_rayimage_psf`, image, blurmatrix, depthmap, depth, custombokeh, type, bokehintensity, bokehlimit, rotation, progbar, channel)
}

convolution_cpp <- function(image, kernel, progbar, channel, bloom_matrix) {
    .Call(`_rayimage_convolution_cpp`, image, kernel, progbar, channel, bloom_matrix)
}

test_that("active area cropping reduces dims and matches subset", {
  img = array(seq_len(4 * 5), dim = c(4, 5))
  out = rayimage:::dng_apply_active_area_cpp(img, c(1, 1, 3, 4))
  expect_equal(dim(out), c(2, 3))
  expect_equal(out, img[2:3, 2:4])
})

test_that("exposure normalization scales values", {
  img = array(c(0.1, 0.2, 0.3, 0.4), dim = c(2, 2))
  out = rayimage:::dng_exposure_normalize_cpp(img, iso = 200, exposure_time = 0.01)
  scale = (100 / 200) * (1 / 0.01)
  expect_equal(out, img * scale)
})

test_that("baseline exposure undo reverses scaling", {
  img = array(seq_len(4), dim = c(2, 2))
  baseline = 1.0
  scaled = img * (2 ^ baseline)
  out = rayimage:::dng_undo_baseline_exposure_cpp(scaled, baseline)
  expect_equal(out, img)
})

#'@title Render Black and White
#'
#'@description Transforms an image to black and white, preserving luminance.
#'
#'@param image 3-layer RGB/4-layer RGBA array, `rayimg` class, or filename of an image.
#'@param rgb_coef Default `c(0.2126, 0.7152, 0.0722)`.
#'Length-3 numeric vector listing coefficients to convert RGB to luminance.
#'@param filename Default `NULL`. The filename of the image to be saved. If this is not given, the image will be plotted instead.
#'@param preview Default `FALSE`. Whether to plot the convolved image, or just to return the values.
#'@return A `rayimg` RGBA array.
#'@export
#'@examplesIf interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")
#'#Plot the image with a title
#'dragon |>
#'  render_title("Dragon", title_offset=c(10,10), title_bar_color="black",
#'            title_size=20, title_color = "white") |>
#'  render_bw(preview = TRUE)
render_bw = function(
  image,
  rgb_coef = c(0.2126, 0.7152, 0.0722),
  filename = NULL,
  preview = FALSE
) {
  stopifnot(length(rgb_coef) == 3 && is.numeric(rgb_coef))
  src = ray_read_image(image, reset_camera_settings = TRUE)
  imagetype = attr(src, "filetype")
  img_source_linear = attr(src, "source_linear")
  colorspace = attr(src, "colorspace")
  white_current = attr(src, "white_current")

  d = dim(src)
  is_array = length(d) == 3

  if (attr(src, "filetype") == "matrix" || (is_array && d[3] == 2)) {
    return(handle_image_output(src, filename = filename, preview = preview))
  }

  # luminance
  lum = rgb_coef[1] *
    src[,, 1] +
    rgb_coef[2] * src[,, 2] +
    rgb_coef[3] * src[,, 3]

  if (length(dim(src)) == 2) {
    out = lum
  } else {
    out = array(1, dim = c(d[1], d[2], 2))
    out[,, 1] = lum
    out[,, 2] = if (d[3] == 4L) src[,, 4] else 1
  }

  out = ray_read_image(
    out,
    filetype = imagetype,
    source_linear = img_source_linear,
    assume_colorspace = colorspace,
    assume_white = white_current
  )

  handle_image_output(out, filename = filename, preview = preview)
}

#' Check Filename
#'
#' @param file Filename to be checked
#' @return Filename
#' @keywords internal
#'
#' @examples
#' #Fake example
get_file_type = function(file) {
  if (is.character(file)) {
    ext = tolower(tools::file_ext(file))
    if (ext == "png") {
      imagetype = "png"
    } else if (ext %in% c("tif", "tiff")) {
      imagetype = "tif"
    } else if (ext %in% c("jpeg", "jpg")) {
      imagetype = "jpg"
    } else if (ext == "exr") {
      imagetype = "exr"
    } else if (ext == "hdr") {
      imagetype = "hdr"
    } else if (ext == "dng") {
      imagetype = "dng"
    } else if (ext == "tga") {
      imagetype = "tga"
    } else if (ext == "bmp") {
      imagetype = "bmp"
    } else if (ext == "psd") {
      imagetype = "psd"
    } else if (ext == "gif") {
      imagetype = "gif"
    } else if (ext == "pic") {
      imagetype = "pic"
    } else if (ext %in% c("pnm", "ppm", "pgm")) {
      imagetype = "pnm"
    } else {
      stop(
        "`",
        file,
        "` not recognized class (png, tiff, jpeg, exr, hdr, dng, tga, bmp, psd, gif, pic, pnm, array, matrix)."
      )
    }
  } else if (length(dim(file)) == 3 && dim(file)[3] == 1) {
    imagetype = "matrix"
  } else if (length(dim(file)) == 3) {
    imagetype = "array"
  } else if (length(dim(file)) == 2) {
    imagetype = "matrix"
  } else if (length(dim(file)) > 3) {
    stop(
      "Images can only be paths to image files, matrices, ",
      "and arrays with fewer than 3 dimensions"
    )
  } else {
    if (is.character(file)) {
      stop(
        "`",
        file,
        "` not recognized class (png, tiff, jpeg, exr, hdr, dng, tga, bmp, psd, gif, pic, pnm, array, matrix)."
      )
    } else {
      stop(
        "input not recognized class (png, tiff, jpeg, exr, hdr, dng, tga, bmp, psd, gif, pic, pnm, array, matrix)."
      )
    }
  }
  imagetype
}

#' @title Select PNG Device
#'
#' @description Internal helper to pick the best available PNG graphics
#' device. Uses `ragg::agg_png()` by default, then falls back to a
#' Cairo-backed `grDevices::png()` device when `use_ragg = FALSE`, and
#' finally to the base PNG device.
#'
#' @param use_ragg Default `TRUE`. If `TRUE`, return `ragg::agg_png()`.
#' Set to `FALSE` to skip `ragg`.
#'
#' @keywords internal
#' @importFrom ragg agg_png
linear_png_device = function(use_ragg = TRUE) {
  if (isTRUE(use_ragg)) {
    return(function(...) {
      args = list(...)
      args$family = NULL
      do.call(agg_png, args)
    })
  }

  if (isTRUE(capabilities("cairo"))) {
    return(function(...) {
      grDevices::png(..., type = "cairo")
    })
  }

  warning(
    "No cairo device available: install the {ragg} package to ensure correct treatment of gamma when adding text to images.",
    call. = FALSE
  )

  grDevices::png
}

#'@title Make Vignette Overlay
#'
#'@description Makes an overlay to simulate vignetting in a camera
#'
#'@param width Width of the image.
#'@param height Height of the image.
#'@param intensity Default `0.4`. `1` is max intensity, `0` is min.
#'@param radius Default `NULL`. Max of height or width, divided by 2.
#'@param radius_multiplier Default `1.3`.
#'@param color Default `black`.
#'@keywords internal
make_vignette_overlay = function(
  width,
  height,
  intensity = 0.3,
  radius_multiplier = 1.3,
  radius = NULL,
  color = "#000000"
) {
  value = array(
    0,
    dim = c(
      floor(width * radius_multiplier),
      floor(height * radius_multiplier),
      4
    )
  )
  hex_color = convert_color(color, as_hex = TRUE)
  color = convert_color(color)
  value[,, 4] = gen_ellipse(
    intensity,
    floor(width * radius_multiplier),
    floor(height * radius_multiplier)
  )
  value[,, 1] = color[1]
  value[,, 2] = color[2]
  value[,, 3] = color[3]

  tempcircle = tempfile(fileext = ".png")
  png::writePNG(value, tempcircle)
  imageval = magick::image_read(tempcircle)
  if (radius_multiplier >= 1) {
    imageval |>
      magick::image_blur(radius = radius / 2, sigma = radius / 4) |>
      magick::image_crop(sprintf(
        "%ix%i+%i+%i",
        height,
        width,
        floor((height * radius_multiplier - height) / 2),
        floor((width * radius_multiplier - width) / 2)
      )) |>
      magick::image_write(tempcircle)
  } else {
    imageval |>
      magick::image_border(
        geometry = sprintf(
          "%ix%i",
          (height - floor(height * radius_multiplier)) / 2,
          (width - floor(width * radius_multiplier)) / 2
        ),
        color = hex_color
      ) |>
      magick::image_blur(radius = radius / 2, sigma = radius / 4) |>
      magick::image_write(tempcircle)
  }
  return(tempcircle)
}

#' @title Convert to Native Raster
#'
#' @description Converts a matrix or image array to a native raster.
#'
#' @param input Matrix or image array to convert to a native raster.
#' @keywords internal
convert_to_native_raster = function(input) {
  ncol = ncol(input)
  nrow = nrow(input)
  if (length(dim(input)) == 3) {
    if (dim(input)[3] == 3) {
      channels = 3
      nr = encode_native_image_rcpp_3(input[,, 1], input[,, 2], input[,, 3])
    } else if (dim(input)[3] == 4) {
      channels = 4
      nr = encode_native_image_rcpp_4(
        input[,, 1],
        input[,, 2],
        input[,, 3],
        input[,, 4]
      )
    } else if (dim(input)[3] == 2) {
      channels = 3
      nr = encode_native_image_rcpp_4(
        input[,, 1],
        input[,, 1],
        input[,, 1],
        input[,, 2]
      )
    } else {
      stop(sprintf(
        "Dimension of array (%d/%d/%d) do not correspond to any image types",
        dim(input)[1],
        dim(input)[2],
        dim(input)[3]
      ))
    }
  } else if (length(dim(input)) == 2) {
    channels = 3
    nr = encode_native_image_rcpp_1(input)
  } else {
    stop(sprintf("Dimensions of array do not correspond to any image types"))
  }
  class(nr) = "nativeRaster"
  dim(nr) = c(nrow, ncol)
  attr(nr, "channels") = channels
  return(nr)
}

prepare_native_raster = function(image, angle = 0, show_linear = FALSE) {
  img = ray_read_image(image, convert_to_array = TRUE)

  display = img
  cs_from = attr(display, "colorspace")
  white_curr = attr(display, "white_current")

  display = render_clamp(display)
  display[is.na(display)] = 0

  alpha_channel = unclass(render_clamp(
    display[,, 4],
    min_value = 0,
    max_value = 1
  ))
  display[,, 4] = alpha_channel

  if (!show_linear) {
    if (
      !is.null(cs_from) &&
        (!identical(cs_from$name, "sRGB") ||
          any(abs(white_curr - CS_SRGB$white_xyz) > 1e-8))
    ) {
      xyz = apply_color_matrix(display, cs_from$rgb_to_xyz)
      if (any(abs(white_curr - CS_SRGB$white_xyz) > 1e-8)) {
        CAT = compute_cat_bradford(white_curr, CS_SRGB$white_xyz)
        xyz = apply_color_matrix(xyz, CAT)
      }
      display = apply_color_matrix(xyz, CS_SRGB$xyz_to_rgb)
      display[,, 1:3][display[,, 1:3] < 0] = 0
    }
    display[,, 1:3] = to_srgb(display[,, 1:3])
  }

  if (!isTRUE(all.equal(angle %% 360, 0))) {
    display = rotate_image_array(display, angle)
  }

  list(
    native_raster = convert_to_native_raster(display),
    display_dim = dim(display)
  )
}

#'@title Add Vignette Effect
#'
#'@description Takes an RGB array/filename and adds a camera vignette effect.
#'
#'@param image 3-layer RGB/4-layer RGBA array, `rayimg` class, or filename of an image.
#'@param filename Default `NULL`. Filename which to save the image. If `NULL` and `preview = FALSE`,
#'returns an RGB array.
#'@param vignette Default `1`. A camera vignetting effect will be added to the image.
#'`1` is the darkest vignetting, while `0` is no vignetting. If vignette is a length-2 vector, the second entry will
#'control the blurriness of the vignette effect (`1` is the default, e.g. `2` would double the blurriness but would take
#'much longer to compute).
#'@param radius Default `1.1`. Multiplier for the size of the vignette. If `1`, the vignette touches
#'the edge of the image.
#'@param color Default `"#000000"` (black). Color of the vignette.
#'@param preview Default `FALSE`. If `TRUE`, it will display the image in addition
#'to returning it.
#'
#'@return A `rayimg` RGBA array.
#'@import grDevices
#'@export
#'@examplesIf interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")
#'#Plot the dragon
#'plot_image(dragon)
#'#Add a vignette effect:
#'render_vignette(dragon, preview = TRUE, vignette = 1, radius=1.1)
#'#Lighten the vignette effect:
#'render_vignette(dragon, preview = TRUE, vignette = 0.5)
#'#Change the radius:
#'render_vignette(dragon, preview = TRUE, vignette = 1, radius=1.5)
#'render_vignette(dragon, preview = TRUE, vignette = 0.1, radius=0.8)
#'#Change the color:
#'render_vignette(dragon, preview = TRUE, vignette = 1, color="white")
#'#Increase the width of the blur by 50%:
#'render_vignette(dragon, preview = TRUE, vignette = c(1,1.5))
render_vignette = function(
  image,
  vignette = 0.5,
  color = "#000000",
  radius = 1.1,
  filename = NULL,
  preview = FALSE
) {
  image = ray_read_image(image, reset_camera_settings = TRUE)
  dimensions = dim(image)

  if (!("magick" %in% rownames(utils::installed.packages()))) {
    stop("`magick` package required for adding overlay")
  }
  if (length(vignette) > 1) {
    if (vignette[2] < 0) {
      stop("vignette[2] must be greater than 0")
    }
    radiusval = min(c(dimensions[1], dimensions[2])) / 2 * vignette[2]
    vignette = vignette[1]
  } else {
    radiusval = min(c(dimensions[1], dimensions[2])) / 2
  }
  if (is.numeric(vignette)) {
    if (vignette[1] > 1 || vignette[1] < 0) {
      stop("vignette value (", vignette[1], ") must be between 0 and 1.")
    }
  } else {
    vignette = 0.4
  }
  image_vignette = make_vignette_overlay(
    width = dimensions[1],
    height = dimensions[2],
    intensity = vignette,
    radius = radiusval,
    radius_multiplier = radius,
    color = color
  )
  temp = render_image_overlay(image, image_vignette)
  handle_image_output(temp, filename = filename, preview = preview)
}

#'@title Plot Native Raster with Custom Aspect Ratio
#'
#'@keywords internal
plot_asp_native_raster = function(
  nr,
  asp = 1,
  show = FALSE,
  return_grob = FALSE,
  gp = grid::gpar(),
  angle = 0
) {
  stopifnot(asp > 0)
  native_dim = dim(nr)
  image_width = native_dim[2]
  image_height = native_dim[1]
  asp = asp * image_width / image_height
  if (asp > 1) {
    gl = grid::grid.layout(
      1,
      1,
      widths = grid::unit(1, "null"),
      heights = grid::unit(1 / asp, "null"),
      respect = TRUE
    )
  } else {
    gl = grid::grid.layout(
      1,
      1,
      widths = grid::unit(1 * asp, "null"),
      heights = grid::unit(1, "null"),
      respect = TRUE
    )
  }

  needs_rotation = !isTRUE(all.equal(angle %% 360, 0))
  clip_setting = if (needs_rotation) "off" else "inherit"

  top.vp = grid::viewport(
    layout = gl,
    name = "grid_asp_container",
    clip = clip_setting
  )
  image_vp = grid::viewport(
    name = "image",
    layout.pos.col = 1,
    layout.pos.row = 1,
    xscale = c(0, image_width),
    yscale = c(image_height, 0),
    gp = gp,
    clip = clip_setting
  )

  image_grob = grid::rasterGrob(
    nr,
    interpolate = FALSE,
    width = grid::unit(1, "npc"),
    height = grid::unit(1, "npc"),
    gp = gp
  )
  if (needs_rotation) {
    angle_norm = abs(angle) %% 180
    if (angle_norm > 90) {
      angle_norm = 180 - angle_norm
    }
    angle_rad = angle_norm * pi / 180
    width_rot = image_width * cos(angle_rad) + image_height * sin(angle_rad)
    height_rot = image_width * sin(angle_rad) + image_height * cos(angle_rad)
    scale_factor = max(width_rot / image_width, height_rot / image_height)
    if (!is.finite(scale_factor) || scale_factor <= 0) {
      scale_factor = 1
    }
    image_grob = grid::editGrob(
      image_grob,
      vp = grid::viewport(
        angle = angle,
        clip = "off",
        width = grid::unit(1 / scale_factor, "npc"),
        height = grid::unit(1 / scale_factor, "npc")
      )
    )
  }

  if (return_grob) {
    image_tree = grid::gTree(
      children = grid::gList(image_grob),
      vp = image_vp
    )
    return(grid::gTree(
      children = grid::gList(image_tree),
      vp = top.vp
    ))
  }

  pushed = 0L
  grid::pushViewport(top.vp)
  pushed = pushed + 1L

  grid::pushViewport(image_vp)
  pushed = pushed + 1L

  on.exit(
    {
      if (pushed > 0L) {
        try(grid::popViewport(pushed), silent = TRUE)
      }
    },
    add = TRUE
  )

  grid::grid.draw(image_grob)
  invisible()
}

snapshot_variant = function() {
  sysname = Sys.info()[["sysname"]]
  if (dir.exists(testthat::test_path("_snaps", sysname))) {
    return(sysname)
  }
  "Darwin"
}

snapshot_tests_enabled = function() {
  identical(Sys.getenv("RAYIMAGE_RUN_SNAPSHOT_TESTS"), "true")
}

skip_unless_snapshot_tests_enabled = function() {
  if (snapshot_tests_enabled()) {
    return(invisible())
  }
  testthat::skip(
    "Visual snapshot tests are local-only. Set RAYIMAGE_RUN_SNAPSHOT_TESTS=true to run them."
  )
}

save_test_png = function(code, path) {
  grDevices::png(filename = path, width = 500, height = 400)
  dev_id = grDevices::dev.cur()

  on.exit(
    {
      try(grid::upViewport(0), silent = TRUE)
      devs = grDevices::dev.list()
      if (!is.null(devs) && dev_id %in% devs) {
        grDevices::dev.off(dev_id)
      }
    },
    add = TRUE
  )

  code()
  path
}

compare_image = function(
  path1,
  path2,
  quantile_diff = 0.001,
  cdf_diff = 0.1,
  max_diff = 0.3
) {
  image1 = unclass(ray_read_image(path1, convert_to_array = TRUE))
  image2 = unclass(ray_read_image(path2, convert_to_array = TRUE))

  dim1 = dim(image1)
  dim2 = dim(image2)

  attributes(image1) = list(dim = dim1)
  attributes(image2) = list(dim = dim2)

  if (!identical(dim1, dim2)) {
    warning(
      sprintf(
        "Dim not the same: %s vs %s",
        paste0(dim1, collapse = "x"),
        paste0(dim2, collapse = "x")
      ),
      call. = FALSE
    )
    return(FALSE)
  }

  diffs = abs(image2 - image1)
  #Ignore small differences
  mostly_identical = sum(diffs > cdf_diff) < length(diffs) * quantile_diff
  diffs_are_minor = max(diffs) < max_diff
  if (!mostly_identical) {
    warning(
      sprintf(
        "Number greater than CDF diff: %i vs %i/%i",
        sum(diffs > cdf_diff),
        length(diffs) * quantile_diff,
        length(diffs)
      ),
      call. = FALSE
    )
  }
  if (!diffs_are_minor) {
    warning(
      sprintf("Max diff: %f vs %f", max(diffs), max_diff),
      call. = FALSE
    )
  }
  return(mostly_identical && diffs_are_minor)
}

expect_snapshot_file_info = function(path, name, compare, variant, args, i) {
  tryCatch(
    {
      expect_snapshot_file(
        path = path,
        name = name,
        compare = compare,
        variant = variant
      )
    },
    error = function(e) {
      is_mat_args = unlist(lapply(args, is.array)) |
        unlist(lapply(args, is.list))
      for (arg_index in seq_along(args)) {
        if (is_mat_args[arg_index]) {
          args[[arg_index]] = "[[ matrix / array ]]"
        }
      }
      stop(sprintf(
        "Snapshot test failed on row %d with arguments: %s\n\nOriginal error: %s",
        i,
        paste(names(args), unname(args), sep = "=", collapse = ", "),
        e$message
      ))
    }
  )
}

expect_no_error_info = function(code, args, i) {
  tryCatch(
    {
      expect_no_error(
        code
      )
    },
    error = function(e) {
      is_mat_args = unlist(lapply(args, is.array)) |
        unlist(lapply(args, is.list)) |
        unlist(lapply(args, \(x) length(x) > 10))
      for (arg_index in seq_along(args)) {
        if (is_mat_args[arg_index]) {
          args[[arg_index]] = "[[ matrix / array ]]"
        }
      }
      stop(sprintf(
        "Snapshot test failed on row %d with arguments: %s\n\nOriginal error: %s",
        i,
        paste(names(args), unname(args), sep = "=", collapse = ", "),
        e$message
      ))
    }
  )
}

run_tests = function(
  func,
  argument_grid,
  plot_prefix = "",
  warning_rows = NULL,
  error_rows = NULL,
  ...
) {
  stopifnot(inherits(argument_grid, "data.frame"))
  variant = snapshot_variant()
  for (i in seq_len(nrow(argument_grid))) {
    args = unlist(argument_grid[i, , drop = FALSE], recursive = FALSE)
    test_filename = sprintf("%s_test%i.png", plot_prefix, i)

    args = append(args, ...)
    if (i %in% warning_rows) {
      expect_snapshot_warning(
        do.call(func, args = args),
        variant = variant
      )
    } else if (i %in% error_rows) {
      expect_snapshot_error(
        do.call(func, args = args),
        variant = variant
      )
    }
    if (interactive()) {
      do.call(func, args = args) |>
        expect_no_error_info(args = args, i = i)
    } else {
      skip_unless_snapshot_tests_enabled()
      path = tempfile(fileext = ".png")
      path = suppressWarnings(suppressMessages(
        save_test_png(
          function() {
            do.call(func, args = args)
          },
          path
        )
      ))
      expect_snapshot_file_info(
        path = path,
        name = test_filename,
        compare = compare_image,
        variant = variant,
        args = args,
        i = i
      )
    }
  }
}

test_that("compare_image normalizes RGB and RGBA PNG channel counts", {
  rgb = array(1, dim = c(8, 9, 3))

  rgba = array(1, dim = c(8, 9, 4))
  rgba[,, 4] = 1

  path_rgb = tempfile(fileext = ".png")
  path_rgba = tempfile(fileext = ".png")

  png::writePNG(rgb, target = path_rgb)
  png::writePNG(rgba, target = path_rgba)

  expect_true(compare_image(path_rgb, path_rgba, max_diff = 1e-8))
})

test_that("compare_image returns false rather than erroring on size mismatch", {
  img1 = array(1, dim = c(8, 9, 3))
  img2 = array(1, dim = c(9, 8, 3))

  path1 = tempfile(fileext = ".png")
  path2 = tempfile(fileext = ".png")

  png::writePNG(img1, target = path1)
  png::writePNG(img2, target = path2)

  expect_warning(expect_false(compare_image(path1, path2)), "Dim not the same")
})

#' @title Add Transparent Padding
#'
#' @description Expands an image or matrix canvas by adding constant-value
#' padding around the original image.
#'
#' @param image 3-layer RGB/4-layer RGBA array, `rayimg` class, matrix, or
#' filename of an image.
#' @param pad Default `10`. Padding in pixels. Use a scalar for equal padding on
#' all sides, or `c(top, right, bottom, left)`.
#' @param color Default `"black"`. Padding color. Can be a scalar, RGB/RGBA
#' numeric vector, or R color string.
#' @param alpha Default `NULL`. Padding alpha for images with an alpha channel.
#' If `NULL`, uses the alpha channel supplied in `color` when present, otherwise
#' defaults to `0` when `color` is omitted and `1` when `color` is supplied.
#' Ignored for RGB and matrix inputs.
#' @param filename Default `NULL`. File to save the image to. If `NULL` and
#' `preview = FALSE`, returns the padded image.
#' @param preview Default `FALSE`. If `TRUE`, display the padded image.
#'
#' @return A padded `rayimg` image or matrix.
#' @export
#' @rdname render_padding
#' @examplesIf interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")
#'render_padding(dragon, pad = 20) |>
#'  plot_image()
render_padding = function(
  image,
  pad = 10,
  color = "black",
  alpha = NULL,
  filename = NULL,
  preview = FALSE
) {
  if (is.null(image)) {
    stop("`image` must be provided.")
  }

  color_missing = missing(color)
  color = convert_color(color, single = TRUE)
  color_alpha = if (length(color) >= 4 && !color_missing) {
    color[4]
  } else {
    NULL
  }
  if (is.null(alpha)) {
    alpha = if (!is.null(color_alpha)) {
      color_alpha
    } else if (color_missing) {
      0
    } else {
      1
    }
  }
  if (!is.numeric(alpha) || length(alpha) != 1 || !is.finite(alpha)) {
    stop("`alpha` must be a finite numeric scalar.")
  }
  if (alpha < 0 || alpha > 1) {
    stop("`alpha` must be in [0,1].")
  }
  pad = normalize_render_padding(pad)
  pad_values = unname(pad)

  temp_image = ray_read_image(image, convert_to_array = FALSE)
  image_dim = dim(temp_image)
  channels = if (length(image_dim) == 2) {
    1L
  } else {
    image_dim[3]
  }
  fill = switch(
    as.character(channels),
    `1` = color[1],
    `2` = c(color[1], alpha),
    `3` = color[1:3],
    `4` = c(color[1:3], alpha),
    stop("`image` must have 1, 2, 3, or 4 channels.")
  )

  if (length(image_dim) == 2) {
    output = matrix(
      fill[1],
      nrow = image_dim[1] + pad_values[1] + pad_values[3],
      ncol = image_dim[2] + pad_values[2] + pad_values[4]
    )
    output[
      (pad_values[1] + 1):(pad_values[1] + image_dim[1]),
      (pad_values[4] + 1):(pad_values[4] + image_dim[2])
    ] = unclass(temp_image)
  } else {
    output = array(
      0,
      dim = c(
        image_dim[1] + pad_values[1] + pad_values[3],
        image_dim[2] + pad_values[2] + pad_values[4],
        image_dim[3]
      )
    )
    for (i in seq_len(image_dim[3])) {
      output[,, i] = fill[i]
    }
    output[
      (pad_values[1] + 1):(pad_values[1] + image_dim[1]),
      (pad_values[4] + 1):(pad_values[4] + image_dim[2]),
      seq_len(image_dim[3])
    ] = unclass(temp_image)
  }

  output = rayimg(
    output,
    filetype = attr(temp_image, "filetype"),
    source_linear = attr(temp_image, "source_linear"),
    colorspace = attr(temp_image, "colorspace"),
    white_current = attr(temp_image, "white_current"),
    exposure = attr(temp_image, "exposure", exact = TRUE),
    iso = attr(temp_image, "iso", exact = TRUE)
  )
  attr(output, "padding") = pad
  handle_image_output(output, filename = filename, preview = preview)
}

#' @keywords internal
normalize_render_padding = function(pad) {
  if (!is.numeric(pad) || !(length(pad) %in% c(1, 4))) {
    stop("`pad` must be a numeric scalar or length-4 vector.")
  }
  if (any(!is.finite(pad)) || any(pad < 0)) {
    stop("`pad` must contain non-negative finite values.")
  }
  pad = as.integer(round(pad))
  if (length(pad) == 1) {
    pad = rep(pad, 4)
  }
  names(pad) = c("top", "right", "bottom", "left")
  pad
}

#' Rotate an image array around its center
#'
#' @keywords internal
rotate_image_array = function(img, angle) {
  if (length(dim(img)) < 2) {
    stop("rotate_image_array() expects a matrix or array input")
  }
  if (is.null(angle) || is.na(angle) || !is.finite(angle)) {
    return(img)
  }
  angle = angle %% 360
  if (isTRUE(all.equal(angle, 0))) {
    return(img)
  }
  theta = angle * pi / 180
  cos_t = cos(theta)
  sin_t = sin(theta)

  dims = dim(img)
  if (length(dims) == 2) {
    img = array(img, dim = c(dims, 1))
    dims = dim(img)
  }
  height = dims[1]
  width = dims[2]
  channels = dims[3]

  width_rot = abs(width * cos_t) + abs(height * sin_t)
  height_rot = abs(width * sin_t) + abs(height * cos_t)
  new_width = max(1L, as.integer(ceiling(width_rot)))
  new_height = max(1L, as.integer(ceiling(height_rot)))

  cx_old = (width + 1) / 2
  cy_old = (height + 1) / 2
  cx_new = (new_width + 1) / 2
  cy_new = (new_height + 1) / 2

  x_indices = matrix(rep(seq_len(new_width), each = new_height), nrow = new_height)
  y_indices = matrix(rep(seq_len(new_height), times = new_width), nrow = new_height)

  x_rel = x_indices - cx_new
  y_rel = cy_new - y_indices

  x_src = x_rel * cos_t + y_rel * sin_t + cx_old
  y_src_rel = -x_rel * sin_t + y_rel * cos_t
  y_src = cy_old - y_src_rel

  x_vec = as.vector(x_src)
  y_vec = as.vector(y_src)

  rotated = array(0, dim = c(new_height, new_width, channels))
  for (ch in seq_len(channels)) {
    channel = img[,, ch, drop = TRUE]
    rotated_channel = bilinear_sample_matrix(channel, x_vec, y_vec)
    rotated[,, ch] = matrix(rotated_channel, nrow = new_height, ncol = new_width)
  }
  return(rotated)
}

bilinear_sample_matrix = function(mat, x, y, fill = 0) {
  height = nrow(mat)
  width = ncol(mat)
  result = rep(fill, length(x))
  valid = !is.na(x) & !is.na(y) & x >= 1 & x <= width & y >= 1 & y <= height
  if (!any(valid)) {
    return(result)
  }
  xv = x[valid]
  yv = y[valid]
  x0 = floor(xv)
  y0 = floor(yv)
  x1 = pmin(x0 + 1, width)
  y1 = pmin(y0 + 1, height)
  dx = xv - x0
  dy = yv - y0

  idx00 = cbind(as.integer(y0), as.integer(x0))
  idx10 = cbind(as.integer(y0), as.integer(x1))
  idx01 = cbind(as.integer(y1), as.integer(x0))
  idx11 = cbind(as.integer(y1), as.integer(x1))

  v00 = mat[idx00]
  v10 = mat[idx10]
  v01 = mat[idx01]
  v11 = mat[idx11]

  interp = v00 * (1 - dx) * (1 - dy) +
    v10 * dx * (1 - dy) +
    v01 * (1 - dx) * dy +
    v11 * dx * dy
  result[valid] = interp
  return(result)
}

#'@title Crop Image
#'
#'@description Crops an image (or matrix) either by specifying axis-aligned limits
#'(`width_limits`/`height_limits`) or by specifying a centered box size via `center`.
#'
#'@param image Default `NULL`. 3-layer RGB/4-layer RGBA array, `rayimg` class, or filename of an image.
#'@param width_limits Default `NULL`. Length-2 numeric `(xmin, xmax)` along the width (x/columns).
#' Values in 0..1 are treated as fractions of image width; otherwise they are pixel indices.
#' Mutually exclusive with `center`. If provided alone, `height_limits` defaults to full height.
#'@param height_limits Default `NULL`. Length-2 numeric `(ymin, ymax)` along the height (y/rows).
#' Values in 0..1 are treated as fractions of image height; otherwise they are pixel indices.
#' Mutually exclusive with `center`. If provided alone, `width_limits` defaults to full width.
#'@param center Default `NULL`. Length-2 numeric `(height, width)` for a center-crop.
#' Values in (0,1] are treated as fractions of `(image height, image width)`; otherwise pixels.
#' Mutually exclusive with `width_limits`/`height_limits`.
#'@param filename Default `NULL`. The filename of the image to be saved. If this is not given, the image will be plotted instead.
#'@param preview Default `FALSE`. Whether to plot the cropped image, or just to return the values.
#'@return A `rayimg` array with the cropped region.
#'@export
#'@examplesIf interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")
#'# Left half of the image
#'render_crop(dragon,
#'            width_limits = c(0, 0.50),
#'            height_limits = c(0, 1)) |>
#'  plot_image()
#'# Right half of the image
#'render_crop(dragon,
#'            width_limits = c(0.50, 1),
#'            height_limits = c(0, 1)) |>
#'  plot_image()
#'# Top middle of the image (25%–50% height, 25%–75% width)
#'render_crop(dragon,
#'            height_limits = c(0.25, 0.50),
#'            width_limits  = c(0.25, 0.75)) |>
#'  plot_image()
#'# Top middle, explicit pixel coords (for a 200x200 image)
#'render_crop(dragon,
#'            height_limits = c(50, 100),
#'            width_limits  = c(50, 150)) |>
#'  plot_image()
#'# Center-crop: 50% height and width of the image
#'render_crop(dragon, center = c(0.5, 0.5)) |>
#'  plot_image()
#'# Center-crop: fixed 50x100 pixels
#'render_crop(dragon, center = c(50, 100)) |>
#'  plot_image()
render_crop = function(
  image = NULL,
  width_limits = NULL,
  height_limits = NULL,
  center = NULL,
  filename = NULL,
  preview = FALSE
) {
  if (is.null(image)) {
    stop("`image` must be provided.", call. = FALSE)
  }

  use_limits = !is.null(width_limits) || !is.null(height_limits)
  use_center = !is.null(center)

  if (use_limits && use_center) {
    stop(
      "Provide either `width_limits`/`height_limits` OR `center`, not both.",
      call. = FALSE
    )
  }
  if (!use_limits && !use_center) {
    stop(
      "You must provide `width_limits`/`height_limits`, or `center`.",
      call. = FALSE
    )
  }

  if (!is.null(filename)) {
    if (tools::file_ext(filename) != "png") {
      filename = paste0(filename, ".png")
    }
  }

  # Normalize input and preserve metadata
  temp_image = ray_read_image(image)
  imagetype = attr(temp_image, "filetype")
  img_source_linear = attr(temp_image, "source_linear")
  colorspace = attr(temp_image, "colorspace")
  white_current = attr(temp_image, "white_current")

  h = nrow(temp_image)
  w = ncol(temp_image)

  to_px_range = function(lims, len) {
    if (length(lims) != 2 || !is.numeric(lims)) {
      stop("Limits must be length-2 numeric.", call. = FALSE)
    }
    a = lims[1]
    b = lims[2]
    if (all(a >= 0, a <= 1, b >= 0, b <= 1)) {
      start = floor(min(a, b) * len) + 1L
      end = ceiling(max(a, b) * len)
    } else {
      start = floor(min(a, b))
      end = ceiling(max(a, b))
    }
    start = max(1L, start)
    end = min(len, end)
    if (end < start) {
      stop(
        "Computed limits are empty after clamping; check your inputs.",
        call. = FALSE
      )
    }
    c(start, end)
  }

  if (use_limits) {
    if (is.null(width_limits)) {
      width_limits = c(0, 1)
    }
    if (is.null(height_limits)) {
      height_limits = c(0, 1)
    }

    x_rng = to_px_range(width_limits, w)
    y_rng = to_px_range(height_limits, h)

    x_start = x_rng[1]
    x_end = x_rng[2]
    y_start = y_rng[1]
    y_end = y_rng[2]
  } else {
    if (length(center) != 2 || !is.numeric(center)) {
      stop(
        "`center` must be a length-2 numeric vector (height, width).",
        call. = FALSE
      )
    }
    box_h = center[1]
    box_w = center[2]
    box_h = if (box_h > 0 && box_h <= 1) {
      round(max(1, box_h * h))
    } else {
      round(box_h)
    }
    box_w = if (box_w > 0 && box_w <= 1) {
      round(max(1, box_w * w))
    } else {
      round(box_w)
    }
    box_h = min(max(1L, box_h), h)
    box_w = min(max(1L, box_w), w)

    y_start = floor((h - box_h) / 2) + 1L
    y_end = y_start + box_h - 1L
    x_start = floor((w - box_w) / 2) + 1L
    x_end = x_start + box_w - 1L
  }

  if (imagetype != "matrix") {
    temp_image = temp_image[y_start:y_end, x_start:x_end, , drop = FALSE]
  } else {
    temp_image = temp_image[y_start:y_end, x_start:x_end, drop = FALSE]
  }

  temp_image = ray_read_image(
    temp_image,
    filetype = imagetype,
    source_linear = img_source_linear,
    assume_colorspace = colorspace,
    assume_white = white_current
  )
  handle_image_output(temp_image, filename = filename, preview = preview)
}

rgb_to_hex = function(x) {
	if (is.null(dim(x))) {
		x = matrix(x, ncol = length(x))
	}
	if (!(ncol(x) %in% c(3L, 4L))) {
		stop("rgb_to_hex(): expected 3 or 4 columns (R,G,B[,A]).")
	}
	x = pmin(pmax(x, 0), 1)
	ints = round(x * 255)
	apply(ints, 1, function(row) {
		paste0("#", paste(sprintf("%02X", row), collapse = ""))
	})
}


array_rgb_to_hex_colmajor = function(arr) {
	d = dim(arr)
	R = as.vector(arr[,, 1])
	G = as.vector(arr[,, 2])
	B = as.vector(arr[,, 3])
	if (d[3] >= 4L) {
		A = as.vector(arr[,, 4])
		return(rgb_to_hex(cbind(R, G, B, A)))
	}
	rgb_to_hex(cbind(R, G, B))
}

#' Render Gamma/Linear
#'
#' Convert sRGB-encoded color data to linear light (and vice versa).
#'
#' @param color A `rayimg`, hex color string (or vector), length-3 numeric RGB vector,
#'   or RGB/RGBA matrix.
#' @param srgb_to_linear Default `TRUE`. If `TRUE`, converts sRGB-encoded data to linear light.
#'   If `FALSE`, converts linear-light data to sRGB.
#' @param as_hex Default `FALSE`. When `srgb_to_linear = FALSE` and input is an RGB/RGBA
#'   matrix or array, returns hex: for matrices a length-`n` vector; for arrays a
#'   length-`prod(dim(x)[1:2])` vector in column-major order. Alpha (if present) is passed through.
#'
#' @return Same shape/type as input unless `as_hex = TRUE` on matrix/array with `srgb_to_linear = FALSE`,
#'   in which case a character vector of hex codes is returned.
#' @export
#' @examples
#' # Numeric RGB vector (sRGB -> linear -> back to sRGB)
#' srgb_vec = c(0.2, 0.4, 0.6)
#' linear_vec = render_gamma_linear(srgb_vec)
#' render_gamma_linear(linear_vec, srgb_to_linear = FALSE)
#'
#' # Hex input (single color)
#' linear_from_hex = render_gamma_linear("#336699")
#' render_gamma_linear(linear_from_hex, srgb_to_linear = FALSE, as_hex = TRUE)
#'
#' # Greyscale matrix (treated as image plane)
#' grey_image = matrix(c(0.1, 0.5, 0.9, 0.3), nrow = 2)
#' linear_grey = render_gamma_linear(grey_image)
#' render_gamma_linear(linear_grey, srgb_to_linear = FALSE, as_hex = TRUE)
#'
#' # Array with a single greyscale channel
#' grey_pixels = array(c(0.2, 0.7), dim = c(1, 2, 1))
#' linear_grey_arr = render_gamma_linear(grey_pixels)
#' render_gamma_linear(linear_grey_arr, srgb_to_linear = FALSE, as_hex = TRUE)
#'
#' # Array with greyscale + alpha
#' grey_alpha = array(c(0.3, 0.9, 0.6, 0.4), dim = c(1, 2, 2))
#' linear_grey_alpha = render_gamma_linear(grey_alpha)
#' render_gamma_linear(linear_grey_alpha, srgb_to_linear = FALSE, as_hex = TRUE)
#'
#' # RGB array (3 channels)
#' rgb_pixels = array(
#'   c(
#'     0.2, 0.4, 0.6,
#'     0.7, 0.1, 0.3
#'   ),
#'   dim = c(1, 2, 3)
#' )
#' linear_rgb = render_gamma_linear(rgb_pixels)
#' render_gamma_linear(linear_rgb, srgb_to_linear = FALSE, as_hex = TRUE)
#'
#' # RGBA array (alpha channel preserved)
#' rgba_pixels = array(
#'   c(
#'     0.2, 0.4, 0.6, 0.5,
#'     0.7, 0.1, 0.3, 0.8
#'   ),
#'   dim = c(1, 2, 4)
#' )
#' linear_rgba = render_gamma_linear(rgba_pixels)
#' render_gamma_linear(linear_rgba, srgb_to_linear = FALSE, as_hex = TRUE)
#'
#' # rayimg input retains metadata and supports hex conversion
#' linear_img = render_gamma_linear(ray_read_image(rgba_pixels))
#' render_gamma_linear(linear_img, srgb_to_linear = FALSE, as_hex = TRUE)
render_gamma_linear = function(color, srgb_to_linear = TRUE, as_hex = FALSE) {
	fxn = if (srgb_to_linear) to_linear else to_srgb

	if (inherits(color, "rayimg")) {
		out = color
		d = dim(out)
		if (is.null(d)) {
			out[] = fxn(out[])
		} else {
			arr = unclass(out)
			c = d[3]
			if (length(d) == 2L) {
				arr = fxn(arr) # greyscale matrix
			} else if (length(d) == 3L) {
				if (c >= 1L) {
					arr[,, 1] = fxn(arr[,, 1])
				} # grey or R
				if (c >= 3L) {
					# RGB
					arr[,, 2] = fxn(arr[,, 2])
					arr[,, 3] = fxn(arr[,, 3])
				}
				# alpha (2nd when c==2, 4th when c==4) untouched
			} else {
				stop("render_gamma_linear(): expected 2D matrix or 3D array in rayimg.")
			}
			out[] = arr
		}
		if (!srgb_to_linear && isTRUE(as_hex)) {
			d = dim(out)
			if (length(d) == 2L) {
				g = as.vector(unclass(out))
				return(rgb_to_hex(cbind(g, g, g)))
			}
			if (length(d) == 3L && d[3] >= 1L && d[3] <= 4L) {
				arr = unclass(out)
				if (d[3] == 1L) {
					g = as.vector(arr[,, 1])
					return(rgb_to_hex(cbind(g, g, g)))
				} else if (d[3] == 2L) {
					g = as.vector(arr[,, 1])
					a = as.vector(arr[,, 2])
					return(rgb_to_hex(cbind(g, g, g, a)))
				} else {
					return(array_rgb_to_hex_colmajor(arr))
				}
			}
			stop("render_gamma_linear(): unsupported rayimg shape for hex.")
		}
		attr(out, "source_linear") = srgb_to_linear
		attr(out, "exposure") = 0
		attr(out, "iso") = 100
		return(out)
	}

	if (is.character(color)) {
		if (!length(color)) {
			stop("render_gamma_linear(): character input cannot be empty.")
		}
		if (!srgb_to_linear) {
			stop(
				"render_gamma_linear(): for hex input, only sRGB -> linear is supported."
			)
		}
		rgb = t(grDevices::col2rgb(color)) / 255
		lin = to_linear(rgb)
		return(if (nrow(lin) == 1L) as.numeric(lin[1, ]) else lin)
	}

	if (is.matrix(color)) {
		if (srgb_to_linear) {
			return(fxn(color)) # greyscale plane
		} else {
			srgb_grey = fxn(color)
			if (as_hex) {
				g = as.vector(srgb_grey)
				return(rgb_to_hex(cbind(g, g, g)))
			}
			return(srgb_grey)
		}
	}

	if (is.array(color)) {
		if (length(dim(color)) != 3L) {
			stop("render_gamma_linear(): array must be h*w*c.")
		}
		d = dim(color)
		c = d[3]
		out = color
		if (c >= 1L) {
			out[,, 1] = fxn(out[,, 1])
		}
		if (c >= 3L) {
			out[,, 2] = fxn(out[,, 2])
			out[,, 3] = fxn(out[,, 3])
		}
		if (!srgb_to_linear && as_hex) {
			if (c == 1L) {
				g = as.vector(out[,, 1])
				return(rgb_to_hex(cbind(g, g, g)))
			}
			if (c == 2L) {
				g = as.vector(out[,, 1])
				a = as.vector(out[,, 2])
				return(rgb_to_hex(cbind(g, g, g, a)))
			}
			return(array_rgb_to_hex_colmajor(out))
		}
		return(out)
	}

	if (is.numeric(color)) {
		if (length(color) != 3L) {
			stop("render_gamma_linear(): numeric input must be length-3 RGB.")
		}
		if (any(!is.finite(color)) || any(color < 0 | color > 1)) {
			stop("render_gamma_linear(): numeric RGB values must be between 0 and 1.")
		}
		out = fxn(color)
		if (!srgb_to_linear && as_hex) {
			return(rgb_to_hex(out))
		}
		return(out)
	}

	stop("render_gamma_linear(): unsupported input type.")
}

test_that("render_gamma_linear round-trips numeric vectors", {
	srgb = c(0.2, 0.4, 0.6)
	linear = render_gamma_linear(srgb)
	expect_type(linear, "double")
	expect_equal(
		render_gamma_linear(linear, srgb_to_linear = FALSE),
		srgb,
		tolerance = 1e-7
	)
	expect_equal(
		render_gamma_linear(linear, srgb_to_linear = FALSE, as_hex = TRUE),
		rayimage:::rgb_to_hex(matrix(srgb, ncol = 3, byrow = TRUE))
	)
})

test_that("render_gamma_linear handles hex colors", {
	hex_single = "#123456"
	linear_single = render_gamma_linear(hex_single)
	expect_equal(
		render_gamma_linear(linear_single, srgb_to_linear = FALSE, as_hex = TRUE),
		toupper(hex_single)
	)

	hex_multi = c("#123456", "#ABCDEF")
	linear_multi = render_gamma_linear(hex_multi)
	expected_srgb = t(grDevices::col2rgb(hex_multi) / 255)
	expect_equal(
		render_gamma_linear(linear_multi, srgb_to_linear = FALSE),
		expected_srgb,
		tolerance = 1e-7
	)
	back_hex = vapply(
		seq_len(nrow(linear_multi)),
		function(i) {
			render_gamma_linear(
				linear_multi[i, ],
				srgb_to_linear = FALSE,
				as_hex = TRUE
			)
		},
		character(1)
	)
	expect_equal(toupper(hex_multi), back_hex)
})

test_that("render_gamma_linear round-trips matrices", {
	grey_mat = matrix(c(0.1, 0.5, 0.9, 0.3), nrow = 2)
	linear_mat = render_gamma_linear(grey_mat)
	expect_equal(
		render_gamma_linear(linear_mat, srgb_to_linear = FALSE),
		grey_mat,
		tolerance = 1e-7
	)
	g_vals = as.vector(grey_mat)
	expected_hex = rayimage:::rgb_to_hex(cbind(g_vals, g_vals, g_vals))
	expect_equal(
		render_gamma_linear(linear_mat, srgb_to_linear = FALSE, as_hex = TRUE),
		expected_hex
	)
})

test_that("render_gamma_linear round-trips arrays of all channel counts", {
	# Greyscale
	arr1 = array(c(0.2, 0.6, 0.4, 0.8), dim = c(2, 2, 1))
	lin1 = render_gamma_linear(arr1)
	expect_equal(
		render_gamma_linear(lin1, srgb_to_linear = FALSE),
		arr1,
		tolerance = 1e-7,
		ignore_attr = TRUE
	)
	g_vals = as.vector(arr1[,, 1])
	expected1 = rayimage:::rgb_to_hex(cbind(g_vals, g_vals, g_vals))
	expect_equal(
		render_gamma_linear(lin1, srgb_to_linear = FALSE, as_hex = TRUE),
		expected1
	)

	# Greyscale + alpha
	arr2 = array(
		c(
			0.3,
			0.7,
			0.5,
			0.9,
			0.4,
			0.2,
			0.6,
			0.1
		),
		dim = c(2, 2, 2)
	)
	lin2 = render_gamma_linear(arr2)
	back2 = render_gamma_linear(lin2, srgb_to_linear = FALSE)
	expect_equal(back2[,, 1], arr2[,, 1], tolerance = 1e-7)
	expect_equal(back2[,, 2], arr2[,, 2], tolerance = 1e-7)
	g2 = as.vector(arr2[,, 1])
	a2 = as.vector(arr2[,, 2])
	expected2 = rayimage:::rgb_to_hex(cbind(g2, g2, g2, a2))
	expect_equal(
		render_gamma_linear(lin2, srgb_to_linear = FALSE, as_hex = TRUE),
		expected2
	)

	# RGB
	arr3 = array(
		c(
			0.2,
			0.4,
			0.6,
			0.7,
			0.1,
			0.3,
			0.5,
			0.8,
			0.2,
			0.9,
			0.3,
			0.4
		),
		dim = c(2, 2, 3)
	)
	lin3 = render_gamma_linear(arr3)
	expect_equal(
		render_gamma_linear(lin3, srgb_to_linear = FALSE),
		arr3,
		tolerance = 1e-7,
		ignore_attr = TRUE
	)
	r3 = as.vector(arr3[,, 1])
	g3 = as.vector(arr3[,, 2])
	b3 = as.vector(arr3[,, 3])
	expected3 = rayimage:::rgb_to_hex(cbind(r3, g3, b3))
	expect_equal(
		render_gamma_linear(lin3, srgb_to_linear = FALSE, as_hex = TRUE),
		expected3
	)

	# RGBA
	arr4 = array(
		c(
			0.2,
			0.4,
			0.6,
			0.5,
			0.7,
			0.1,
			0.3,
			0.8,
			0.5,
			0.8,
			0.2,
			0.4,
			0.9,
			0.3,
			0.4,
			0.6
		),
		dim = c(2, 2, 4)
	)
	lin4 = render_gamma_linear(arr4)
	back4 = render_gamma_linear(lin4, srgb_to_linear = FALSE)
	expect_equal(
		back4[,, 1:3, drop = FALSE],
		arr4[,, 1:3, drop = FALSE],
		tolerance = 1e-7
	)
	expect_equal(back4[,, 4], arr4[,, 4], tolerance = 1e-7)
	r4 = as.vector(arr4[,, 1])
	g4 = as.vector(arr4[,, 2])
	b4 = as.vector(arr4[,, 3])
	a4 = as.vector(arr4[,, 4])
	expected4 = rayimage:::rgb_to_hex(cbind(r4, g4, b4, a4))
	expect_equal(
		render_gamma_linear(lin4, srgb_to_linear = FALSE, as_hex = TRUE),
		expected4
	)
})

test_that("render_gamma_linear round-trips rayimg inputs and preserves metadata", {
	rgba_pixels = array(
		c(
			0.2,
			0.4,
			0.6,
			0.5,
			0.7,
			0.1,
			0.3,
			0.8
		),
		dim = c(1, 2, 4)
	)
	img = rayimage:::rayimg(
		rgba_pixels,
		source_linear = FALSE,
		colorspace = CS_SRGB,
		white_current = CS_SRGB$white_xyz
	)

	linear_img = render_gamma_linear(img)
	expect_true(isTRUE(attr(linear_img, "source_linear")))
	expect_equal(
		linear_img[,, 4],
		img[,, 4],
		tolerance = 1e-7,
		ignore_attr = TRUE
	)

	back_img = render_gamma_linear(linear_img, srgb_to_linear = FALSE)
	expect_false(isTRUE(attr(back_img, "source_linear")))
	expect_equal(
		unclass(back_img),
		unclass(img),
		tolerance = 1e-7,
		ignore_attr = TRUE
	)
	expect_identical(attr(back_img, "colorspace"), attr(img, "colorspace"))
	expect_identical(attr(back_img, "white_current"), attr(img, "white_current"))

	r_vals = as.vector(rgba_pixels[,, 1])
	g_vals = as.vector(rgba_pixels[,, 2])
	b_vals = as.vector(rgba_pixels[,, 3])
	a_vals = as.vector(rgba_pixels[,, 4])
	expected_hex = rayimage:::rgb_to_hex(cbind(r_vals, g_vals, b_vals, a_vals))
	hex_from_linear = render_gamma_linear(
		linear_img,
		srgb_to_linear = FALSE,
		as_hex = TRUE
	)
	expect_equal(hex_from_linear, expected_hex)
})

#' Create a rayimg
#'
#' This class will always be:
#'  * A matrix or a 2/3/4 layer array
#'  * It tracks whether the source data was linearized
#'  * It carries working colorspace/white
#'  * It carries informational exposure/ISO metadata
#'
#' @param x Default `NULL`. The underlying array/matrix to wrap.
#' @param filetype Default `NULL`. Original source type to record (e.g., "png").
#' @param source_linear Default `FALSE`. Whether the original source was linearized.
#' @param colorspace Default `CS_ACESCG`. Working space descriptor (see `make_colorspace()`).
#' @param white_current Default `colorspace$white_xyz`. Current assumed scene/display white (XYZ, Y=1).
#' @param exposure Default `0`. Informational exposure compensation in stops.
#' @param iso Default `100`. Informational ISO value.
#'
#' @return An object of class `c("rayimg", <original class>)` with attributes.
#' @keywords internal
rayimg = function(
	x = NULL,
	filetype = NULL,
	source_linear = FALSE,
	colorspace = CS_ACESCG,
	white_current = colorspace$white_xyz,
	exposure = attr(x, "exposure", exact = TRUE),
	iso = attr(x, "iso", exact = TRUE)
) {
	if (is.null(x)) {
		stop("rayimg(): 'x' cannot be NULL.")
	}
	stopifnot(is.array(x))

	x_new = rayimg_validate_dimensions(x)

	# channels attribute
	if (is.null(attr(x, "channels"))) {
		attr(x_new, "channels") = rayimg_detect_channels(x_new)
	} else {
		attr(x_new, "channels") = attr(x, "channels")
	}

	# required attrs
	attr(x_new, "filetype") = filetype
	attr(x_new, "source_linear") = source_linear
	attr(x_new, "exposure") = rayimg_exposure_value(exposure)
	attr(x_new, "iso") = rayimg_iso_value(iso)

	# new attrs for color mgmt
	stopifnot(
		is.list(colorspace),
		all(c("rgb_to_xyz", "xyz_to_rgb", "white_xyz") %in% names(colorspace))
	)
	attr(x_new, "colorspace") = colorspace
	attr(x_new, "white_current") = white_current
	if (!is.null(attr(x, "dng", exact = TRUE))) {
		attr(x_new, "dng") = attr(x, "dng", exact = TRUE)
	}
	class(x_new) = c("rayimg", setdiff(class(x), "rayimg"))
	x_new
}

rayimg_exposure_value = function(exposure) {
	if (is.null(exposure) || length(exposure) == 0L) {
		return(0)
	}
	if (
		!is.numeric(exposure) ||
			length(exposure) != 1L ||
			is.na(exposure) ||
			!is.finite(exposure)
	) {
		stop("rayimg(): 'exposure' must be a finite numeric value.")
	}
	as.numeric(exposure)
}

rayimg_iso_value = function(iso) {
	if (is.null(iso) || length(iso) == 0L) {
		return(100)
	}
	if (
		!is.numeric(iso) ||
			length(iso) != 1L ||
			is.na(iso) ||
			!is.finite(iso) ||
			iso <= 0
	) {
		stop("rayimg(): 'iso' must be a positive finite numeric value.")
	}
	as.numeric(iso)
}

rayimg_channels_from_count = function(n) {
	switch(
		as.character(n),
		"1" = c("Greyscale"),
		"2" = c("Greyscale", "Alpha"),
		"3" = c("Red", "Green", "Blue"),
		"4" = c("Red", "Green", "Blue", "Alpha"),
		stop("This should never be hit.")
	)
}

rayimg_detect_channels = function(x) {
	d = dim(x)
	if (length(d) < 3L) {
		return("Greyscale")
	}
	rayimg_channels_from_count(d[3])
}

#Always return a multidimensional (d[3] = 2,3,4) array, or a matrix.
rayimg_validate_dimensions = function(x) {
	d = dim(x)
	stopifnot(!is.null(d), length(d) %in% c(2, 3), d[1] > 0, d[2] > 0)
	if (length(d) == 3) {
		if (d[3] == 1) {
			x = x[,, 1] # drop extra
		} else {
			stopifnot(d[3] %in% c(2, 3, 4))
		}
	}
	x
}

#' @export
`[.rayimg` = function(x, ..., drop = TRUE) {
	ch = rayimg_detect_channels(x)
	caller_env = parent.frame()
	mc = match.call(expand.dots = FALSE)
	dots_expr = as.list(mc$`...`)
	nd = length(dim(x))

	if (length(dots_expr) < nd) {
		dots_expr = c(dots_expr, rep(list(quote(expr = )), nd - length(dots_expr)))
	}
	idx = lapply(dots_expr, function(z) {
		if (identical(z, quote(expr = ))) TRUE else eval(z, envir = caller_env)
	})

	if (nd >= 3L) {
		if (!isTRUE(idx[[3L]]) && length(idx[[3L]]) > 4L) {
			stop("A rayimg can only have four or fewer channels selected at once.")
		}
		if (!isTRUE(idx[[3L]]) && any(idx[[3L]] > 4L)) {
			stop("rayimg: channel index > 4.")
		}
	}

	y = NextMethod("[", drop = FALSE)

	if (drop) {
		d = dim(y)
		dn = dimnames(y)
		drop_idx = which(d == 1L & seq_along(d) >= 3L)
		if (length(drop_idx)) {
			keep = setdiff(seq_along(d), drop_idx)
			y = array(y, dim = d[keep], dimnames = dn[keep])
		}
	}

	# carry attrs
	attr(y, "filetype") = attr(x, "filetype")
	attr(y, "source_linear") = attr(x, "source_linear")
	attr(y, "exposure") = attr(x, "exposure")
	attr(y, "iso") = attr(x, "iso")
	attr(y, "colorspace") = attr(x, "colorspace")
	attr(y, "white_current") = attr(x, "white_current")
	if (!is.null(attr(x, "dng", exact = TRUE))) {
		attr(y, "dng") = attr(x, "dng", exact = TRUE)
	}

	if (!is.null(ch)) {
		if (nd >= 3L) {
			ch_idx = idx[[3L]]
			if (isTRUE(ch_idx)) {
				ch_idx = seq_along(ch)
			}
			attr(y, "channels") = ch[ch_idx]
		} else {
			attr(y, "channels") = ch[seq_len(min(1L, length(ch)))]
		}
	} else {
		attr(y, "channels") = rayimg_detect_channels(y)
	}

	if (!is.null(dim(y))) {
		class(y) = unique(c("rayimg", setdiff(class(y), "rayimg")))
	}
	y
}

test_that("render_text_image trims glyph bounds and adds padding", {
  old_device = getOption("device")
  on.exit(options(device = old_device), add = TRUE)
  options(
    device = function(...)
      grDevices::pdf(file = tempfile(fileext = ".pdf"), ...)
  )

  args = list(
    text = "A",
    size = 40,
    width = 160,
    height = 160,
    background_alpha = 0,
    check_text_width = FALSE,
    check_text_height = FALSE
  )

  loose = do.call(render_text_image, c(args, list(trim = FALSE)))
  tight = do.call(render_text_image, c(args, list(trim = TRUE)))
  padded = do.call(
    render_text_image,
    c(args, list(trim = TRUE, trim_padding = 5))
  )
  implied_trim = do.call(
    render_text_image,
    c(args, list(trim = FALSE, trim_padding = c(1, 2, 3, 4)))
  )

  expect_true(dim(tight)[1] < dim(loose)[1])
  expect_true(dim(tight)[2] < dim(loose)[2])
  expect_equal(dim(padded)[1:2], dim(tight)[1:2] + 10)
  expect_equal(dim(implied_trim)[1:2], dim(tight)[1:2] + c(4, 6))

  padded_array = unclass(padded)
  expect_true(all(
    padded_array[c(1:5, (nrow(padded_array) - 4):nrow(padded_array)), , 4] == 0
  ))
  expect_true(all(
    padded_array[, c(1:5, (ncol(padded_array) - 4):ncol(padded_array)), 4] == 0
  ))
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/flipfunctions.R
\name{fliplr}
\alias{fliplr}
\title{Flip Left-Right}
\usage{
fliplr(x)
}
\arguments{
\item{x}{Matrix}
}
\value{
Flipped matrix
}
\description{
Flip Left-Right
}
\examples{
#Fake example
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_asp_native_raster.R
\name{plot_asp_native_raster}
\alias{plot_asp_native_raster}
\title{Plot Native Raster with Custom Aspect Ratio}
\usage{
plot_asp_native_raster(
  nr,
  asp = 1,
  show = FALSE,
  return_grob = FALSE,
  gp = grid::gpar(),
  angle = 0
)
}
\description{
Plot Native Raster with Custom Aspect Ratio
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{dragon}
\alias{dragon}
\title{Dragon Image}
\format{
An RGB 3-layer HDR rayimg array with 200 rows and 200 columns,
generated using the rayrender package, with gamma correction.
}
\usage{
dragon
}
\description{
Dragon Image
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_alpha_outline.R
\name{render_alpha_outline}
\alias{render_alpha_outline}
\title{Render Alpha Outline}
\usage{
render_alpha_outline(
  image = NULL,
  mask = NULL,
  expand = 0,
  edge_softness = 0.1,
  blur = 0,
  gap_fill = 1,
  gap_fill_alpha_threshold = 0.25,
  color = "black",
  alpha = 1,
  pad = 0,
  composite = TRUE,
  filename = NULL,
  preview = FALSE
)
}
\arguments{
\item{image}{Default \code{NULL}. Image whose alpha channel will be used as the
mask. If \code{mask} is supplied, \code{image} is only used for metadata.}

\item{mask}{Default \code{NULL}. Optional logical or numeric matrix to use as the
outline mask instead of \code{image} alpha.}

\item{expand}{Default \code{0}. Number of pixels to expand the mask.}

\item{edge_softness}{Default \code{0.1}. Width of the softened halo edge
transition, in pixels.}

\item{blur}{Default \code{0}. Gaussian blur standard deviation applied to the
final outline alpha.}

\item{gap_fill}{Default \code{1}. Maximum alpha gap width, in pixels, to bridge
in the expanded outline alpha. This fills border-connected breaks and notches
caused by discrete distance quantization. Set to \code{0} to disable.}

\item{gap_fill_alpha_threshold}{Default \code{0.25}. Alpha threshold used to
classify border gaps. Gap filling only modifies pixels below this threshold
that are connected to the image exterior; enclosed interior holes surrounded
by pixels at or above this threshold are left unchanged.}

\item{color}{Default \code{"black"}. Outline color.}

\item{alpha}{Default \code{1}. Overall outline alpha multiplier.}

\item{pad}{Default \code{0}. Padding in pixels before computing the outline. If
\code{NULL}, padding is computed from \code{expand}, \code{edge_softness}, and \code{blur}.}

\item{composite}{Default \code{TRUE}. If \code{TRUE} and \code{image} is supplied, composite
the outline below the original image. Set to \code{FALSE} to return the outline
alone.}

\item{filename}{Default \code{NULL}. File to save the image to. If \code{NULL} and
\code{preview = FALSE}, returns the output image.}

\item{preview}{Default \code{FALSE}. If \code{TRUE}, display the outline image.}
}
\value{
A \code{rayimg} RGBA output image with a \code{padding} attribute.
}
\description{
Creates a colored RGBA outline or halo from an image alpha
channel or a supplied mask. When \code{image} is supplied, the original image is
composited over the outline by default.
}
\examples{
\dontshow{if (interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}
txt = render_text_image("\U0001F409",  size = 120, background_alpha = 0)
render_alpha_outline(txt, expand = 2, blur = 1, color = "purple") |>
 plot_image()

#Make a rainbow outline 
render_alpha_outline(txt, expand = 2, blur = 0.1, color = "darkgreen") |>
 render_alpha_outline(expand = 2, blur = 0.1, color = "yellow") |>
 render_alpha_outline(expand = 2, blur = 0.1, color = "orange") |>
 render_alpha_outline(expand = 2, blur = 0.1, color = "red") |>
 render_alpha_outline(expand = 2, blur = 0.1, color = "purple") |>
 render_alpha_outline(expand = 2, blur = 0.1, color = "blue") |>
 plot_image()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colorspace_descriptors.R
\name{make_rgb_xyz_matrices}
\alias{make_rgb_xyz_matrices}
\title{Build RGB-to-XYZ matrices from primaries + white}
\usage{
make_rgb_xyz_matrices(
  primaries = list(r = c(0.713, 0.293), g = c(0.165, 0.83), b = c(0.128, 0.044)),
  white_xy = c(0.32168, 0.33767)
)
}
\arguments{
\item{primaries}{Default \code{list(r=c(0.713,0.293), g=c(0.165,0.830), b=c(0.128,0.044))}. xy chromaticities (AP1).}

\item{white_xy}{Default \code{c(0.32168,0.33767)}. White xy (D60).}
}
\value{
List with 3x3 numeric matrices: \code{rgb_to_xyz}, \code{xyz_to_rgb}.
}
\description{
Build RGB-to-XYZ matrices from primaries + white
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colorspace_helpers.R
\name{to_linear}
\alias{to_linear}
\title{sRGB to Linear}
\usage{
to_linear(x)
}
\arguments{
\item{x}{Values}
}
\value{
Transformed values
}
\description{
sRGB to Linear
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate_2d_gaussian.R
\name{generate_2d_gaussian}
\alias{generate_2d_gaussian}
\title{Generate 2D Gaussian Distribution}
\usage{
generate_2d_gaussian(
  sd = 1,
  power = 1,
  dim = c(11, 11),
  width = 3,
  rescale_unity = FALSE
)
}
\arguments{
\item{sd}{Default \code{1}. Standard deviation of the normal distribution}

\item{power}{Default \code{1}. Power to take the distribution. Higher values will result in a sharper peak.}

\item{dim}{Default \code{c(11, 11)}. The dimensions of the matrix.}

\item{width}{Default \code{3} (\code{-10} to \code{10}). The range in which to compute the distribution.}

\item{rescale_unity}{Default \code{FALSE}. If \code{TRUE}, this will rescale the max value to one. Useful
if wanting to plot the distribution with \code{\link[=plot_image]{plot_image()}}.}
}
\description{
Generates a 2D gaussian distribution, with an optional
argument to take the gaussian to a user-defined power.
}
\examples{
\dontshow{if (interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}
image(generate_2d_gaussian(1,1,31), asp=1)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_file_type.R
\name{get_file_type}
\alias{get_file_type}
\title{Check Filename}
\usage{
get_file_type(file)
}
\arguments{
\item{file}{Filename to be checked}
}
\value{
Filename
}
\description{
Check Filename
}
\examples{
#Fake example
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_alpha_outline.R
\name{offset_alpha}
\alias{offset_alpha}
\title{Offset Alpha Matrix}
\usage{
offset_alpha(alpha, row_offset, col_offset)
}
\arguments{
\item{alpha}{Numeric alpha matrix.}

\item{row_offset}{Integer row offset. Positive values sample from rows below
each output pixel.}

\item{col_offset}{Integer column offset. Positive values sample from columns
to the right of each output pixel.}
}
\value{
Numeric alpha matrix with the same dimensions as \code{alpha}.
}
\description{
Internal shift helper used by \code{fill_alpha_gaps()}. It returns a
same-size alpha matrix shifted by integer row and column offsets, filling
newly exposed pixels with zero instead of wrapping. This lets the gap-filling
scan compare pixels a fixed distance away on opposite sides of a candidate
gap.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_padding.R
\name{render_padding}
\alias{render_padding}
\title{Add Transparent Padding}
\usage{
render_padding(
  image,
  pad = 10,
  color = "black",
  alpha = NULL,
  filename = NULL,
  preview = FALSE
)
}
\arguments{
\item{image}{3-layer RGB/4-layer RGBA array, \code{rayimg} class, matrix, or
filename of an image.}

\item{pad}{Default \code{10}. Padding in pixels. Use a scalar for equal padding on
all sides, or \code{c(top, right, bottom, left)}.}

\item{color}{Default \code{"black"}. Padding color. Can be a scalar, RGB/RGBA
numeric vector, or R color string.}

\item{alpha}{Default \code{NULL}. Padding alpha for images with an alpha channel.
If \code{NULL}, uses the alpha channel supplied in \code{color} when present, otherwise
defaults to \code{0} when \code{color} is omitted and \code{1} when \code{color} is supplied.
Ignored for RGB and matrix inputs.}

\item{filename}{Default \code{NULL}. File to save the image to. If \code{NULL} and
\code{preview = FALSE}, returns the padded image.}

\item{preview}{Default \code{FALSE}. If \code{TRUE}, display the padded image.}
}
\value{
A padded \code{rayimg} image or matrix.
}
\description{
Expands an image or matrix canvas by adding constant-value
padding around the original image.
}
\examples{
\dontshow{if (interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}
render_padding(dragon, pad = 20) |>
 plot_image()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colorspace_helpers.R
\name{get_whitepoint_xyz}
\alias{get_whitepoint_xyz}
\title{Normalize/label a white point}
\usage{
get_whitepoint_xyz(white_point = NULL)
}
\arguments{
\item{white_point}{Default \code{NULL}. Named ("D65","D60","D55","D50","D75","E") or XYZ (length 3, any Y).}
}
\value{
list(value = XYZ with Y=1, white_name = name or "Custom")
}
\description{
Normalize/label a white point
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_reorient.R
\name{render_reorient}
\alias{render_reorient}
\title{Reorient Image}
\usage{
render_reorient(
  image,
  flipx = FALSE,
  flipy = FALSE,
  transpose = FALSE,
  filename = NULL,
  preview = FALSE
)
}
\arguments{
\item{image}{3-layer RGB/4-layer RGBA array, \code{rayimg} class, or filename of an image.}

\item{flipx}{Default \code{FALSE}. Flip horizontally}

\item{flipy}{Default \code{FALSE}. Flip vertically.}

\item{transpose}{Default \code{FALSE}. Transpose image.}

\item{filename}{Default \code{NULL}. The filename of the image to be saved. If this is not given, the image will be plotted instead.}

\item{preview}{Default \code{FALSE}. Whether to plot the convolved image, or just to return the values.}
}
\value{
3-layer RGB reoriented array or matrix.
}
\description{
Reorients an image or matrix. Transformations are applied in this order: x, y, and transpose.
}
\examples{
\dontshow{if (interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}
#Original orientation
plot_image(dragon)
#Flip the dragon image horizontally
dragon |>
 render_reorient(flipx = TRUE) |>
 plot_image()
#Flip the dragon image vertically
dragon |>
 render_reorient(flipy = TRUE) |>
 plot_image()
#Transpose the dragon image
dragon |>
 render_reorient(transpose = TRUE) |>
 plot_image()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_stack.R
\name{render_stack}
\alias{render_stack}
\title{Stack Images}
\usage{
render_stack(
  input_list,
  stack = c("v", "h", "vertical", "horizontal"),
  align = c("center", "start", "end", "left", "right", "top", "bottom"),
  background = c(0, 0, 0, 0),
  convert_colorspace = TRUE,
  filename = NULL,
  preview = FALSE
)
}
\arguments{
\item{input_list}{List of 3-layer RGB/4-layer RGBA array, \code{rayimg} class, or image filenames.}

\item{stack}{Default \code{"v"}. Stack direction: \code{"v"}/\code{"vertical"} or \code{"h"}/\code{"horizontal"}.}

\item{align}{Default \code{"center"}. Alignment on the non-stacking axis. Options:
\code{"start"}/\code{"left"}/\code{"top"}, \code{"center"}, \code{"end"}/\code{"right"}/\code{"bottom"}.}

\item{background}{Default \code{c(0,0,0,0)}. Background color for padding, as numeric
\code{c(r,g,b[,a])} in 0..1 or a color string.}

\item{convert_colorspace}{Default \code{TRUE}. Whether to convert all images to match
the colorspace of the first image in the list.}

\item{filename}{Default \code{NULL}. File to save the image to. If \code{NULL} and \code{preview = FALSE},
returns an RGB array.}

\item{preview}{Default \code{FALSE}. If \code{TRUE}, it will display the image in addition
to returning it.}
}
\value{
A \code{rayimg} RGBA array.
}
\description{
Stacks a list of images vertically or horizontally into a single image.
}
\examples{
\dontshow{if (interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}
render_stack(list(dragon, 1 - dragon), stack = "h") |>
 plot_image()
render_stack(list(dragon, render_bw(dragon)), stack = "v") |>
 plot_image()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colorspace_descriptors.R
\name{make_colorspace}
\alias{make_colorspace}
\title{Build a colorspace descriptor}
\usage{
make_colorspace(
  name = "ACEScg",
  primaries = list(r = c(0.713, 0.293), g = c(0.165, 0.83), b = c(0.128, 0.044)),
  white_xy = c(0.32168, 0.33767),
  white_name = "D60"
)
}
\arguments{
\item{name}{Default \code{"ACEScg"}.}

\item{primaries}{Default AP1.}

\item{white_xy}{Default D60.}
}
\value{
list(name, rgb_to_xyz, xyz_to_rgb, white_xyz, white_name)
}
\description{
Build a colorspace descriptor
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_vignette.R
\name{render_vignette}
\alias{render_vignette}
\title{Add Vignette Effect}
\usage{
render_vignette(
  image,
  vignette = 0.5,
  color = "#000000",
  radius = 1.1,
  filename = NULL,
  preview = FALSE
)
}
\arguments{
\item{image}{3-layer RGB/4-layer RGBA array, \code{rayimg} class, or filename of an image.}

\item{vignette}{Default \code{1}. A camera vignetting effect will be added to the image.
\code{1} is the darkest vignetting, while \code{0} is no vignetting. If vignette is a length-2 vector, the second entry will
control the blurriness of the vignette effect (\code{1} is the default, e.g. \code{2} would double the blurriness but would take
much longer to compute).}

\item{color}{Default \code{"#000000"} (black). Color of the vignette.}

\item{radius}{Default \code{1.1}. Multiplier for the size of the vignette. If \code{1}, the vignette touches
the edge of the image.}

\item{filename}{Default \code{NULL}. Filename which to save the image. If \code{NULL} and \code{preview = FALSE},
returns an RGB array.}

\item{preview}{Default \code{FALSE}. If \code{TRUE}, it will display the image in addition
to returning it.}
}
\value{
A \code{rayimg} RGBA array.
}
\description{
Takes an RGB array/filename and adds a camera vignette effect.
}
\examples{
\dontshow{if (interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}
#Plot the dragon
plot_image(dragon)
#Add a vignette effect:
render_vignette(dragon, preview = TRUE, vignette = 1, radius=1.1)
#Lighten the vignette effect:
render_vignette(dragon, preview = TRUE, vignette = 0.5)
#Change the radius:
render_vignette(dragon, preview = TRUE, vignette = 1, radius=1.5)
render_vignette(dragon, preview = TRUE, vignette = 0.1, radius=0.8)
#Change the color:
render_vignette(dragon, preview = TRUE, vignette = 1, color="white")
#Increase the width of the blur by 50\%:
render_vignette(dragon, preview = TRUE, vignette = c(1,1.5))
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_crop.R
\name{render_crop}
\alias{render_crop}
\title{Crop Image}
\usage{
render_crop(
  image = NULL,
  width_limits = NULL,
  height_limits = NULL,
  center = NULL,
  filename = NULL,
  preview = FALSE
)
}
\arguments{
\item{image}{Default \code{NULL}. 3-layer RGB/4-layer RGBA array, \code{rayimg} class, or filename of an image.}

\item{width_limits}{Default \code{NULL}. Length-2 numeric \verb{(xmin, xmax)} along the width (x/columns).
Values in 0..1 are treated as fractions of image width; otherwise they are pixel indices.
Mutually exclusive with \code{center}. If provided alone, \code{height_limits} defaults to full height.}

\item{height_limits}{Default \code{NULL}. Length-2 numeric \verb{(ymin, ymax)} along the height (y/rows).
Values in 0..1 are treated as fractions of image height; otherwise they are pixel indices.
Mutually exclusive with \code{center}. If provided alone, \code{width_limits} defaults to full width.}

\item{center}{Default \code{NULL}. Length-2 numeric \verb{(height, width)} for a center-crop.
Values in (0,1] are treated as fractions of \verb{(image height, image width)}; otherwise pixels.
Mutually exclusive with \code{width_limits}/\code{height_limits}.}

\item{filename}{Default \code{NULL}. The filename of the image to be saved. If this is not given, the image will be plotted instead.}

\item{preview}{Default \code{FALSE}. Whether to plot the cropped image, or just to return the values.}
}
\value{
A \code{rayimg} array with the cropped region.
}
\description{
Crops an image (or matrix) either by specifying axis-aligned limits
(\code{width_limits}/\code{height_limits}) or by specifying a centered box size via \code{center}.
}
\examples{
\dontshow{if (interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}
# Left half of the image
render_crop(dragon,
           width_limits = c(0, 0.50),
           height_limits = c(0, 1)) |>
 plot_image()
# Right half of the image
render_crop(dragon,
           width_limits = c(0.50, 1),
           height_limits = c(0, 1)) |>
 plot_image()
# Top middle of the image (25\%–50\% height, 25\%–75\% width)
render_crop(dragon,
           height_limits = c(0.25, 0.50),
           width_limits  = c(0.25, 0.75)) |>
 plot_image()
# Top middle, explicit pixel coords (for a 200x200 image)
render_crop(dragon,
           height_limits = c(50, 100),
           width_limits  = c(50, 150)) |>
 plot_image()
# Center-crop: 50\% height and width of the image
render_crop(dragon, center = c(0.5, 0.5)) |>
 plot_image()
# Center-crop: fixed 50x100 pixels
render_crop(dragon, center = c(50, 100)) |>
 plot_image()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_padding.R
\name{pad_to_fit}
\alias{pad_to_fit}
\title{pad to fit}
\usage{
pad_to_fit(dim, kernel)
}
\value{
Hillshade with edges trimmed
}
\description{
Pads to a rect
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cli_helpers.R
\name{rayimg_vertical_ellipsis}
\alias{rayimg_vertical_ellipsis}
\title{Vertical ellipsis symbol}
\usage{
rayimg_vertical_ellipsis(use_cli = TRUE)
}
\arguments{
\item{use_cli}{Default \code{TRUE}. If \code{TRUE}, consult \code{cli::is_utf8_output()} for Unicode support.}
}
\value{
vertical ellipses if supported
}
\description{
Vertical ellipsis symbol
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_to_display.R
\name{render_to_display}
\alias{render_to_display}
\title{Render To Display}
\usage{
render_to_display(image, display = CS_SRGB, encode = TRUE, adapt_white = TRUE)
}
\arguments{
\item{image}{A \code{rayimg}, array, or filename.}

\item{display}{Default \code{CS_SRGB}. Target display RGB space.}

\item{encode}{Default \code{TRUE}. If \code{TRUE}, apply sRGB OETF (only valid when \code{display} is sRGB-like).}

\item{adapt_white}{Default \code{TRUE}. Whether to perform CAT from working white to display white.}
}
\value{
A \code{rayimg} when \code{encode=FALSE}, or a \code{rayimg} with sRGB-encoded RGB when \code{encode=TRUE}.
}
\description{
Convert an image from its working space to a display space, and optionally apply sRGB OETF for viewing.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rayimg.R
\name{rayimg}
\alias{rayimg}
\title{Create a rayimg}
\usage{
rayimg(
  x = NULL,
  filetype = NULL,
  source_linear = FALSE,
  colorspace = CS_ACESCG,
  white_current = colorspace$white_xyz,
  exposure = attr(x, "exposure", exact = TRUE),
  iso = attr(x, "iso", exact = TRUE)
)
}
\arguments{
\item{x}{Default \code{NULL}. The underlying array/matrix to wrap.}

\item{filetype}{Default \code{NULL}. Original source type to record (e.g., "png").}

\item{source_linear}{Default \code{FALSE}. Whether the original source was linearized.}

\item{colorspace}{Default \code{CS_ACESCG}. Working space descriptor (see \code{make_colorspace()}).}

\item{white_current}{Default \code{colorspace$white_xyz}. Current assumed scene/display white (XYZ, Y=1).}

\item{exposure}{Default \code{0}. Informational exposure compensation in stops.}

\item{iso}{Default \code{100}. Informational ISO value.}
}
\value{
An object of class \verb{c("rayimg", <original class>)} with attributes.
}
\description{
This class will always be:
\itemize{
\item A matrix or a 2/3/4 layer array
\item It tracks whether the source data was linearized
\item It carries working colorspace/white
\item It carries informational exposure/ISO metadata
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{sunset_image}
\alias{sunset_image}
\title{Sunset Image}
\format{
An RGBA 4-layer LDR rayimg array from a JPEG image
with 400 rows and 400 columns.
}
\usage{
sunset_image
}
\description{
Sunset Image
}
\author{
Tyler Morgan-Wall
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ray_write_image.R
\name{ray_write_image}
\alias{ray_write_image}
\title{Write Image}
\usage{
ray_write_image(image, filename, clamp = FALSE, write_linear = NA, ...)
}
\arguments{
\item{image}{3-layer RGB/4-layer RGBA array, \code{rayimg} class, or filename of an image.}

\item{filename}{File to write to, with filetype determined by extension. Filetype can be
\code{PNG}, \code{JPEG}, \code{TIFF}, \code{EXR}, or \code{DNG}.}

\item{clamp}{Default \code{FALSE}, automatically determined. Whether to clamp the image to 0-1. If the file extension is \code{PNG} of \code{JPEG},
this is forced to \code{TRUE}.}

\item{write_linear}{Default \code{NA}, automatically determined. By default, images will be gamma corrected (write_linear = FALSE) for
all file types except \code{EXR} and \code{DNG} (which are linear formats), unless otherwise specified.}

\item{...}{Arguments to pass to either \code{jpeg::writeJPEG}, \code{png::writePNG}, \code{libopenexr::write_exr}, or \code{tiff::writeTIFF}.}
}
\value{
A \code{rayimg} RGBA array.
}
\description{
Takes an RGB array/filename and writes it to file.
}
\examples{
\dontshow{if (interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}
#Write as a png
tmparr = tempfile(fileext=".png")
ray_read_image(dragon) |>
 ray_write_image(tmparr)
ray_read_image(tmparr) |>
 plot_image()
#Write as a JPEG (passing quality arguments via ...)
tmparr = tempfile(fileext=".jpg")
ray_read_image(dragon) |>
 ray_write_image(tmparr, quality = 0.2)
ray_read_image(tmparr) |>
 plot_image()
#Write as a tiff
tmparr = tempfile(fileext=".tiff")
ray_read_image(dragon) |>
 ray_write_image(tmparr)
ray_read_image(tmparr) |>
  plot_image()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colorspace_helpers.R
\name{apply_color_matrix}
\alias{apply_color_matrix}
\title{Apply a 3x3 color transform to an image}
\usage{
apply_color_matrix(img, M = diag(3))
}
\arguments{
\item{img}{3/4-channel array (RGB/A). 2D (grey) or 2-channel (grey+alpha) pass through.}

\item{M}{3x3 matrix.}
}
\value{
Array with transformed RGB; alpha preserved.
}
\description{
Apply a 3x3 color transform to an image
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_text_image.R
\name{render_text_image}
\alias{render_text_image}
\title{Generate Text Image}
\usage{
render_text_image(
  text,
  lineheight = 1,
  color = "black",
  size = 12,
  font = "sans",
  just = "left",
  background_color = "white",
  background_alpha = 1,
  use_ragg = TRUE,
  width = NA,
  height = NA,
  filename = NULL,
  check_text_width = TRUE,
  check_text_height = TRUE,
  trim = FALSE,
  trim_padding = 0,
  preview = FALSE
)
}
\arguments{
\item{text}{Text to turn into an image.}

\item{lineheight}{Default \code{1}. Multiplier for the lineheight.}

\item{color}{Default \code{"black"}. String specifying the color of the text.}

\item{size}{Default \code{12}. Numeric value specifying the font size of the text.}

\item{font}{Default \code{"sans"}. String specifying the font family for the text.
Common options include \code{"sans"}, \code{"mono"}, \code{"serif"}, \code{"Times"}, \code{"Helvetica"}, etc.}

\item{just}{Default \code{"left"}. Horizontal alignment of the text: \code{"left"},
\code{"center"}, or \code{"right"}.}

\item{background_color}{Default \code{"white"}. Color of the background.}

\item{background_alpha}{Default \code{1}. Transparency of the background. A value
between \code{0} (fully transparent) and \code{1} (fully opaque).}

\item{use_ragg}{Default \code{TRUE}. Whether to use the \code{ragg} package as the graphics device. Required for emojis.}

\item{width}{Default \code{NA}. User-defined textbox width.}

\item{height}{Default \code{NA}. User-defined textbox height.}

\item{filename}{Default \code{NULL}. String specifying the file path to save the resulting image.
If \code{NULL} and \code{preview = FALSE}, the function returns the processed RGB array.}

\item{check_text_width}{Default \code{TRUE}. Whether to manually adjust the bounding
box of the resulting image to ensure the string bbox is wide enough for the text.
Not all systems provide accurate font sizes: this ensures the string is not cut off
at the edges, at the cost of needing to repeatedly render the image internally until
a suitable image is found.}

\item{check_text_height}{Default \code{FALSE}. Whether to manually adjust the bounding
box of the resulting image to ensure the string bbox is tall enough for the text.
This will ensure a tight vertical bounding box on the text.}

\item{trim}{Default \code{FALSE}. If \code{TRUE}, post-process the rendered image by
cropping it to the tight bounding box of the text or emoji pixels. The
bounding box is measured from a transparent text mask, so this can also trim
images with opaque backgrounds.}

\item{trim_padding}{Default \code{0}. Padding in pixels to add after trimming. Use
a scalar for equal padding on all sides, or \code{c(top, right, bottom, left)}.
Nonzero padding implies \code{trim = TRUE}.}

\item{preview}{Default \code{FALSE}. Boolean indicating whether to display the image after processing.
If \code{TRUE}, the image is displayed but not saved or returned.}
}
\value{
A 3-layer RGB array of the processed image if \code{filename = NULL} and \code{preview = FALSE}.
Otherwise, writes the image to the specified file or displays it if \code{preview = TRUE}.
}
\description{
Generates an image which tightly fits text.
}
\examples{
\dontshow{if (interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}
  #Generate an image of some text
  render_text_image("Some text", preview = TRUE)
  #Change the font size
  render_text_image("Some text", size = 100, preview = TRUE)
  #Change the font color
  render_text_image("Some text", size = 100, color="red",preview = TRUE)
  #Change the background color and transparency
  render_text_image("Some text", size = 50, color="purple",
                      background_color="purple", background_alpha = 0.5,
                      preview = TRUE)
  # Plot an emoji with the agg device.
  render_text_image("\U0001F600\U0001F680", size = 50, color = "purple", use_ragg = TRUE,
                    background_alpha = 0,
                    preview = TRUE)

  # Plot an emoji with the agg device and adjust the height and width (which
  # is on by default) to be a tight fit.
  render_text_image("\U0001F600\U0001F680", size = 50, color = "purple", use_ragg = TRUE,
                    background_alpha = 0, check_text_width = TRUE,
                    check_text_height = TRUE,
                    preview = TRUE)

  # Trim the image to the rendered glyphs and then add transparent padding.
  render_text_image("\U0001F409", size = 80, use_ragg = TRUE,
                    background_alpha = 0, trim = TRUE, trim_padding = 8,
                    preview = TRUE)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cli_helpers.R
\name{rayimg_composite_checker}
\alias{rayimg_composite_checker}
\title{Compose RGB over checkerboard using alpha}
\usage{
rayimg_composite_checker(rgb, alpha = NULL, light = 0.92, dark = 0.75)
}
\description{
Compose RGB over checkerboard using alpha
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_color_correction.R
\name{render_color_correction}
\alias{render_color_correction}
\title{Render Color Correction Matrix (3x3, linear RGB)}
\usage{
render_color_correction(
  image,
  matrix = diag(3),
  filename = NULL,
  preview = FALSE
)
}
\arguments{
\item{image}{Default \code{NULL}. 3-layer RGB/4-layer RGBA array, \code{rayimg}, or filename.}

\item{matrix}{Default \code{diag(3)}. 3x3 numeric matrix.}

\item{filename}{Default \code{NULL}. Output path.}

\item{preview}{Default \code{FALSE}. If \code{TRUE}, display the image.}
}
\value{
A \code{rayimg} RGBA array.
}
\description{
Apply a technical/look 3x3 matrix to an image
}
\examples{
# We will start with an image that's too warm--we want to correct it to
# match the color of the second as closely as possible.
dragon_D50 = render_white_balance(dragon, target_white = "D50", bake = TRUE)
dragon_D75 = render_white_balance(dragon, target_white = "D75", bake = TRUE)
plot_image(dragon_D50)
plot_image(dragon_D75)

# Fucntion to fit a color correction matrix
fit_cc_matrix = function(src, tgt) {
	stopifnot(ncol(src) == 3, ncol(tgt) == 3, nrow(src) == nrow(tgt))
	M_t = apply(tgt, 2, function(y) qr.solve(src, y))
	t(M_t)
}

# Sample N RGBs from an image for fit
rand_samples = function(img, n = 5000, seed=1) {
	set.seed(seed)
	d = dim(img)
	ys = sample.int(d[1], n, TRUE); xs = sample.int(d[2], n, TRUE)
	cbind(img[cbind(ys,xs,1)], img[cbind(ys,xs,2)], img[cbind(ys,xs,3)])
}

S = rand_samples(dragon_D50)
T = rand_samples(dragon_D75)

M = fit_cc_matrix(S, T)
# Optionally, regularize toward identity to avoid overfitting
#  M = 0.8 * M + 0.2 * diag(3)

updated_source = render_color_correction(dragon_D50, matrix = M)

# Matching the old image (grey) to the target image (orange),
# giving a new image (black).
plot(S,pch=16,cex=0.5,col="grey")
points(T,pch=16,cex=0.5,col="orange")
points(rand_samples(updated_source,seed=2),pch=16,cex=0.5,col="black")

# Plot the images
plot_image(render_title(dragon_D50, "OG Image",
           title_color = "white", title_size = 12))
plot_image(render_title(dragon_D75, "Target Image",
           title_color = "white", title_size = 12))
plot_image(render_title(updated_source, "Corrected OG Image",
           title_color = "white", title_size = 12))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colorspace_helpers.R
\name{compute_cat_bradford}
\alias{compute_cat_bradford}
\title{Bradford CAT matrix from source to target white}
\usage{
compute_cat_bradford(
  source_xyz = c(0.95264, 1, 1.00827),
  target_xyz = c(0.95047, 1, 1.08883),
  eps = 1e-08
)
}
\arguments{
\item{source_xyz}{Default \code{c(0.95264,1,1.00827)}. Source XYZ white, Y=1.}

\item{target_xyz}{Default \code{c(0.95047,1,1.08883)}. Target XYZ white, Y=1.}

\item{eps}{Default \code{1e-8}. Small epsilon.}
}
\value{
3x3 numeric CAT matrix to apply in XYZ.
}
\description{
Bradford CAT matrix from source to target white
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_image.R
\name{plot_image}
\alias{plot_image}
\title{Plot Image}
\usage{
plot_image(
  image,
  draw_grid = FALSE,
  ignore_alpha = FALSE,
  asp = 1,
  new_page = TRUE,
  return_grob = FALSE,
  gp = grid::gpar(),
  angle = 0,
  show_linear = FALSE
)
}
\arguments{
\item{image}{3-layer RGB/4-layer RGBA array, \code{rayimg} class, or filename of an image.}

\item{draw_grid}{Default \code{FALSE}. If \code{TRUE}, this will draw a grid in the background to help
disambiguate the actual image from the device (helpful if the image background is the same as the
device's background).}

\item{ignore_alpha}{Default \code{FALSE}. Whether to ignore the alpha channel when plotting.}

\item{asp}{Default \code{1}. Aspect ratio of the pixels in the plot. For example, an aspect ratio of \code{4/3} will
slightly widen the image.}

\item{new_page}{Default \code{TRUE}. Whether to call \code{grid::grid.newpage()} before plotting the image.}

\item{return_grob}{Default \code{FALSE}. Whether to return the grob object.}

\item{gp}{A \code{grid::gpar()} object to include for the grid viewport displaying the image.}

\item{angle}{Default \code{0}. Counter-clockwise rotation (in degrees) applied before plotting.}

\item{show_linear}{Default \code{FALSE}. Most data should be gamma corrected before displaying on a screen.
Set to \code{TRUE} to turn off this gamma correction.}
}
\description{
Displays the image in the current device.
}
\examples{
\dontshow{if (interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}
#Plot the dragon array
plot_image(dragon)
#Make pixels twice as wide as tall
plot_image(dragon, asp = 2)
#Plot non-square images
plot_image(dragon[1:100,,])
#Make pixels twice as tall as wide
plot_image(dragon[1:100,,], asp = 1/2)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_alpha_outline.R
\name{composite_alpha_outline_image}
\alias{composite_alpha_outline_image}
\title{Composite Alpha Outline Image}
\usage{
composite_alpha_outline_image(halo_image, source_image)
}
\arguments{
\item{halo_image}{RGBA outline image used as the compositing backdrop.}

\item{source_image}{RGBA source image, padded to the same dimensions as
\code{halo_image}, used as the compositing foreground.}
}
\value{
A \code{rayimg} RGBA image containing the source composited over the
outline.
}
\description{
Internal helper that composites the original source image over
the generated outline using source-over alpha compositing. The outline is the
backdrop and the source image is the foreground. Colors are treated as
straight, non-premultiplied alpha, using
\code{Ao = Af + Ab * (1 - Af)} and
\code{Co = (Cf * Af + Cb * Ab * (1 - Af)) / Ao}. Fully transparent output pixels
are forced to black to avoid carrying arbitrary color values.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_distance.R
\name{render_boolean_distance}
\alias{render_boolean_distance}
\title{Render Boolean Distance}
\usage{
render_boolean_distance(boolean, rescale = FALSE)
}
\arguments{
\item{boolean}{Logical matrix (or matrix of 1s and 0s), where distance will be measured to the \code{TRUE} values.}

\item{rescale}{Default \code{FALSE}. Rescales the calculated distance to a range of 0-1.
Useful for visualizing the distance matrix.}
}
\value{
Matrix of distance values.
}
\description{
Takes a logical matrix and returns the nearest distance to each TRUE value.
}
\examples{
\dontshow{if (interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}
#Measure distance to
plot_image(render_boolean_distance(t(volcano) > 150))
plot_image(render_boolean_distance(t(volcano) < 150))
#If we want to rescale this to zero to one (to visualize like an image), set rescale=TRUE
plot_image(render_boolean_distance(t(volcano) > 150,rescale=TRUE))
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_alpha_outline.R
\name{fill_alpha_gaps}
\alias{fill_alpha_gaps}
\title{Fill Border Alpha Gaps}
\usage{
fill_alpha_gaps(alpha, max_gap = 1, alpha_threshold = 0.25)
}
\arguments{
\item{alpha}{Numeric alpha matrix.}

\item{max_gap}{Default \code{1}. Maximum low-alpha run width, in pixels, to bridge.}

\item{alpha_threshold}{Default \code{0.25}. Alpha threshold used to decide which
pixels are low-alpha border gaps.}
}
\value{
Numeric alpha matrix with eligible border gaps filled.
}
\description{
Internal post-process for small distance-transform
quantization artifacts on outline edges. The algorithm first calls
\code{find_border_alpha_gaps()} to mark only low-alpha pixels connected to the
image exterior. It then scans horizontal, vertical, and diagonal directions.
For each gap width up to \code{max_gap}, \code{offset_alpha()} samples the alpha values
on opposite sides of each candidate gap. A candidate pixel is raised to the
weaker of the two neighboring alpha values, and only if that pixel was marked
as border-connected. This bridges edge notches while leaving enclosed
interior holes unchanged.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{dragondepth}
\alias{dragondepth}
\title{Dragon Depthmap}
\format{
An matrix with 200 rows and 200 columns, representing
the depth into the \code{dragon} image scene. Generated using the rayrender package.
Distances range from 847 to 1411.
}
\usage{
dragondepth
}
\description{
Dragon Depthmap
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_convolution.R
\name{render_convolution}
\alias{render_convolution}
\title{Render Convolution}
\usage{
render_convolution(
  image,
  kernel = "gaussian",
  kernel_dim = 11,
  kernel_extent = 3,
  absolute = TRUE,
  min_value = NULL,
  include_alpha = FALSE,
  filename = NULL,
  preview = FALSE,
  progress = FALSE
)
}
\arguments{
\item{image}{3-layer RGB/4-layer RGBA array, \code{rayimg} class, or filename of an image.}

\item{kernel}{Default \code{gaussian}. By default, an 11x11 Gaussian kernel with a mean
of \code{0} and a standard deviation of \code{1}, running from \code{-kernel_extent} to \code{kernel_extent}.
If numeric, this will be the standard deviation of the normal distribution. If a
matrix, it will be used directly as the convolution kernel (but resized always to be an odd number
of columns and rows).}

\item{kernel_dim}{Default \code{11}. The dimension of the \code{gaussian} kernel. Ignored
if user specifies their own kernel.}

\item{kernel_extent}{Default \code{3}. Extent over which to calculate the kernel.}

\item{absolute}{Default \code{TRUE}. Whether to take the absolute value of the convolution.}

\item{min_value}{Default \code{NULL}. If numeric, specifies the minimum value (for any color channel)
for a pixel to have the convolution performed.}

\item{include_alpha}{Default \code{FALSE}. Whether to include the alpha channel in the convolution.}

\item{filename}{Default \code{NULL}. The filename of the image to be saved. If this is not given, the image will be plotted instead.}

\item{preview}{Default \code{TRUE}. Whether to plot the convolved image, or just to return the values.}

\item{progress}{Default \code{TRUE}. Whether to display a progress bar.}
}
\value{
A \code{rayimg} RGBA array.
}
\description{
Takes an image and applies a convolution operation to it, using
a user-supplied or built-in kernel. Edges are calculated by limiting the size
of the kernel to only that overlapping the actual image (renormalizing the
kernel for the edges).
}
\examples{
\dontshow{if (interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}
#Perform a convolution with the default gaussian kernel
plot_image(dragon)
#Perform a convolution with the default gaussian kernel
render_convolution(dragon, preview = TRUE)
#Increase the width of the kernel
render_convolution(dragon, kernel = 2, kernel_dim=21,kernel_extent=6, preview = TRUE)
#Perform edge detection using an edge detection kernel
edge = matrix(c(-1,-1,-1,-1,8,-1,-1,-1,-1),3,3)
render_convolution(render_bw(dragon), kernel = edge, preview = TRUE, absolute=FALSE)
#Perform edge detection with Sobel matrices
sobel1 = matrix(c(1,2,1,0,0,0,-1,-2,-1),3,3)
sobel2 = matrix(c(1,2,1,0,0,0,-1,-2,-1),3,3,byrow=TRUE)
sob1 = render_convolution(render_bw(dragon), kernel = sobel1)
sob2 = render_convolution(render_bw(dragon), kernel = sobel2)
sob_all = sob1 + sob2
plot_image(sob1)
plot_image(sob2)
plot_image(sob_all)

#Only perform the convolution on bright pixels (bloom)
render_convolution(dragon, kernel = 5, kernel_dim=24, kernel_extent=24,
                  min_value=1, preview = TRUE)
#Use a built-in kernel:
render_convolution(dragon, kernel = generate_2d_exponential(falloff=2, dim=31, width=21),
                  preview = TRUE)
#We can also apply this function to matrices:
volcano |> image()
volcano |>
 render_convolution(kernel=generate_2d_gaussian(sd=1,dim=31)) |>
 image()
#Use a custom kernel (in this case, an X shape):
custom = diag(10) + (diag(10)[,10:1])
plot_image(custom)
render_convolution(dragon, kernel = custom, preview = TRUE)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_bokeh_size.R
\name{calc_bokeh_size}
\alias{calc_bokeh_size}
\title{Calculate Bokeh Size}
\usage{
calc_bokeh_size(z, zfocus, f, N, ramp)
}
\arguments{
\item{z}{Depth matrix.}

\item{zfocus}{Depth in which to blur. Minimum 0, maximum 1.}

\item{f}{Focal length of the virtual camera. .}

\item{N}{F-stop. Focal length of the virtual camera.}
}
\value{
Matrix of bokeh sizes.
}
\description{
Calculates the amount of blurring at each point with camera characteristics.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_padding.R
\name{trim_padding}
\alias{trim_padding}
\title{trim_padding}
\usage{
trim_padding(input, pad = 1)
}
\arguments{
\item{input}{A two-dimensional matrix.}

\item{pad}{Number of padding entries}
}
\value{
Matrix with edges trimmed
}
\description{
Trims padding
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{print.rayimg}
\alias{print.rayimg}
\title{Print method for rayimg}
\usage{
\method{print}{rayimg}(x = NULL, preview_n = 10, decimals = 3, color = TRUE, ...)
}
\arguments{
\item{x}{Default \code{NULL}. A \code{rayimg} object.}

\item{preview_n}{Default \code{10}. Max rows/cols to display in numeric preview.}

\item{decimals}{Default \code{3}. Number of decimal places to display in numeric preview.}

\item{color}{Default \code{TRUE}. Colorize headers and R/G/B/A numeric cells when supported.}

\item{...}{Default ``. Ignored.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Displays the working colorspace and non-neutral camera metadata alongside a numeric preview.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert_to_native_raster.R
\name{convert_to_native_raster}
\alias{convert_to_native_raster}
\title{Convert to Native Raster}
\usage{
convert_to_native_raster(input)
}
\arguments{
\item{input}{Matrix or image array to convert to a native raster.}
}
\description{
Converts a matrix or image array to a native raster.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_vibrance.R
\name{render_vibrance}
\alias{render_vibrance}
\title{Render Vibrance}
\usage{
render_vibrance(
  image,
  vibrance = 0,
  protect_luminance = 0.25,
  filename = NULL,
  preview = FALSE
)
}
\arguments{
\item{image}{Image filename, 3-layer RGB array, 4-layer RGBA array, or matrix. If a filename,
it will be read via \code{ray_read_image()}.}

\item{vibrance}{Default \code{0}. Numeric scalar in -1..1 (values are clamped). Positive values
increase vibrance primarily in low-saturation regions; negative values decrease vibrance primarily
in high-saturation regions.}

\item{protect_luminance}{Default \code{0.25}. Numeric scalar in 0..1. Reduces the vibrance effect
in deep shadows and bright highlights to avoid harsh color shifts. Set to \code{0} to disable.}

\item{filename}{Default \code{NULL}. If not \code{NULL}, the processed image will be written to this file.}

\item{preview}{Default \code{FALSE}. Whether to plot the processed image.}
}
\value{
4-layer RGBA array (or matrix if passed a matrix and \code{ray_read_image()} is not used).
}
\description{
Adjusts image vibrance (adaptive saturation), similar in spirit to the iPhone "Vibrance" control:
it boosts muted colors more than already-saturated colors, and reduces saturation more strongly
in already-saturated regions when negative. Works on HDR arrays as well (values can exceed 1).
}
\examples{
\dontshow{if (interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}
  dragon |>
    render_vibrance(vibrance = 0.4, preview = TRUE)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make_vignette_overlay.R
\name{make_vignette_overlay}
\alias{make_vignette_overlay}
\title{Make Vignette Overlay}
\usage{
make_vignette_overlay(
  width,
  height,
  intensity = 0.3,
  radius_multiplier = 1.3,
  radius = NULL,
  color = "#000000"
)
}
\arguments{
\item{width}{Width of the image.}

\item{height}{Height of the image.}

\item{intensity}{Default \code{0.4}. \code{1} is max intensity, \code{0} is min.}

\item{radius_multiplier}{Default \code{1.3}.}

\item{radius}{Default \code{NULL}. Max of height or width, divided by 2.}

\item{color}{Default \code{black}.}
}
\description{
Makes an overlay to simulate vignetting in a camera
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_image_overlay.R
\name{render_image_overlay}
\alias{render_image_overlay}
\title{Add Overlay}
\usage{
render_image_overlay(
  image,
  image_overlay = NULL,
  rescale_original = FALSE,
  convert_overlay_colorspace = TRUE,
  alpha = NA,
  filename = NULL,
  preview = FALSE
)
}
\arguments{
\item{image}{3-layer RGB/4-layer RGBA array, \code{rayimg} class, or filename of an image.}

\item{image_overlay}{Default \code{NULL}. 3-layer RGB/4-layer RGBA array, \code{rayimg} class, or filename of an image.
This image will be resized to the dimensions of \code{image} if it does not match
exactly, unless \code{rescale_original = TRUE}.}

\item{rescale_original}{Default \code{FALSE}. If \code{TRUE}, function will resize the original image to match
the overlay.}

\item{convert_overlay_colorspace}{Default \code{TRUE}. Whether to convert the overlay's colorspace
to match the underlying image.}

\item{alpha}{Default \code{NA}, using overlay's alpha channel. Otherwise, this sets the alpha transparency
by multiplying the existing alpha channel by this value (between 0 and 1).}

\item{filename}{Default \code{NULL}. File to save the image to. If \code{NULL} and \code{preview = FALSE},
returns an RGB array.}

\item{preview}{Default \code{FALSE}. If \code{TRUE}, it will display the image in addition
to returning it.}
}
\value{
A \code{rayimg} RGBA array.
}
\description{
Takes an RGB array/filename and composites an image overlay over
the full image. For pixel-positioned sprite, label, glyph, or billboard
overlays, use \code{\link[=render_sprite_overlay]{render_sprite_overlay()}}.
}
\examples{
\dontshow{if (interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}
#Plot the dragon
plot_image(dragon)
#Add an overlay of a red semi-transparent circle:
circlemat = generate_2d_disk(min(dim(dragon)[1:2]))
circlemat = circlemat/max(circlemat)

#Create RGBA image, with a transparency of 0.5
rgba_array = array(1, dim=c(nrow(circlemat),ncol(circlemat),4))
rgba_array[,,1] = circlemat
rgba_array[,,2] = 0
rgba_array[,,3] = 0
dragon_clipped = dragon
dragon_clipped[dragon_clipped > 1] = 1
render_image_overlay(dragon_clipped, image_overlay = rgba_array,
                 alpha=0.5, preview = TRUE)

	# Read photo, convert to ACEScg with CAT (scene)
	photo = ray_read_image(sunset_image, normalize = FALSE)
	photo_aces = render_convert_colorspace(
		photo,
		to_mats = CS_ACESCG,
		adapt_white = TRUE
	)
	tmp_txt = tempfile(fileext = ".png")
	render_text_image(
		"Sunset",
		size = 60,
		filename = tmp_txt,
		color = "#c300ff",
		background_alpha = 0
	)
	# Read logo (display-referred), convert primaries only (no CAT)
	logo = ray_read_image(tmp_txt, normalize = FALSE) # sRGB/D65
	logo_aces = render_convert_colorspace(
		logo,
		to_mats = CS_ACESCG,
		adapt_white = FALSE
	)

	# Composite in ACEScg, then display (plot_image converts to sRGB/D65 + OETF)
	# Here, we also turn overlay conversion in [render_image_overlay()] off,
 # to show what happens when you don't account for the colorspace difference.
 # By default [render_image_overlay()] will do this for you.
	comp1 = render_image_overlay(
		photo_aces,
		logo_aces,
		convert_overlay_colorspace = FALSE
	) |>
   render_title(title_text = "#c300ff: Match",
                title_bar_color = "white", title_color = "#c300ff", title_bar_alpha=1)
	comp2 = render_image_overlay(
		photo_aces,
		logo,
		convert_overlay_colorspace = FALSE
	) |>
	 render_title(title_text = "#c300ff: Incorrect",
               title_bar_color = "white", title_color = "#c300ff", title_bar_alpha=1)

	plot_image_grid(list(comp1, comp2), dim = c(1, 2))
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colorspace_descriptors.R
\name{colorspace_descriptors}
\alias{colorspace_descriptors}
\alias{CS_ACESCG}
\alias{CS_SRGB}
\alias{CS_P3D65}
\alias{CS_BT2020}
\alias{CS_ADOBE}
\title{Prebuilt RGB Working/Display Color Spaces}
\usage{
CS_ACESCG

CS_SRGB

CS_P3D65

CS_BT2020

CS_ADOBE
}
\description{
These objects are \strong{color space descriptors} used by \code{rayimage} to tag images
and to convert between RGB spaces. Each descriptor is a named list with:
\itemize{
\item \code{name}: Character. Human-readable name of the space.
\item \code{primaries}: List with \code{r}, \code{g}, \code{b} entries giving xy chromaticities.
\item \code{rgb_to_xyz}: 3x3 numeric matrix (linear RGB -> CIE XYZ, Y=1 white).
\item \code{xyz_to_rgb}: 3x3 numeric matrix (CIE XYZ -> linear RGB).
\item \code{white_xyz}: Length-3 numeric XYZ of the reference white (normalized s.t. Y = 1).
\item \code{white_name}: Character label for the white (e.g., \code{"D60"}, \code{"D65"}).
}

These are used as:
\itemize{
\item \code{attr(img, "colorspace")}: the current working/display space for an image.
\item \code{attr(img, "white_current")}: the current assumed scene/display white (XYZ, Y=1).
}
}
\details{
Prebuilt RGB Working/Display Color Spaces
\itemize{
\item \strong{CS_ACESCG}: ACEScg (AP1 primaries, D60, scene-linear). Recommended working space.
\item \strong{CS_SRGB}: sRGB/Rec.709 (D65). Typical display space; used for PNG/JPEG/TIFF output.
\item \strong{CS_P3D65}: Display P3 (D65). Wide-gamut display space.
\item \strong{CS_BT2020}: BT.2020 (D65). Very wide-gamut; common for HDR mastering.
\item \strong{CS_ADOBE}: Adobe RGB (D65). Photo workflows.
}
}
\examples{
\dontshow{if (interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}
  # Tag a raw array as sRGB, then convert to ACEScg:
  arr = array(runif(64*64*4), dim = c(64, 64, 4))
  ri  = ray_read_image(arr, assume_colorspace = CS_SRGB, assume_white = CS_SRGB$white_xyz)
  riA = render_convert_colorspace(ri, to_mats = CS_ACESCG)
  plot_image(riA) # display-converted to sRGB automatically

  # Convert an sRGB JPEG to ACEScg on ingest:
  # img = ray_read_image("photo.jpg", normalize = FALSE)
  # img_aces = render_convert_colorspace(img, to_mats = CS_ACESCG)
\dontshow{\}) # examplesIf}
}
\seealso{
\itemize{
\item \code{\link[=render_convert_colorspace]{render_convert_colorspace()}} to convert images between spaces.
\item \code{\link[=render_white_balance]{render_white_balance()}} for Bradford CAT within a working space.
\item \code{\link[=plot_image]{plot_image()}}, \code{\link[=ray_write_image]{ray_write_image()}} for display/output conversion.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_clamp.R
\name{render_clamp}
\alias{render_clamp}
\title{Clamp Image}
\usage{
render_clamp(
  image,
  min_value = 0,
  max_value = 1,
  filename = NA,
  preview = FALSE,
  ...
)
}
\arguments{
\item{image}{3-layer RGB/4-layer RGBA array, \code{rayimg} class, or filename of an image.}

\item{min_value}{Default \code{0}. Minimum value to clamp the RGB channels in the image to.}

\item{max_value}{Default \code{1}. Maximum value to clamp the RGB channels in the image to.}

\item{filename}{Default \code{NA}. Filename}

\item{preview}{Default \code{FALSE}. If \code{TRUE}, it will display the image in addition
to returning it.}

\item{...}{Arguments to pass to either \code{jpeg::readJPEG}, \code{png::readPNG}, or \code{tiff::readTIFF}.}
}
\value{
A \code{rayimg} RGBA array.
}
\description{
Clamps an image to a user-specified range, ignoring the alpha channel.
}
\examples{
\dontshow{if (interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}
#The image of the unchanged image
range(dragon)
#Clamp the maximum and minimum values to one and zero
render_clamp(dragon) |>
 range()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interpolate_array.R
\name{interpolate_array}
\alias{interpolate_array}
\title{Matrix/Array Interpolation}
\usage{
interpolate_array(image, x, y)
}
\arguments{
\item{image}{3-layer RGB/4-layer RGBA array, \code{rayimg} class, or filename of an image.}

\item{x}{X indices (or fractional index) to interpolate.}

\item{y}{Y indices (or fractional index) to interpolate.}
}
\value{
Either a vector of values (if image is a matrix) or a list of interpolated values
from each layer.
}
\description{
Given a series of X and Y coordinates and an array/matrix, interpolates the Z coordinate
using bilinear interpolation.
}
\examples{
#if(interactive()){
#Interpolate a matrix
interpolate_array(volcano,c(10,10.1,11),c(30,30.5,33))
#Interpolate a 3-layer array (returns list for each channel)
interpolate_array(dragon,c(10,10.1,11),c(30,30.5,33))
#end}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colorspace_helpers.R
\name{to_srgb}
\alias{to_srgb}
\title{Linear to sRGB}
\usage{
to_srgb(x)
}
\arguments{
\item{x}{Values}
}
\value{
Transformed values
}
\description{
Linear to sRGB
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_title.R
\name{render_title}
\alias{render_title}
\title{Render a Title on an Image}
\usage{
render_title(
  image,
  title_text = "",
  title_size = 30,
  title_offset = rep(title_size/2, 2),
  title_lineheight = 1,
  title_color = "black",
  title_font = "Arial",
  title_style = "plain",
  title_bar_color = NA,
  title_bar_alpha = 0.5,
  title_bar_width = NULL,
  title_position = NA,
  title_just = "left",
  use_magick = FALSE,
  filename = NULL,
  preview = FALSE
)
}
\arguments{
\item{image}{3-layer RGB/4-layer RGBA array, \code{rayimg} class, or filename of an image.}

\item{title_text}{Default \code{""}. Text string to be added as the title to the image.}

\item{title_size}{Default \code{30}. Numeric value specifying the font size of the title text.}

\item{title_offset}{Default \code{c(15,15)}. Numeric vector specifying the horizontal
and vertical offset of the title text, relative to its anchor position.}

\item{title_lineheight}{Default \code{1}. Multiplier for the lineheight.}

\item{title_color}{Default \code{"black"}. String specifying the color of the title text.}

\item{title_font}{Default \code{"Arial"}. String specifying the font family for the title text.
Common options include \code{"sans"}, \code{"mono"}, \code{"serif"}, \code{"Times"}, \code{"Helvetica"}, etc.}

\item{title_style}{Default \code{"plain"}. String specifying the font style, such as
\code{"plain"}, \code{"italic"}, or \code{"bold"}.}

\item{title_bar_color}{Default \code{NULL}. Color of the optional title bar. If \code{NULL}, no bar is added.}

\item{title_bar_alpha}{Default \code{0.5}. Transparency level of the title bar. A value
between \code{0} (fully transparent) and \code{1} (fully opaque).}

\item{title_bar_width}{Default \code{NULL}. Numeric value for the height of the title bar in pixels.
If \code{NULL}, it is automatically calculated based on the text size and line breaks.}

\item{title_position}{Default \code{"northwest"}. String specifying the position of the title text.
Only used when \code{use_magick = TRUE}. Common options include \code{"northwest"}, \code{"center"}, \code{"south"}, etc.}

\item{title_just}{Default \code{"left"}. Horizontal alignment of the title text: \code{"left"},
\code{"center"}, or \code{"right"}.}

\item{use_magick}{Default \code{FALSE}. Boolean indicating whether to use the \code{magick} package for
rendering titles. This option will be deprecated in future versions.}

\item{filename}{Default \code{NULL}. String specifying the file path to save the resulting image.
If \code{NULL} and \code{preview = FALSE}, the function returns the processed RGB array.}

\item{preview}{Default \code{FALSE}. Boolean indicating whether to display the image after processing.
If \code{TRUE}, the image is displayed but not saved or returned.}
}
\value{
A 3-layer RGB array of the processed image if \code{filename = NULL} and \code{preview = FALSE}.
Otherwise, writes the image to the specified file or displays it if \code{preview = TRUE}.
}
\description{
Adds a title with optional styling and a title bar to an image.
The image can be previewed or saved to a file. Supports both the \code{grid}-based
method and (deprecated) \code{magick} package for rendering the title.
}
\note{
The \code{use_magick} parameter and all functionality tied to the \code{magick} package are
planned for deprecation. It is recommended to use the \code{grid}-based method for
future compatibility.
}
\examples{
\dontshow{if (interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}
#Plot the dragon
render_title(dragon, preview = TRUE, title_text = "Dragon", title_size=20)
#That's hard to see--let's add a title bar:
render_title(dragon, preview = TRUE, title_text = "Dragon", title_size=20,
         title_bar_color="white")
#Change the width of the bar:
render_title(dragon, preview = TRUE, title_text = "Dragon", title_size=20,
         title_bar_color="white", title_offset = c(8,8))
#The width of the bar will also automatically adjust for newlines:
render_title(dragon, preview = TRUE, title_text = "Dragon\n(Blue)", title_size=20,
         title_bar_color="white")
#Change the color and title color:
render_title(dragon, preview = TRUE, title_text = "Dragon", title_size=20,
         title_bar_color="red", title_color = "white")
#Change the transparency:
render_title(dragon, preview = TRUE, title_text = "Dragon",
            title_size=20, title_bar_alpha = 0.8,
            title_bar_color="red", title_color = "white")
#Read directly from a file
temp_image = tempfile(fileext = ".png")
ray_write_image(dragon, temp_image)
render_title(temp_image, preview = TRUE, title_text = "Dragon",
             title_size=20, title_bar_alpha = 0.8,
             title_bar_color="red", title_color = "white")
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colorspace_helpers.R
\name{array_rgb_to_hex_colmajor}
\alias{array_rgb_to_hex_colmajor}
\title{Array hex vector in column-major order}
\usage{
array_rgb_to_hex_colmajor(arr)
}
\arguments{
\item{arr}{Array with c>=3}
}
\value{
Hex vector length h*w
}
\description{
Array hex vector in column-major order
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_tonemap.R
\name{render_tonemap}
\alias{render_tonemap}
\title{Render Tonemap}
\usage{
render_tonemap(
  image,
  method = c("raw", "reinhard", "uncharted", "hbd"),
  exposure_bias = 2,
  W = 11.2,
  filename = NULL,
  preview = FALSE
)
}
\arguments{
\item{image}{Default \code{NULL}. 3/4-channel array, \code{rayimg}, or filename.}

\item{method}{Default \code{raw}. One of \code{"raw"}, \code{"reinhard"}, \code{"uncharted"}, \code{"hbd"}.
Chooses the global tone-mapping curve. \code{"raw"} leaves the image unchanged.
\code{"reinhard"} applies a classic shoulder that gently compresses highlights.
\code{"uncharted"} (Hable filmic) gives a film-style toe/shoulder and keeps mid-tones contrasty.
\code{"hbd"} is a fast filmic-like curve with a soft toe/shoulder.
All methods operate on \strong{linear RGB} and return \strong{display-referred linear} (no OETF).}

\item{exposure_bias}{Default \code{2}. Only for \code{"uncharted"}.
Scalar applied to the linear RGB channels \strong{before} the Uncharted/Hable curve
(i.e., a pre-curve exposure gain). Values >1 brighten; <1 darken.
\code{exposure_bias=2} is about +1 stop; \code{0.5} is about -1 stop.}

\item{W}{Default \code{11.2}. White (linear scene value) that maps to 1.0 \strong{after}
the Uncharted/Hable/Reinhard curve. Acts like a “white point” for the shoulder: larger values
preserve more highlight detail (later roll-off), smaller values roll off earlier and harder.}

\item{filename}{Default \code{NULL}. Output path.}

\item{preview}{Default \code{FALSE}. Show result.}
}
\value{
A \code{rayimg} RGBA array.
}
\description{
Reinhard / Uncharted(Hable) / HBD operators in linear
}
\examples{
# Plot unchanged image
render_tonemap(dragon, preview = TRUE)
# Plot Reinhard tonemapped image
render_tonemap(dragon, method = "reinhard", preview = TRUE)
# Plot Uncharted/Hable tonemapped image
render_tonemap(dragon, method = "uncharted",preview = TRUE)
# Plot Uncharted/Hable tonemapped image, white point adjusted
render_tonemap(dragon, method = "uncharted", W=11.4, preview = TRUE)
# Plot Uncharted/Hable tonemapped image, exposure adjusted
render_tonemap(dragon, method = "uncharted", exposure_bias = 1,preview = TRUE)
# Plot Hejl-Burgess-Dawson tonemapped image, exposure adjusted
render_tonemap(dragon, method = "hbd",preview = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_image_overlay.R
\name{render_sprite_overlay}
\alias{render_sprite_overlay}
\title{Add Sprite Overlay}
\usage{
render_sprite_overlay(
  image,
  image_overlay = NULL,
  convert_overlay_colorspace = TRUE,
  alpha = NA,
  overlay_coords = c(1, 1),
  overlay_dims = NULL,
  overlay_anchor = "nw",
  overlay_just = NULL,
  hjust = NULL,
  vjust = NULL,
  preserve_channels = TRUE,
  filename = NULL,
  preview = FALSE
)
}
\arguments{
\item{image}{3-layer RGB/4-layer RGBA array, \code{rayimg} class, or filename of an image.}

\item{image_overlay}{Default \code{NULL}. 3-layer RGB/4-layer RGBA array, \code{rayimg}
class, or filename of the sprite overlay.}

\item{convert_overlay_colorspace}{Default \code{TRUE}. Whether to convert the overlay's colorspace
to match the underlying image.}

\item{alpha}{Default \code{NA}, using overlay's alpha channel. Otherwise, this sets the alpha transparency
by multiplying the existing alpha channel by this value (between 0 and 1).}

\item{overlay_coords}{Default \code{c(1, 1)}. Pixel coordinate \code{c(x, y)} used to
position the overlay. \code{x} increases from left to right and \code{y} from top to
bottom, starting at 1.}

\item{overlay_dims}{Default \code{NULL}. Dimensions for the overlay in pixels. If
provided, the overlay is resized with \code{render_resized()} before compositing.}

\item{overlay_anchor}{Default \code{"nw"}. Which corner of the overlay is placed at
\code{overlay_coords} when no numeric justification is supplied. Options: \code{"nw"},
\code{"ne"}, \code{"sw"}, \code{"se"}.}

\item{overlay_just}{Default \code{NULL}. Optional numeric \code{c(hjust, vjust)}
justification used to place the overlay relative to \code{overlay_coords}. If set,
the overlay's left/top pixel is computed as
\code{round(x - hjust * overlay_width)} and
\code{round(y - vjust * overlay_height)}.}

\item{hjust}{Default \code{NULL}. Optional horizontal justification. Overrides the
first value of \code{overlay_just}.}

\item{vjust}{Default \code{NULL}. Optional vertical justification. Overrides the
second value of \code{overlay_just}.}

\item{preserve_channels}{Default \code{TRUE}. If \code{TRUE}, RGB input images return
RGB output and RGBA input images return RGBA output.}

\item{filename}{Default \code{NULL}. File to save the image to. If \code{NULL} and
\code{preview = FALSE}, returns an image array.}

\item{preview}{Default \code{FALSE}. If \code{TRUE}, it will display the image in addition
to returning it.}
}
\value{
A \code{rayimg} array.
}
\description{
Places an image overlay at a pixel location without resizing it to
the full destination image. This is intended for sprite-like or billboard-like
overlays such as labels, icons, glyphs, and screen-space annotations.
}
\examples{
\dontshow{if (interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}
dragon_clipped = dragon
dragon_clipped[dragon_clipped > 1] = 1

# Render a transparent dragon emoji and center it on a pixel with hjust/vjust:
dragon_emoji = render_text_image(
 "\U0001F409",
 size = 80,
 background_alpha = 0,
 use_ragg = TRUE,
 trim = TRUE,
 trim_padding = 8
)
render_sprite_overlay(
 dragon_clipped,
 image_overlay = dragon_emoji,
 overlay_coords = c(ncol(dragon_clipped) / 2, nrow(dragon_clipped) / 2),
 hjust = 0.5,
 vjust = 0.5,
 preview = TRUE
)

# The same justification can be supplied as overlay_just = c(hjust, vjust):
render_sprite_overlay(
 dragon_clipped,
 image_overlay = dragon_emoji,
 overlay_coords = c(ncol(dragon_clipped), nrow(dragon_clipped)),
 overlay_just = c(1, 1),
 preview = TRUE
)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_alpha_outline.R
\name{find_border_alpha_gaps}
\alias{find_border_alpha_gaps}
\title{Find Border Alpha Gaps}
\usage{
find_border_alpha_gaps(alpha, alpha_threshold = 0.25)
}
\arguments{
\item{alpha}{Numeric alpha matrix.}

\item{alpha_threshold}{Default \code{0.25}. Pixels below this alpha are considered
low-alpha gap pixels.}
}
\value{
Logical matrix marking low-alpha pixels connected to the image
exterior.
}
\description{
Internal helper that classifies which low-alpha pixels are part
of the exterior border region. Pixels below \code{alpha_threshold} are considered
low alpha. The algorithm starts from low-alpha pixels on the image boundary
and performs an 8-connected flood fill through other low-alpha pixels.
Low-alpha regions enclosed by pixels at or above \code{alpha_threshold} are not
reached, so they are protected from \code{fill_alpha_gaps()}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_bokeh.R
\name{render_bokeh}
\alias{render_bokeh}
\title{Render Bokeh}
\usage{
render_bokeh(
  image,
  depthmap,
  focus = 0.5,
  focallength = 100,
  fstop = 4,
  filename = NULL,
  preview = TRUE,
  preview_focus = FALSE,
  bokehshape = "circle",
  bokehintensity = 1,
  bokehlimit = 0.8,
  rotation = 0,
  aberration = 0,
  progress = interactive(),
  ...
)
}
\arguments{
\item{image}{3-layer RGB/4-layer RGBA array, \code{rayimg} class, or filename of an image.}

\item{depthmap}{Depth map filename or 1d array.}

\item{focus}{Defaults \code{0.5}. Depth in which to blur.}

\item{focallength}{Default \code{100}. Focal length of the virtual camera.}

\item{fstop}{Default \code{4}. F-stop of the virtual camera.}

\item{filename}{Default \code{NULL}. The filename of the image to be saved. If this is not given, the image will be plotted instead.}

\item{preview}{Default \code{TRUE}. If \code{FALSE}, it will not display the image and just return the RGB array.}

\item{preview_focus}{Default \code{FALSE}. If \code{TRUE}, a red line will be drawn across the image
showing where the camera will be focused.}

\item{bokehshape}{Default \code{circle}. Also built-in: \code{hex}. The shape of the bokeh. If the user
passes in a 2D matrix, that matrix will control the shape of the bokeh.}

\item{bokehintensity}{Default \code{1}. Intensity of the bokeh when the pixel intensity is greater than \code{bokehlimit}.}

\item{bokehlimit}{Default \code{0.8}. Limit after which the bokeh intensity is increased by \code{bokehintensity}.}

\item{rotation}{Default \code{0}. Number of degrees to rotate the hexagonal bokeh shape.}

\item{aberration}{Default \code{0}. Adds chromatic aberration to the image. Maximum of \code{1}.}

\item{progress}{Default \code{TRUE}. Whether to display a progress bar.}

\item{...}{Additional arguments to pass to \code{\link[=plot_image]{plot_image()}} if \code{preview = TRUE}.}
}
\value{
A \code{rayimg} RGBA array.
}
\description{
Takes an image and a depth map to render the image with depth of field
(i.e. similar to "Portrait Mode" in an iPhone). User can specify a custom bokeh shape,
or use one of the built-in bokeh types.
}
\examples{
\dontshow{if (interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}
#Plot the dragon
plot_image(dragon)
#Plot the depth map
plot_image(dragondepth/1500)
#Preview the focal plane:
render_bokeh(dragon, dragondepth, focus=950, preview_focus = TRUE)
#Change the focal length:
render_bokeh(dragon, dragondepth, focus=950, focallength=300)
#Add chromatic aberration:
render_bokeh(dragon, dragondepth, focus=950, focallength=300, aberration = 0.5)
#Change the focal distance:
render_bokeh(dragon, dragondepth, focus=600, focallength=300)
render_bokeh(dragon, dragondepth, focus=1300, focallength=300)
#Change the bokeh shape to a hexagon:
render_bokeh(dragon, dragondepth, bokehshape = "hex",
            focallength=300, focus=600)
#Change the bokeh intensity:
render_bokeh(dragon, dragondepth,
            focallength=400, focus=900, bokehintensity = 1)
render_bokeh(dragon, dragondepth,
            focallength=400, focus=900, bokehintensity = 3)
#Rotate the hexagonal shape:
render_bokeh(dragon, dragondepth, bokehshape = "hex", rotation=15, bokehintensity=3,
            focallength=400, focus=800)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ray_read_image.R
\name{ray_read_image}
\alias{ray_read_image}
\title{Read Image}
\usage{
ray_read_image(
  image,
  convert_to_array = FALSE,
  preview = FALSE,
  source_linear = NA,
  normalize = FALSE,
  dng_normalize = FALSE,
  normalize_to = CS_ACESCG,
  normalize_adapt_white = TRUE,
  assume_colorspace = NULL,
  assume_white = NULL,
  reset_camera_settings = FALSE,
  ...
)
}
\arguments{
\item{image}{3-layer RGB/4-layer RGBA array, \code{rayimg} class, or filename of an image.}

\item{convert_to_array}{Default \code{FALSE}. Whether to convert 2D B&W images/matrices to RGBA arrays.}

\item{preview}{Default \code{FALSE}. If \code{TRUE}, it will display the image in addition
to returning it.}

\item{source_linear}{Default \code{NA}, automatically determined based on the image type.
Whether the image source is linear data or sRGB. \code{TRUE} for matrices, arrays, and
floating-point HDR formats (\code{EXR}, \code{HDR}), and \code{DNG}; \code{FALSE} for all other file-based formats
(e.g., \code{jpeg}, \code{png}, \code{tiff}, \code{tga}, \code{bmp}, \code{psd}, \code{gif}, \code{pic}, \code{pnm}).}

\item{normalize}{Default \code{FALSE}. If \code{TRUE}, convert to \code{normalize_to} space on read. Note that
rayimg inputs will keep their colorspace and ignore this option--use \code{render_convert_colorspace()}
to change an existing rayimg to a new colorspace.}

\item{dng_normalize}{Default \code{FALSE}. If \code{TRUE}, DNG data is normalized to 0..1,
using black/white levels; if \code{FALSE}, raw values are returned.}

\item{normalize_to}{Default \code{CS_ACESCG}. Target colorspace when \code{normalize=TRUE}.}

\item{normalize_adapt_white}{Default \code{TRUE}. If \code{TRUE}, Bradford-adapt source white to target white.}

\item{assume_colorspace}{Default \code{NULL}. A colorspace descriptor (e.g., \code{CS_SRGB},
\code{CS_ACESCG}). If given, the loaded image will be \emph{tagged} with this space
rather than the default inferred by file type. No pixel conversion is performed
unless \code{normalize=TRUE}.}

\item{assume_white}{Default \code{NULL}. Scene/display white for the loaded image.
Either a named white (\verb{"D60","D65","D50","D55","D75","E"}) or XYZ with Y=1.
If \code{NULL}, uses \code{assume_colorspace$white_xyz}.}

\item{reset_camera_settings}{Default \code{FALSE}. If \code{TRUE}, reset informational
exposure/ISO metadata to neutral values (\code{exposure = 0}, \code{iso = 100}) when
wrapping an existing \code{rayimg} or attributed array.}

\item{...}{Arguments to pass to \code{jpeg::readJPEG}, \code{png::readPNG}, \code{tiff::readTIFF},
or \code{libopenexr::read_exr()} for supported formats.}
}
\value{
A \code{rayimg} RGBA array.
}
\description{
Reads an image from a file/array/matrix. From files, supports
\code{JPEG}, \code{PNG}, \code{TIFF}, \code{TGA}, \code{BMP}, \code{PSD}, \code{GIF}, \code{HDR}, \code{PIC}, \code{PNM}, \code{EXR},
and \code{DNG} images.
}
\examples{
\dontshow{if (interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}
#Write as a png
tmparr = tempfile(fileext=".png")
ray_read_image(dragon) |>
 ray_write_image(tmparr)
ray_read_image(tmparr) |>
 plot_image()
#Write as a JPEG (passing quality arguments via ...)
tmparr = tempfile(fileext=".jpg")
ray_read_image(dragon) |>
 ray_write_image(tmparr, quality = 0.2)
ray_read_image(tmparr) |>
 plot_image()
#Write as a tiff
tmparr = tempfile(fileext=".tiff")
ray_read_image(dragon) |>
 ray_write_image(tmparr)
ray_read_image(tmparr) |>
  plot_image()

#Write as an exr
tmparr = tempfile(fileext=".exr")
ray_read_image(dragon) |>
 ray_write_image(tmparr)
ray_read_image(tmparr) |>
  plot_image()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colorspace_helpers.R
\name{rgb_to_hex}
\alias{rgb_to_hex}
\title{Encode RGB/RGBA to hex (sRGB domain)}
\usage{
rgb_to_hex(x)
}
\arguments{
\item{x}{Values. 3 or 4 columns (R,G,B,A) in 0 to 1.}
}
\value{
Hex vector (#RRGGBB or #RRGGBBAA).
}
\description{
Encode RGB/RGBA to hex (sRGB domain)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_bw.R
\name{render_bw}
\alias{render_bw}
\title{Render Black and White}
\usage{
render_bw(
  image,
  rgb_coef = c(0.2126, 0.7152, 0.0722),
  filename = NULL,
  preview = FALSE
)
}
\arguments{
\item{image}{3-layer RGB/4-layer RGBA array, \code{rayimg} class, or filename of an image.}

\item{rgb_coef}{Default \code{c(0.2126, 0.7152, 0.0722)}.
Length-3 numeric vector listing coefficients to convert RGB to luminance.}

\item{filename}{Default \code{NULL}. The filename of the image to be saved. If this is not given, the image will be plotted instead.}

\item{preview}{Default \code{FALSE}. Whether to plot the convolved image, or just to return the values.}
}
\value{
A \code{rayimg} RGBA array.
}
\description{
Transforms an image to black and white, preserving luminance.
}
\examples{
\dontshow{if (interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}
#Plot the image with a title
dragon |>
 render_title("Dragon", title_offset=c(10,10), title_bar_color="black",
           title_size=20, title_color = "white") |>
 render_bw(preview = TRUE)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_gamma_linear.R
\name{render_gamma_linear}
\alias{render_gamma_linear}
\title{Render Gamma/Linear}
\usage{
render_gamma_linear(color, srgb_to_linear = TRUE, as_hex = FALSE)
}
\arguments{
\item{color}{A \code{rayimg}, hex color string (or vector), length-3 numeric RGB vector,
or RGB/RGBA matrix.}

\item{srgb_to_linear}{Default \code{TRUE}. If \code{TRUE}, converts sRGB-encoded data to linear light.
If \code{FALSE}, converts linear-light data to sRGB.}

\item{as_hex}{Default \code{FALSE}. When \code{srgb_to_linear = FALSE} and input is an RGB/RGBA
matrix or array, returns hex: for matrices a length-\code{n} vector; for arrays a
length-\code{prod(dim(x)[1:2])} vector in column-major order. Alpha (if present) is passed through.}
}
\value{
Same shape/type as input unless \code{as_hex = TRUE} on matrix/array with \code{srgb_to_linear = FALSE},
in which case a character vector of hex codes is returned.
}
\description{
Convert sRGB-encoded color data to linear light (and vice versa).
}
\examples{
# Numeric RGB vector (sRGB -> linear -> back to sRGB)
srgb_vec = c(0.2, 0.4, 0.6)
linear_vec = render_gamma_linear(srgb_vec)
render_gamma_linear(linear_vec, srgb_to_linear = FALSE)

# Hex input (single color)
linear_from_hex = render_gamma_linear("#336699")
render_gamma_linear(linear_from_hex, srgb_to_linear = FALSE, as_hex = TRUE)

# Greyscale matrix (treated as image plane)
grey_image = matrix(c(0.1, 0.5, 0.9, 0.3), nrow = 2)
linear_grey = render_gamma_linear(grey_image)
render_gamma_linear(linear_grey, srgb_to_linear = FALSE, as_hex = TRUE)

# Array with a single greyscale channel
grey_pixels = array(c(0.2, 0.7), dim = c(1, 2, 1))
linear_grey_arr = render_gamma_linear(grey_pixels)
render_gamma_linear(linear_grey_arr, srgb_to_linear = FALSE, as_hex = TRUE)

# Array with greyscale + alpha
grey_alpha = array(c(0.3, 0.9, 0.6, 0.4), dim = c(1, 2, 2))
linear_grey_alpha = render_gamma_linear(grey_alpha)
render_gamma_linear(linear_grey_alpha, srgb_to_linear = FALSE, as_hex = TRUE)

# RGB array (3 channels)
rgb_pixels = array(
  c(
    0.2, 0.4, 0.6,
    0.7, 0.1, 0.3
  ),
  dim = c(1, 2, 3)
)
linear_rgb = render_gamma_linear(rgb_pixels)
render_gamma_linear(linear_rgb, srgb_to_linear = FALSE, as_hex = TRUE)

# RGBA array (alpha channel preserved)
rgba_pixels = array(
  c(
    0.2, 0.4, 0.6, 0.5,
    0.7, 0.1, 0.3, 0.8
  ),
  dim = c(1, 2, 4)
)
linear_rgba = render_gamma_linear(rgba_pixels)
render_gamma_linear(linear_rgba, srgb_to_linear = FALSE, as_hex = TRUE)

# rayimg input retains metadata and supports hex conversion
linear_img = render_gamma_linear(ray_read_image(rgba_pixels))
render_gamma_linear(linear_img, srgb_to_linear = FALSE, as_hex = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_padding.R
\name{add_multi_padding}
\alias{add_multi_padding}
\title{add_multi_padding}
\usage{
add_multi_padding(input, pad = 1)
}
\arguments{
\item{input}{A two-dimensional matrix or a 3D array (including \code{rayimg}).}

\item{pad}{Number of padding entries}
}
\value{
Matrix with edges padded
}
\description{
Adds multiple levels padding to the matrix
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preview_focus.R
\name{preview_focus}
\alias{preview_focus}
\title{Preview Focus}
\usage{
preview_focus(image, depthmap, focus, imagetype, depthtype)
}
\arguments{
\item{image}{3-layer RGB/4-layer RGBA array, \code{rayimg} class, or filename of an image.}

\item{depthmap}{Depth map filename or 1d array.}

\item{focus}{Defaults \code{0.5}. Depth in which to blur. Minimum 0, maximum 1.}
}
\description{
Displays the focal point
}
\examples{
#Plot the dragon
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/handle_image_output.R
\name{handle_image_output}
\alias{handle_image_output}
\title{Handle Image Output}
\usage{
handle_image_output(image, filename = NULL, preview = FALSE)
}
\arguments{
\item{image}{3-layer RGB/4-layer RGBA array, \code{rayimg} class, or filename of an image.}

\item{filename}{Default \code{NULL}. String specifying the file path to save the image.
If \code{NULL}, the image is returned or displayed.}

\item{preview}{Default \code{FALSE}. If \code{TRUE}, the image is displayed.}
}
\value{
If \code{filename = NULL} and \code{preview = FALSE}, returns the image array invisibly.
Otherwise, saves to file or displays the image.
}
\description{
Handles image output logic: returning, saving to a file, or previewing.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert_color.R
\name{convert_color}
\alias{convert_color}
\title{Convert Color}
\usage{
convert_color(color, as_hex = FALSE, single = FALSE)
}
\arguments{
\item{color}{The color to convert. Can be either an R color string, or a
numeric scalar, RGB vector, RGBA vector, or RGB/RGBA matrix with values
between \code{0} and \code{1}.}

\item{as_hex}{Default \code{FALSE}. If \code{TRUE}, return hex colors.}

\item{single}{Default \code{FALSE}. If \code{TRUE}, require exactly one color.}
}
\value{
RGB or RGBA color vector, or a color matrix for multiple colors.
}
\description{
Convert Color
}
\examples{
#none
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_resized.R
\name{render_resized}
\alias{render_resized}
\title{Resize Image}
\usage{
render_resized(
  image,
  mag = 1,
  dims = NULL,
  filename = NULL,
  preview = FALSE,
  method = "tri"
)
}
\arguments{
\item{image}{3-layer RGB/4-layer RGBA array, \code{rayimg} class, or filename of an image.}

\item{mag}{Default \code{1}. Amount to magnify the image, preserving aspect ratio. Overridden if
\code{dim} is not \code{NULL}.}

\item{dims}{Default \code{NULL}. Exact resized dimensions.}

\item{filename}{Default \code{NULL}. The filename of the image to be saved. If this is not given, the image will be plotted instead.}

\item{preview}{Default \code{FALSE}. Whether to plot the convolved image, or just to return the values.}

\item{method}{Default \code{trilinear}. Filters to up/downsample the image. Options: \code{bilinear}, \code{box}, \code{trilinear},
\code{catmull}, \code{mitchell}.}
}
\value{
A \code{rayimg} RGBA array.
}
\description{
Resizes an image or a matrix, using bilinear interpolation.
}
\examples{
\dontshow{if (interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}
#Plot the image with a title
dragon |>
 render_title("Dragon", title_offset=c(10,10), title_bar_color="black",
           title_size=20, title_color = "white") |>
 plot_image()
#Half of the resolution
render_resized(dragon, mag = 1/2) |>
 render_title("Dragon (half res)", title_offset=c(5,5), title_bar_color="black",
           title_size=10, title_color = "white") |>
 plot_image()
#Double the resolution
render_resized(dragon, mag = 2) |>
 render_title("Dragon (2x res)", title_offset=c(20,20), title_bar_color="black",
           title_size=40, title_color = "white") |>
 plot_image()
#Specify the exact resulting dimensions
render_resized(dragon, dim = c(160,320)) |>
 render_title("Dragon (custom size)", title_offset=c(10,10), title_bar_color="black",
           title_size=20, title_color = "white") |>
 plot_image()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_string_dimensions.R
\name{get_string_dimensions}
\alias{get_string_dimensions}
\title{Get String Dimensions}
\usage{
get_string_dimensions(string, font = "sans", size = 12, align = "center", ...)
}
\arguments{
\item{string}{The string to be measured.}

\item{font}{Default \code{"sans"}.}

\item{size}{A numeric value specifying the size of the font in points.}

\item{align}{Default `"left"``. The string alignment.}

\item{...}{Other arguments to pass to `systemfonts::shape_string()``}
}
\value{
A data.frame listing the string dimensions.
}
\description{
Calculates font metrics for a specified font, font size, and style.
}
\details{
The function renders specific characters (\code{"d"} for ascender, \code{"g"} for descender, \code{"x"} for neutral) using the specified font parameters. It calculates the bounding box of each character to determine the necessary adjustments for accurate text positioning.
}
\examples{
# Get height of basic sans font
get_string_dimensions("This is a string", size=24)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rotate_image_array.R
\name{rotate_image_array}
\alias{rotate_image_array}
\title{Rotate an image array around its center}
\usage{
rotate_image_array(img, angle)
}
\description{
Rotate an image array around its center
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate_2d_disk.R
\name{generate_2d_disk}
\alias{generate_2d_disk}
\title{Generate 2D Disk}
\usage{
generate_2d_disk(dim = c(11, 11), radius = 1, rescale_unity = FALSE)
}
\arguments{
\item{dim}{Default \code{c(11, 11)}. The dimensions of the matrix.}

\item{radius}{Default \code{1}. Radius of the disk, compared to the dimensions. Should be less than one.}

\item{rescale_unity}{Default \code{FALSE}. If \code{TRUE}, this will rescale the max value to one. Useful
if wanting to plot the distribution with \code{\link[=plot_image]{plot_image()}}.}
}
\description{
Generates a 2D disk with a gradual falloff.

Disk generated using the following formula:

\ifelse{html}{\out{(-22.35 * cos(1.68 * r<sup>2</sup>) + 85.91 * sin(1.68 * r<sup>2</sup>) ) * exp(-4.89 * r<sup>2</sup>) +
(35.91 * cos(4.99 * r<sup>2</sup>) - 28.87 * sin(4.99 * r<sup>2</sup>)) * exp(-4.71 * r<sup>2</sup>) +
(-13.21 * cos(8.24 * r<sup>2</sup>) - 1.57 * sin(8.24 * r<sup>2</sup>)) * exp(-4.05 * r<sup>2</sup>) +
(0.50 * cos(11.90 * r<sup>2</sup>) + 1.81 * sin(11.90 * r<sup>2</sup>)) * exp(-2.92 * r<sup>2</sup>) +
(0.13 * cos(16.11 * r<sup>2</sup>) - 0.01 * sin(16.11 * r<sup>2</sup>)) * exp(-1.51 * r<sup>2</sup>)}}{(-22.35 cos(1.68 r^2) + 85.91 sin(1.68 r^2) ) exp(-4.89 r^2) +
(35.91 cos(4.99 r^2) - 28.87 sin(4.99 r^2)) exp(-4.71 r^2) +
(-13.21 cos(8.24 r^2) - 1.57 sin(8.24 r^2)) exp(-4.05 r^2) +
(0.50 cos(11.90 r^2) + 1.81 sin(11.90 r^2)) exp(-2.92 r^2) +
(0.13 cos(16.11 r^2) - 0.01 sin(16.11 r^2)) exp(-1.51 r^2)}

The origin of the coordinate system is the center of the matrix.
}
\examples{
\dontshow{if (interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}
image(generate_2d_disk(101), asp=1)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linear_png_device.R
\name{linear_png_device}
\alias{linear_png_device}
\title{Select PNG Device}
\usage{
linear_png_device(use_ragg = TRUE)
}
\arguments{
\item{use_ragg}{Default \code{TRUE}. If \code{TRUE}, return \code{ragg::agg_png()}.
Set to \code{FALSE} to skip \code{ragg}.}
}
\description{
Internal helper to pick the best available PNG graphics
device. Uses \code{ragg::agg_png()} by default, then falls back to a
Cairo-backed \code{grDevices::png()} device when \code{use_ragg = FALSE}, and
finally to the base PNG device.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_convert_colorspace.R
\name{render_convert_colorspace}
\alias{render_convert_colorspace}
\title{Render Convert Colorspace}
\usage{
render_convert_colorspace(
  image,
  from_mats = NA,
  to_mats = CS_ACESCG,
  adapt_white = TRUE,
  from_white = NA,
  to_white = NA,
  filename = NULL,
  preview = FALSE
)
}
\arguments{
\item{image}{3-layer RGB/4-layer RGBA array, \code{rayimg}, or filename.}

\item{from_mats}{Default \code{NA}. Source space object; when \code{NA}, pulled from \code{attr(src,"colorspace")}.}

\item{to_mats}{Default \code{CS_ACESCG}. Target space object.}

\item{adapt_white}{Default \code{TRUE}. Apply Bradford CAT from \code{from_white} to \code{to_white}.}

\item{from_white}{Default \code{NA}. Source white (name or XYZ, Y=1). When \code{NA}, uses \code{attr(src,"white_current")}.}

\item{to_white}{Default \code{NA}. Target white (name or XYZ, Y=1). When \code{NA}, uses \code{to_mats$white_xyz}.}

\item{filename}{Default \code{NULL}. Output path.}

\item{preview}{Default \code{FALSE}. If \code{TRUE}, display the image.}
}
\value{
A \code{rayimg} RGBA array tagged with the \strong{target} space.
}
\description{
Convert between two RGB spaces via XYZ, with optional white adaptation. Operates on \strong{linear} RGB.
}
\examples{
\dontshow{if (interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}
# Read photo, convert to ACEScg with CAT (scene)
	photo = ray_read_image(sunset_image, normalize = FALSE)
	photo_aces = render_convert_colorspace(
		photo,
		to_mats = CS_ACESCG,
		adapt_white = TRUE
	)
	tmp_txt = tempfile(fileext = ".png")
	render_text_image(
		"Sunset",
		size = 60,
		filename = tmp_txt,
		color = "#c300ffff",
		background_alpha = 0
	)
	# Read logo (display-referred), convert primaries only (no CAT)
	logo = ray_read_image(tmp_txt, normalize = FALSE) # sRGB/D65
	logo_aces = render_convert_colorspace(
		logo,
		to_mats = CS_ACESCG,
		adapt_white = FALSE
	)

	# Composite in ACEScg, then display (plot_image converts to sRGB/D65 + OETF)
	# Here, we also turn overlay conversion in [render_image_overlay()] off,
 # to show what happens when you don't account for the colorspace difference.
 # By default [render_image_overlay()] will do this for you.
	comp1 = render_image_overlay(
		photo_aces,
		logo_aces,
		convert_overlay_colorspace = FALSE
	)
	comp2 = render_image_overlay(
		photo_aces,
		logo,
		convert_overlay_colorspace = FALSE
	)
  #Note the color differences, which arise from the mismatched colorspace on the right.
	 plot_image_grid(list(comp1, comp2), dim = c(1, 2))
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_image_grid.R
\name{plot_image_grid}
\alias{plot_image_grid}
\title{Plot Image Grid}
\usage{
plot_image_grid(
  input_list,
  dim = c(1, 1),
  asp = 1,
  draw_grid = FALSE,
  gp = grid::gpar()
)
}
\arguments{
\item{input_list}{List of 3-layer RGB/4-layer RGBA array, \code{rayimg} class, or image filenames.}

\item{dim}{Default \code{c(1,1)}. Width by height of output grid.}

\item{asp}{Default \code{1}. Aspect ratio of the pixels(s). For example, an aspect ratio of \code{4/3} will
slightly widen the image. This can also be a vector the same length of
\code{input_list} to specify an aspect ratio for each image in the grid.}

\item{draw_grid}{Default \code{FALSE}. If \code{TRUE}, this will draw a grid in the background to help
disambiguate the actual image from the device (helpful if the image background is the same as the
device's background).}

\item{gp}{A \code{grid::gpar()} object to include for the grid viewport displaying the image.}
}
\description{
Displays the image in the current device.
}
\examples{
\dontshow{if (interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}
#Plot the dragon array
plot_image_grid(list(dragon, 1-dragon), dim = c(1,2))
plot_image_grid(list(dragon, 1-dragon), dim = c(2,1))
plot_image_grid(list(dragon, NULL, 1-dragon), dim = c(2,2), asp = c(2,1,1/2))
plot_image_grid(list(dragon, NULL, NULL, dragon), dim = c(2,2), asp = c(2,1,1,1/2))
#Plot alongside the depth matrix
dragon_depth_reoriented = render_reorient(dragondepth,
                                         transpose = TRUE,
                                         flipx = TRUE)/2000
plot_image_grid(list(dragondepth/2000, dragon, dragon, dragondepth/2000),
               dim = c(2,2))
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_exposure.R
\name{render_exposure}
\alias{render_exposure}
\title{Apply exposure compensation and ISO gain}
\usage{
render_exposure(
  image,
  exposure = 0,
  iso = NA,
  auto = FALSE,
  percentile = 0.99,
  verbose = FALSE,
  filename = NA,
  preview = FALSE,
  ...
)
}
\arguments{
\item{image}{3-layer RGB/4-layer RGBA array, \code{rayimg} class, or filename of an image.}

\item{exposure}{Default \code{0}. Exposure compensation in stops; applied directly to RGB values.}

\item{iso}{Default \code{NA}. ISO gain to apply directly to RGB values. \code{NA} applies no ISO gain
and preserves the current informational ISO metadata.}

\item{auto}{Default \code{FALSE}. If \code{TRUE}, automatically add the exposure needed to
place a luminance percentile at 1.0 for LDR display/output. \code{exposure}
is then treated as an additional compensation value.}

\item{percentile}{Default \code{0.99}. Luminance percentile used when \code{auto = TRUE}.
Must be greater than 0 and less than or equal to 1. Use \code{1} for max luminance.}

\item{verbose}{Default \code{FALSE}. If \code{TRUE} and \code{auto = TRUE}, print the
automatically computed exposure adjustment.}

\item{filename}{Default \code{NA}. Output path.}

\item{preview}{Default \code{FALSE}. If \code{TRUE}, display the image.}

\item{...}{Additional args passed to \code{\link[=plot_image]{plot_image()}} (when \code{preview=TRUE})
or to \code{\link[=ray_write_image]{ray_write_image()}} (when \code{filename} is given).}
}
\value{
A \code{rayimg} RGBA array.
}
\description{
Apply exposure compensation and ISO gain
}
\examples{
\dontshow{if (interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}
# LDR/sRGB (auto): decodes to linear, applies EV, records camera metadata
  render_exposure(dragon, exposure = +1, preview = TRUE)
# Force linear/HDR behavior
  render_exposure(dragon * 2, exposure = -1, preview = TRUE)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_white_balance.R
\name{render_white_balance}
\alias{render_white_balance}
\title{Render White Balance (Bradford CAT)}
\usage{
render_white_balance(
  image,
  reference_white = NA,
  target_white = "D60",
  bake = FALSE,
  filename = NULL,
  preview = FALSE
)
}
\arguments{
\item{image}{Default \code{NULL}. 3-layer RGB/4-layer RGBA array, \code{rayimg}, or filename.}

\item{reference_white}{Default \code{NA}. Source white (XYZ Y=1 or named); when \code{NA}, uses \code{attr(img, "white_current")}.}

\item{target_white}{Default \code{"D60"}. Target white (XYZ Y=1 or named).}

\item{bake}{Default \code{FALSE}. If \code{TRUE}, this will bake the color change into the image, and return the
old tag. This can be used for stylistic adjustments.}

\item{filename}{Default \code{NULL}. Output path.}

\item{preview}{Default \code{FALSE}. If \code{TRUE}, display the image.}
}
\description{
Chromatic-adapt an image from \code{reference_white} to \code{target_white} \strong{in its working RGB space}.

When \code{bake = FALSE} (the default), this should not result in any visible changes to the image when
plotted in rayimage via \code{\link[=plot_image]{plot_image()}}, as that function then corrects the white balance for
display. Set
}
\examples{
# Not specifying a target white balance returns an unchanged image
render_white_balance(dragon, preview = TRUE)

# Default white "D60", converting to D50 produces a warmer image
render_white_balance(dragon, preview = TRUE, target_white = "D50", bake = TRUE)

# Default white "D60", converting to "D75" produces a cooler image
render_white_balance(dragon, preview = TRUE, target_white = "D75", bake = TRUE)

# Set reference white to colder "D75", converting to "D50" produces a very warm image
render_white_balance(dragon, preview = TRUE,
                     reference_white= "D75", target_white = "D50", bake = TRUE)

# With a real, non-rendered image:
render_white_balance(sunset_image, preview=TRUE)

# Cooler
render_white_balance(sunset_image, preview=TRUE, target_white = "D75", bake = TRUE)

# Warmer
render_white_balance(sunset_image, preview=TRUE, target_white = "D50", bake = TRUE)
}
